#include "MdlImport.h"
#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <simpobj.h>
#include "../Shared/StdLib.h"
#include "../Shared/Memory.h"
#include "../Shared/String.h"
#include "../Shared/StringBuffer.h"
#include "../Shared/FileHelpers.h"
#include "../Shared/Buffer.h"
#include "../Shared/Config.h"
#include "resource.h"
#include <ilayer.h>
#include <ilayermanager.h>
#include "SR3IL.h"

#define SR3IMP_FIXAXIS

extern HINSTANCE	g_hInstance;

bool SplitPath(const char* pName, CStringA* pPath, CString* pFileName, CStringA* pExtension){
	TStringBufferA<MAX_PATH> tmp, buf;
	const char* s, *s2;
	tmp = pName;
	s = strrchr(tmp.GetBufferPtr(), '/');
	s2 = strrchr(tmp.GetBufferPtr(), '\\');
	if (s2 > s)
		s = s2;
	if (s){
		tmp.Left(buf, (int)(s-tmp.GetBufferPtr()));
		//tmp = buf;
		if (pPath)
			*pPath = buf;
		s++;
	} else {
		if (pPath)
			*pPath = "";
		s = pName;
	}
	tmp = s;
	s = strrchr(tmp.GetBufferPtr(), '.');
	if (s){
		tmp.Left(buf, (int)(s-tmp.GetBufferPtr()));
		if (pFileName)
			*pFileName = buf;
		if (pExtension)
			*pExtension = s+1;
	} else {
		if (pFileName)
			*pFileName = buf;
		if (pExtension)
			*pExtension = "";
	}
	return true;
}

//_M_ISUNICODE

const MCHAR* CStringConvert::UTF8ToM(const char* pStr){
#ifdef _M_ISUNICODE
	m_WBuf = pStr;
	return m_WBuf;
#else //_M_ISUNICODE
	return pStr;
#endif //!_M_ISUNICODE
}

const MCHAR* CStringConvert::UTF16ToM(const wchar_t* pStr){
#ifdef _M_ISUNICODE
	return pStr;
#else //_M_ISUNICODE
	m_ABuf = pStr;
	return m_ABuf;
#endif //!_M_ISUNICODE
}

const char* CStringConvert::MToUTF8(const MCHAR* pStr){
#ifdef _M_ISUNICODE
	m_ABuf = pStr;
	return m_ABuf;
#else //_M_ISUNICODE
	return pStr;
#endif //!_M_ISUNICODE
}

const wchar_t* CStringConvert::MToUTF16(const MCHAR* pStr){
#ifdef _M_ISUNICODE
	return pStr;
#else //_M_ISUNICODE
	m_WBuf = pStr;
	return m_WBuf;
#endif //!_M_ISUNICODE
}

TriObject* CreateTriObjectAndAddToRoot(INode** ppNode = NULL){
	INode *pNode;
	TriObject *pTri;
	pTri = CreateNewTriObject();
	if (!pTri)
		return NULL;
	pNode = GetCOREInterface()->CreateObjectNode(pTri);
	if (!pNode){
		pTri->DeleteThis();
		return NULL;
	}
	GetCOREInterface()->GetRootNode()->AttachChild(pNode);
	if (ppNode)
		*ppNode = pNode;
	return pTri;
}

class CSkin{
public:
	ISkinImportData*	m_pImportData;
	ISkin*				m_pSkin;
	INode*				m_pNode;
	bool Create(INode* pNode){
		IDerivedObject* pDObj;
		Modifier* pMod;
		Object* pObj;
		m_pNode = pNode;

		pMod = (Modifier*)GetCOREInterface()->CreateInstance(OSM_CLASS_ID, /*SKIN_CLASSID*/ Class_ID(9815843,87654));
		if (!pMod)
			return false;

		pObj = pNode->GetObjectRef();
		if (!pObj)
			return true;
		if (pObj->SuperClassID() == GEN_DERIVOB_CLASS_ID)
			pDObj = (IDerivedObject*)pObj;
		else {
			pDObj = CreateDerivedObject(pObj);
			if (!pDObj){
				pMod->DeleteThis();
				return false;
			}
			pNode->SetObjectRef(pDObj);
		}
		pDObj->SetAFlag(A_LOCK_TARGET);
		pDObj->AddModifier(pMod);
		pDObj->ClearAFlag(A_LOCK_TARGET);

		m_pSkin = (ISkin*)pMod->GetInterface(I_SKIN);
		m_pImportData = (ISkinImportData*)pMod->GetInterface(I_SKINIMPORTDATA);
		m_pSkin->GetContextInterface(pNode);

	
		GetCOREInterface()->SetCommandPanelTaskMode(TASK_MODE_MODIFY);
		GetCOREInterface()->SelectNode(pNode);
		return true;
	}
	void AddBone(INode* pBone, bool update = false){
		m_pImportData->AddBoneEx(pBone, update ? TRUE : FALSE);
	}
	BOOL AddWeights(UINT32 *indices, float* weights, INode** ppBones, UINT32 count, UINT32 vert){
		Tab<INode*> nodetab;
		Tab<float> weighttab;
		UINT32 i;
		for (i = 0; i < count; i++){
			if (weights[i] == 0)
				continue;
			nodetab.Append(1, &ppBones[indices[i]]);
			//nodetab.Append(1, &ppBones[indices[i]]);
			weighttab.Append(1, &weights[i]);
		}
		if (weighttab.Count() > 0)
			return m_pImportData->AddWeights(m_pNode, (int)vert, nodetab, weighttab);
		return FALSE;
	}
	void AddWeights(float* weights, INode** ppBones, UINT32 count, UINT32 vert){
		Tab<INode*> nodetab;
		Tab<float> weighttab;
		UINT32 i;
		for (i = 0; i < count; i++){
			if (weights[i] == 0 || !ppBones[i])
				continue;
			//nodetab.Append(1, &ppBones[indices[i]]);
			nodetab.Append(1, &ppBones[i]);
			weighttab.Append(1, &weights[i]);
		}
		if (weighttab.Count() > 0){
			m_pImportData->AddWeights(m_pNode, (int)vert, nodetab, weighttab);
		}
	}
};


INode* FindNode(INode* pNode, const MCHAR* pName){
	int i, c;
	INode* pFind;
	if (_M_STRCMPI(pNode->GetName(), pName) == 0)
		return pNode;
	c = pNode->NumChildren();
	for (i = 0; i < c; i++){
		pFind = FindNode(pNode->GetChildNode(i), pName);
		if (pFind)
			return pFind;
	}
	return NULL;
}

INode* FindNodeWithClass(INode* pNode, const MCHAR* pName, const Class_ID &cls){
	int i, c;
	INode* pFind;
	Object* pObj;
	if (_M_STRCMPI(pNode->GetName(), pName) == 0){
		if (pObj = pNode->GetObjectRef()){
			if (pObj->ClassID() == cls)
				return pNode;
		}
	}
	c = pNode->NumChildren();
	for (i = 0; i < c; i++){
		pFind = FindNodeWithClass(pNode->GetChildNode(i), pName, cls);
		if (pFind)
			return pFind;
	}
	return NULL;
}

class CTransform {
public:
	Matrix3	Transform;
	bool	Initialized;
	CTransform() : Initialized(false){
	}
};

ILayerManager* GetLayerMgr(){
#if MAX_SDK <= 2010
	#define SCENE_REF_LAYERREF		10
	return (ILayerManager*)(GetCOREInterface()->GetScenePointer()->GetReference(SCENE_REF_LAYERREF));
#else //MAX_SDK <= 2010
	return GetCOREInterface13()->GetLayerManager();
#endif //!MAX_SDK <= 2010
}
/*
void InitializeLayers(CModelState &state, ILayer** ppLayers){
	int i;
	TStringBufferA<128> buf;
	ILayerManager* pMgr;
	MSTR name;
	bool firstlayer;
	firstlayer = true;
	pMgr = GetLayerMgr();
	if (!state.UseLayers || !pMgr){
		memset(ppLayers, 0, sizeof(ILayer*)*(MAX_NUM_LODS+1));
		return;
	}
	for (i = 0; i < state.VvdHdr->numLODs; i++){
		buf.Print("Mdl LOD %d", i);
		if (!state.LODs[i]){
			ppLayers[i] = NULL;
			continue;
		}
		if (ppLayers[i] = pMgr->GetLayer(state.CharToMChar(buf)))
			continue;
		name = state.CharToMChar(buf);
		ppLayers[i] = pMgr->CreateLayer(name);
		if (!ppLayers[i])
			continue;
		ppLayers[i]->Hide(!firstlayer);
		firstlayer = false;
	}
	for (; i < MAX_NUM_LODS; i++){
		ppLayers[i] = NULL;
	}
	if (!(ppLayers[MAX_NUM_LODS] = pMgr->GetLayer(_M("Mdl Bones"))))
		ppLayers[MAX_NUM_LODS] = pMgr->CreateLayer(MSTR(_M("Mdl Bones")));
}*/


class CMemoryBuffer {
protected:
	void*	m_pBuffer;
	size_t	m_Size;
	CMemoryBuffer(CMemoryBuffer &){}
public:
	inline CMemoryBuffer() : m_pBuffer(NULL), m_Size(0){}
	inline ~CMemoryBuffer(){
		if (m_pBuffer)
			free(m_pBuffer);
	}
	inline bool Alloc(size_t size){
		if (size <= m_Size && m_pBuffer)
			return true;
		if (m_pBuffer)
			free(m_pBuffer);		
		return (m_pBuffer = malloc(m_Size = size)) != NULL;
	}
	inline void* GetBuffer(){ return m_pBuffer; }
	inline size_t GetSize() const { return m_Size; };
	inline const void* GetBuffer() const { return m_pBuffer; }
	inline PUINT8 operator()(){ return (PUINT8)m_pBuffer; }
};

class CBufferReader {
protected:
	CMemoryBuffer*	m_pBuffer;
	size_t			m_Offset;
	CBufferReader(){}
public:
	inline CBufferReader(CMemoryBuffer *pBuf) : m_pBuffer(pBuf), m_Offset(0){
	}
	inline ~CBufferReader(){
	}
	inline size_t Read(void* pDst, size_t size){
		if (!pDst)
			return 0;
		if (size > m_pBuffer->GetSize()-m_Offset)
			size = m_pBuffer->GetSize()-m_Offset;
		memcpy(pDst, (*m_pBuffer)()+m_Offset, size);
		m_Offset += size;
		return size;
	}
	inline bool ReadB(void* pDst, size_t size){
		if (!pDst)
			return 0;
		if (size > m_pBuffer->GetSize()-m_Offset)
			return false;
		memcpy(pDst, (*m_pBuffer)()+m_Offset, size);
		m_Offset += size;
		return true;
	}
	inline void* GetCurrent(){
		return (*m_pBuffer)()+m_Offset;
	}
	template<typename T> inline bool Get(T &dst){
		return (ReadB(&dst, sizeof(T)));
	}
	inline size_t SeekA(size_t offset){
		if (offset > m_pBuffer->GetSize())
			offset = m_pBuffer->GetSize();
		return m_Offset = offset;
	}
	inline size_t SeekR(ptrdiff_t delta){
		return SeekA(m_Offset + delta);
	}
	inline size_t Offset(){
		return m_Offset;
	}
	inline size_t Align(size_t align){
		return SeekA((m_Offset+(align-1))/align*align);
	}
	template<size_t align> size_t Align(){
		return SeekA((m_Offset+(align-1))/align*align);
	}
};


bool LoadFileToBuffer(const char* pPath, CMemoryBuffer &buf){
	size_t len;
	FILE* f;
	if (fopen_s(&f, pPath, "rb") != 0)
		return false;
	fseek(f, 0, SEEK_END);
	len = ftell(f);
	fseek(f, 0, SEEK_SET);
	if (!buf.Alloc(len))
		return fclose(f), false;
	if (fread(buf.GetBuffer(), 1, len, f) != len)
		return fclose(f), false;
	return fclose(f), true;
}


union UPointer {
	void* __ptr64 __uptr Ptr;
	INT32	Offset;
	inline void Init(void* pBase){
		if (Offset == -1)
			Ptr = NULL;
		else {
			Ptr = ((PUINT8)pBase)+Offset;
		}
	}
};
template<typename T> union TUPointer {
	T* __ptr64 __uptr Ptr;
	INT32	Offset;
	inline void Init(void* pBase){
		if (Offset == -1)
			Ptr = NULL;
		else {
			T = (T*)(((PUINT8)pBase)+Offset);
		}
	}
};

struct SRigBone{
	TUPointer<char>		Name;	
	float				tx;
	float				ty;
	float				tz;
	float				rx;
	float				ry;
	float				rz;
	INT32				Parent;
	INT32				Unknown;
};

struct SRigHelper {
	TUPointer<char>		Name;	
	float				tx;
	float				ty;
	float				tz;
	float				rx;
	float				ry;
	float				rz;
	float				MinX;
	float				MinY;
	float				MinZ;
	float				MaxX;
	float				MaxY;
	float				MaxZ;
	INT32				Parent;
	INT32				Unknown;
};

struct SRigHeader {
	UINT32		Field_0;
	UINT32		Field_4;
	UINT32		Field_8;
	UINT32		Field_C;
	UINT32		Field_10;
	UINT32		Field_14;
	UINT32		Field_18;
	UINT32		Field_1C;
	UINT32		Field_20;
	UINT32		BoneCount;
	UINT32		Field_28;
	UINT32		Field_2C;
	UINT32		HelperCount;
	UINT32		Field_34;
	UPointer	Field_38;
	TUPointer<SRigBone>		Bones;
	TUPointer<SRigHelper>	Helpers;
};

void InitUPointer(void* pData, UPointer& p){
	if (p.Offset == -1)
		p.Ptr = NULL;
	else {
		p.Ptr = ((UINT8*)pData)+p.Offset;
	}
}

template<typename T> void InitTUPointer(void* pData, TUPointer<T>& p){
	if (p.Offset == -1)
		p.Ptr = NULL;
	else {
		p.Ptr = (T*)(((UINT8*)pData)+p.Offset);
	}
}




struct SCMeshHeader {
	UINT16	Tag; //2
	UINT16	Version; //4
	UINT32	PreloadListLength; //8
	UINT32	PreloadListOffset; //C
	UINT32	PreloadListCount; //10
	UINT8	Pad[16];
	enum {
		TAG = '8T',
		VERSION = 4,
	};
};

struct SMeshDrawCall {
	UINT16	MaterialIndex;
	UINT16	VertexBufferIndex;
	UINT32	IndexStart;
	UINT32	IndexCount;
	UINT32	MinIndex;
	UINT32	MaxIndex;
};

struct SMeshObject{
	UINT32	NumDrawCalls;
	//Should be floats
	//Possibly bbox
	float	Field_4;
	float	Field_8;
	float	Field_C;
	float	Field_10;
	float	Field_14;
	float	Field_18;
	UINT32	Field_1C;
	TUPointer<SMeshDrawCall>	Drawcalls;
	TUPointer<TUPointer<UINT8>>	BoneData;
};

struct SVertexStream{
	UINT32	VertexCount;
	UINT8	Stream0Stride;
	UINT8	InputLayout;
	UINT8	InputLayoutType;
	UINT8	Size_3; //Could be align bytes
	UINT32	Field_8;
	UINT32	Field_C;
	UPointer	VertexData;
};

//Flag 1 might mean inline data
struct SMesh {
	UINT32		Flags;
	UINT32		NumObjects;
	TUPointer<SMeshObject>	Objects;
	UINT32		VertexStreamCount;
	UINT32		Field_14;
	TUPointer<SVertexStream>	VertexStreams;
	UINT32		IndexCount;
	UINT32		Field_24;
	UPointer	IndexData;
	UINT8		IndexSize;
	//3b space
	UINT32		Field_34;
	UINT16		BoneRemapIndexCount;
	//2b space
	UINT32		Field_3C;
	TUPointer<UINT8>	BoneRemapIndices;
	UINT16		BoneRemapDataCount;
	//2b space
	UINT32		Field_4C;
	TUPointer<UINT8>	BoneRemapData; //{ UINT8 Count; UINT8 Offset; }[BoneRemapDataCount]
	float		Field_58;
	float		Field_5C;
	float		Field_60;
	UINT32		Field_64;
	UINT32		Field_68;
	UINT32		Field_6C;
};

struct SMeshHeader {
	UINT32		FourCC;
	UINT16		Version;
	UINT16		Flags;
	INT16		Field_8;
	UINT16		RigBoneCount;
	UINT16		Field_C;
	UINT16		Field_E;
	UINT16		LodCount;
	UINT16		UpgradeCount;
	UINT32		Field_14;
	UPointer	Field_18;
	UPointer	Field_20;
	UINT32		Field_28;
	UINT32		Field_2C;
	UINT32		Field_30;
	UINT32		Field_34;
	UINT16		Field_38;
	UINT16		Field_3A;
	UINT32		Field_3C;
	UPointer	Field_40;
	UPointer	Field_48;
	//UPointer	m_pD3DMesh;
	TUPointer<SMesh>	m_pMesh; //There's typically the D3D mesh in this place
	UPointer	Field_58;
	UPointer	Field_60;
	UPointer	Field_68;
	UINT32		NumUpgradeLods;
	//UINT16	Field_72;
	UINT32		Field_74;
	TUPointer<UINT32>	UpgradeNames; //Numeric "names" found in items_3d.xtbl
	TUPointer<INT16>	LodMeshIndices; //Upgrades * INT16[LodCount]
	enum { FCC = 0x424BD00D };
};

struct SMtlMap{
	TUPointer<UINT32>	Field_0;
	UINT32				m_Count;
	UINT32				Field_C;
};

void InitLayout(UINT8 &il, UINT32 iltype){
	switch (il){
	case 0:
		il = 0x1B + iltype;
		break;
	case 1:
		il = 0x1F + iltype;
		break;
	case 100:
		il = 0x65 + iltype;
		break;
	case 2:
		il = 0x23 + iltype;
		break;
	case 3:
		il = 0x27 + iltype;
		break;
	case 101:
		il = 0x69 + iltype;
		break;
	case 4:
		il = 0x2B + iltype;
		break;
	case 14:
		il = 0x2F + iltype;
		break;
	case 15:
		il = 0x33 + iltype;
		break;
	case 16:
		il = 0x37 + iltype;
		break;
	case 17:
		il = 0x3B + iltype;
		break;
	case 18:
		il = 0x3F + iltype;
		break;
	case 19:
		il = 0x43 + iltype;
		break;
	case 20:
		il = 0x47 + iltype;
		break;
	case 21:
		il = 0x4B + iltype;
		break;
	case 22:
		il = 0x4F + iltype;
		break;
	case 23:
		il = 0x53 + iltype;
		break;
	case 24:
		il = 0x57 + iltype;
		break;
	case 25:
		il = 0x5B + iltype;
		break;
	case 26:
		il = 0x5F + iltype;
		break;
	}
}

/*
Flags:
1 = External data (g?mesh_pc)
2 = skinned
4 = ?
*/

SMesh* MeshParseSub(CBufferReader &r, UINT32 &ctag, UINT32 &glen){
	UINT32 smeshlen = sizeof(SMesh); 
	UINT16 meshversion;
	UINT32 tag, tag2; //Same tag at the end
	UINT32 len;
	size_t offset;
	UINT32 i, j;
	offset = r.Offset();
	r.Get(meshversion);
	r.Align<4>();
	if (meshversion != 9)
		return NULL;
	r.Get(tag);
	r.Get(len); //Len is from version to end tag
	r.Get(glen);
	r.Align<8>();
	SMesh* pMesh = (SMesh*)r.GetCurrent();
	r.SeekR(sizeof(SMesh));

	pMesh->VertexStreams.Ptr = (SVertexStream*)r.GetCurrent();
	r.SeekR(pMesh->VertexStreamCount*sizeof(SVertexStream));

	for (i = 0; i < pMesh->VertexStreamCount; i++){
		InitLayout(pMesh->VertexStreams.Ptr[i].InputLayout, pMesh->VertexStreams.Ptr[i].InputLayoutType);
	}
	if (pMesh->Flags & 2){
		if (pMesh->BoneRemapIndexCount > 0){
			pMesh->BoneRemapIndices.Ptr = (UINT8*)r.GetCurrent();
			r.SeekR(pMesh->BoneRemapIndexCount);
		} else
			pMesh->BoneRemapIndices.Ptr = NULL;
		
		if (pMesh->BoneRemapDataCount > 0){
			pMesh->BoneRemapData.Ptr = (UINT8*)r.GetCurrent();
			r.SeekR(pMesh->BoneRemapDataCount*2);
		} else
			pMesh->BoneRemapData.Ptr = NULL;

	} else {
		pMesh->BoneRemapData.Ptr = NULL;
		pMesh->BoneRemapIndices.Ptr = NULL;
	}

	r.Align<16>();
	

	//Meshes, drawacalls, bone map
	if (pMesh->NumObjects > 0){
		pMesh->Objects.Ptr = (SMeshObject*)r.GetCurrent();
		r.SeekR(pMesh->NumObjects*sizeof(SMeshObject));
		for (i = 0; i < pMesh->NumObjects; i++){
			pMesh->Objects.Ptr[i].Drawcalls.Ptr = (SMeshDrawCall*)r.GetCurrent();
			r.SeekR(pMesh->Objects.Ptr[i].NumDrawCalls*sizeof(SMeshDrawCall));
		}
		for (i = 0; i < pMesh->NumObjects; i++){
			if (pMesh->BoneRemapDataCount > 0){
				pMesh->Objects.Ptr[i].BoneData.Ptr = (TUPointer<UINT8>*)r.GetCurrent();
				r.SeekR(sizeof(UPointer)*pMesh->Objects.Ptr[i].NumDrawCalls);
				for (j = 0; j < pMesh->Objects.Ptr[i].NumDrawCalls; j++){
					pMesh->Objects.Ptr[i].BoneData.Ptr[j].Ptr = pMesh->BoneRemapData.Ptr+pMesh->Objects.Ptr[i].BoneData.Ptr[j].Offset*2;
				}
			} else {
				pMesh->Objects.Ptr[i].BoneData.Ptr = NULL;
			}
		}
	} else {
		pMesh->Objects.Ptr = NULL;
	}

	if (pMesh->NumObjects && !pMesh->Objects.Ptr)
		return NULL;

	if (!(pMesh->Flags & 1)){
		r.Align<16>();
		pMesh->IndexData.Ptr = r.GetCurrent();
		r.SeekR(pMesh->IndexSize*pMesh->IndexCount);
		for (i = 0; i < pMesh->VertexStreamCount; i++){
			r.Align<16>();
			pMesh->VertexStreams.Ptr[i].VertexData.Ptr = r.GetCurrent();
			r.SeekR((pMesh->VertexStreams.Ptr[i].Stream0Stride+pMesh->VertexStreams.Ptr[i].Size_3)*pMesh->VertexStreams.Ptr[i].VertexCount);
		}
	}
	if (pMesh->Flags & 4 && pMesh->IndexCount > 0){
		//FIXME:0000
	}
	r.Align<4>();
	r.Get(tag2);
	ctag = tag;

	return tag == tag2 ? pMesh : NULL;

	//return NULL;
}

bool InitializeMesh(SMesh* p, CBufferReader &cr, CBufferReader &gr, UINT32 ctag, UINT32 glen){
	UINT32 tag;
	size_t start;
	UINT32 i;
	start = gr.Offset();
	gr.Get(tag);
	SVertexStream* pStream;
	if (tag != ctag)
		return false;
	if (p->Flags & 4){
	} else {
		gr.Align<16>();
		p->IndexData.Ptr = gr.GetCurrent();
		gr.SeekR(p->IndexSize*p->IndexCount);
		for (i = 0; i < p->VertexStreamCount; i++){
			PTR_WARN
			pStream = (SVertexStream*)(p->VertexStreams.Ptr+i);
			
			gr.Align<16>();
			pStream->VertexData.Ptr = gr.GetCurrent();
			gr.SeekR((pStream->Stream0Stride+pStream->Size_3)*pStream->VertexCount);
		}
	}
	gr.Align<4>();
	gr.Get(tag);
	if (tag != ctag)
		return false;
	return (gr.Offset()-start == glen);
}

SMesh* MeshParse2(CBufferReader &cr, CBufferReader &gr){
	UINT32 tag, glen;
	SMesh* p = MeshParseSub(cr, tag, glen);
	if (!p)
		return NULL;
	if (p->Flags & 1 && !InitializeMesh(p, cr, gr, tag, glen))
		return NULL;
	return p;
}

struct SMtlHeader {
	UINT32		ShaderID; //Lower case CRC32 of shader name (exluding suffix)
	UINT32		Field_4;
	UINT32		Field_8;
	UINT16		Field_C;
	UINT8		Field_E;
	UINT8		Field_F;
	UPointer	Field_10;
	UPointer	Field_18;
	UPointer	Field_20;
	UPointer	Field_28;
};

SMtlHeader* ParseMtl(CBufferReader &r){
	UINT32 size;
	size_t pos;
	SMtlHeader* pHdr;
	pos = r.Offset();
	r.Get(size);

	r.Align<8>();
	pHdr = (SMtlHeader*)r.GetCurrent();
	r.SeekR(sizeof(SMtlHeader));

	r.Align<4>();
	pHdr->Field_10.Ptr = r.GetCurrent();
	r.SeekR(12*pHdr->Field_C);

	r.Align<4>();
	pHdr->Field_18.Ptr = r.GetCurrent();
	r.SeekR(4*pHdr->Field_E);

	r.Align<16>();
	pHdr->Field_20.Ptr = r.GetCurrent();
	r.SeekR(16*pHdr->Field_F);
	if (r.Offset()-size != pos)
		return NULL;
	//Initialize material here
	return pHdr;
}

bool MeshParse(CBufferReader &cr, CBufferReader &gr, SMeshHeader &mhdr){
	const char* plist;
	UINT32 listlen, i;
	SMesh* pMesh;
	cr.Align<8>();
	cr.Get(mhdr);
	cr.Get(listlen);
	cr.Align<16>();
	cr.SeekR(1);
	SMtlMap* pMtlMap;
	if (listlen)
		plist = (const char*)cr.GetCurrent();
	else
		plist = NULL;
	cr.SeekR(listlen);

	if (mhdr.Version < 0x2A)
		mhdr.UpgradeCount = mhdr.NumUpgradeLods;
	if (mhdr.UpgradeNames.Offset != -1 && mhdr.UpgradeCount < mhdr.NumUpgradeLods)
		return false;
	if (mhdr.Field_8 > 0){
		cr.Align<16>();
		mhdr.Field_18.Ptr = cr.GetCurrent();
		cr.SeekR(mhdr.Field_8 * 96);
	}
	if (mhdr.Field_38 > 0){
		cr.Align<16>();
		mhdr.Field_40.Ptr = cr.GetCurrent();
		cr.SeekR(mhdr.Field_38 * 24);
	}
	if (mhdr.Field_3A > 0){
		cr.Align<16>();
		mhdr.Field_48.Ptr = cr.GetCurrent();
		cr.SeekR(mhdr.Field_3A * 40);
	}
	if (mhdr.RigBoneCount > 0){
		cr.Align<16>();
		mhdr.Field_20.Ptr = cr.GetCurrent();
		cr.SeekR(mhdr.RigBoneCount * 4);
	}

	cr.Align<8>();
	pMesh = MeshParse2(cr, gr);
	if (!pMesh)
		return false;
	mhdr.m_pMesh.Ptr = pMesh;



	if (mhdr.Field_E > 0){
		cr.Align<8>();
		mhdr.Field_58.Ptr = cr.GetCurrent();
		cr.SeekR(8*mhdr.Field_E);
		mhdr.Field_60.Ptr = cr.GetCurrent();
		cr.SeekR(4*mhdr.Field_E);
	}
	if (mhdr.Field_C > 0){
		cr.Align<8>();
		mhdr.Field_68.Ptr = cr.GetCurrent();
		cr.SeekR(mhdr.Field_C*8);
		for (i = 0; i < mhdr.Field_C; i++){
			if (!ParseMtl(cr))
				return false;
		}
	}
	if (mhdr.Field_E > 0){
		for (i = 0; i < mhdr.Field_E; i++){
			cr.Align<8>();
			pMtlMap = (SMtlMap*)cr.GetCurrent();
			cr.SeekR(sizeof(SMtlMap));
			pMtlMap->Field_0.Ptr = (UINT32*)cr.GetCurrent();
			cr.SeekR(sizeof(UINT32)*pMtlMap->m_Count);
		}
	}
	//
	if (mhdr.UpgradeNames.Offset == -1){
		mhdr.UpgradeNames.Ptr = NULL;
	} else {
		cr.Align<4>();
		mhdr.UpgradeNames.Ptr = (UINT32*)cr.GetCurrent();
		cr.SeekR(sizeof(UINT32)* mhdr.UpgradeCount);
	}
	if (mhdr.LodMeshIndices.Offset == -1){
		mhdr.LodMeshIndices.Ptr = NULL;
	} else {
		cr.Align<2>();
		mhdr.LodMeshIndices.Ptr = (INT16*)cr.GetCurrent();
		cr.SeekR(mhdr.NumUpgradeLods * mhdr.LodCount * sizeof(INT16));
	}
	return true;
}


bool AdjustCar(CBufferReader &r, bool &iscar){
	size_t start;
	UINT32 id;
	UINT32 offset;
	start = r.Offset();
	r.Get(id);
	if (id != 0x38){
		r.SeekA(start);
		iscar = false;
		return true;
	}
	r.Get(offset);
	if (offset == 0xFFFFFFFF)
		return false;
	r.SeekA(offset);
	iscar = true;
	return true;
}


Matrix3 D3DMatrixToMatrix3(D3DMATRIX m){
	Matrix3 tm;
	int i;
	tm.IdentityMatrix();
	for (i = 0; i < 4; i++)
		tm.SetRow(i, Point3(m.m[i][0], m.m[i][1], m.m[i][2]));
	return tm;
}


Matrix3 SwapMatrixComponents(Matrix3 m){
	Point3 row1, row2;
	Point4 col1, col2;
	row1 = m.GetRow(1);
	row2 = m.GetRow(2);
	m.SetRow(1, row2);
	m.SetRow(2, row1);
	col1 = m.GetColumn(1);
	col2 = m.GetColumn(2);
	m.SetColumn(1, col2);
	m.SetColumn(2, col1);
	return m;
}

class CModelState{
public:
	CMemoryBuffer	m_CCMesh;
	CMemoryBuffer	m_GCMesh;
	SMeshHeader		m_Data;
	bool			m_IsCar;
	bool GetExtensions(const char* pext, const char*&cext, const char*&gext){
		if (_stricmp(pext, "gcmesh_pc") == 0 || _stricmp(pext, "ccmesh_pc") == 0){
			cext = "ccmesh_pc";
			gext = "gcmesh_pc";
			return true;
		}
		if (_stricmp(pext, "ccar_pc") == 0 || _stricmp(pext, "gcar_pc") == 0){
			cext = "ccar_pc";
			gext = "gcar_pc";
			return true;
		}
		return false;
	}
	bool Init(const char* pPath){
		const char* gext, *cext;
		CStringA path, name, ext;
		size_t start;
		TStringBufferA<MAX_PATH> buf;
		SplitPath(pPath, &path, &name, &ext);
		if (!GetExtensions(ext, cext, gext))
			return false;
		buf.Print("%s\\%s.%s", path(), name(), cext);
		if (!LoadFileToBuffer(buf, m_CCMesh))
			return false;
		buf.Print("%s\\%s.%s", path(), name(), gext);
		if (!LoadFileToBuffer(buf, m_GCMesh))
			return false;
		CBufferReader cr(&m_CCMesh);
		CBufferReader gr(&m_GCMesh);
		SCMeshHeader chdr;
		if (!AdjustCar(cr, m_IsCar))
			return false;
		//Header load
		start = cr.Offset();
		cr.Get(chdr);
		if (chdr.Tag != SCMeshHeader::TAG || chdr.Version != SCMeshHeader::VERSION)
			cr.SeekA(start);
		else {
			cr.SeekR(chdr.PreloadListLength+1);
			if (chdr.Version >= 2)
				cr.Align<16>();
		}
		//Mesh load
		return MeshParse(cr, gr, m_Data);
	}
	PUINT8 CData(){
		return (PUINT8)m_CCMesh.GetBuffer();
	}
	const char* PartName(INT32 offset){
		if (offset == -1)
			return NULL;
		return (const char*)(CData()+offset);
	}
	INT32 GetEntryCount(){
		return m_IsCar ? *(UINT32*)(CData()+0x390) : 0;
	}
	PUINT8 GetEntry(INT32 i){
		if (i < 0 || i >= GetEntryCount())
			return NULL;
		return CData()+0x3A0+(i*0xE0);
	}
	const char* GetPartName(INT32 part){
		PUINT8 pEntry;
		INT32 i;
		if (!m_IsCar)
			return NULL;
		if (*(INT32*)(CData()+0x34) == part){
			return PartName(*(INT32*)(CData()+0x3C));
		}
		for (i = 0; i < GetEntryCount(); i++){
			pEntry = GetEntry(i);
			if (!pEntry)
				return NULL;
			if (*(INT32*)(pEntry+0x5C) == part){
				return PartName(*(INT32*)(pEntry+0x40));
			}
		}
		return NULL;
	}
	Matrix3 GetPartTransform(INT32 part){
		PUINT8 pEntry;
		Matrix3 tm, stm;
		tm.SetRotateX(DegToRad(-90.0f));
		stm.SetScale(Point3(1.0, 1.0, -1.0));
		INT32 i;
		Matrix3 m;
		if (!m_IsCar)
			return NULL;
		if (*(INT32*)(CData()+0x34) == part){
			m.IdentityMatrix();
			return m;
			//return PartName(*(INT32*)(CData()+0x3C));
		}
		for (i = 0; i < GetEntryCount(); i++){
			pEntry = GetEntry(i);
			if (!pEntry)
				return NULL;
			if (*(INT32*)(pEntry+0x5C) == part){
				m = D3DMatrixToMatrix3(*(D3DMATRIX*)pEntry);
				return m;
				//return PartName(*(INT32*)(pEntry+0x40));
			}
		}
		m.IdentityMatrix();
		return m;
	}
};


UINT32 IndexReaderU16(const void* pData, UINT32 &offset){
	return ((UINT16*)pData)[offset++];
}

UINT32 IndexReaderU32(const void* pData, UINT32 &offset){
	return ((UINT32*)pData)[offset++];
}

typedef UINT32 (*pfnIndexReader)(const void* pData, UINT32 &offset);

class CTriStripReader {
protected:
	pfnIndexReader	m_pfnIndexReader;
	UINT32			m_Offset;
	UINT32			m_Index;
	const void*		m_pData;
	UINT32			m_Indices[3];
	UINT32			m_StoredIndices[3];
	CTriStripReader(){}
	CTriStripReader(CTriStripReader&){}
public:
	CTriStripReader(const void* pData, UINT32 size) : m_pData(pData), m_Offset(0), m_Index(1){
		m_pfnIndexReader = size == 2 ? IndexReaderU16 : IndexReaderU32;
		m_StoredIndices[0] = m_Indices[0] = m_pfnIndexReader(m_pData, m_Offset);
		m_StoredIndices[1] = m_Indices[1] = m_pfnIndexReader(m_pData, m_Offset);
		m_StoredIndices[2] = m_Indices[2] = m_pfnIndexReader(m_pData, m_Offset);
	}
	inline CTriStripReader &operator++(){
		m_Indices[0] = m_StoredIndices[0] = m_StoredIndices[1];
		m_StoredIndices[1] = m_StoredIndices[2];
		m_StoredIndices[2] = m_pfnIndexReader(m_pData, m_Offset);
		
		if (m_Index & 1){
			m_Indices[1] = m_StoredIndices[2];
			m_Indices[2] = m_StoredIndices[1];
		} else {			
			m_Indices[1] = m_StoredIndices[1];
			m_Indices[2] = m_StoredIndices[2];
		}
		m_Index++;
		return *this;
	}
	void Inc(){
		m_Indices[0] = m_StoredIndices[0] = m_StoredIndices[1];
		m_StoredIndices[1] = m_StoredIndices[2];
		m_StoredIndices[2] = m_pfnIndexReader(m_pData, m_Offset);
		
		if (m_Index & 1){
			m_Indices[1] = m_StoredIndices[2];
			m_Indices[2] = m_StoredIndices[1];
		} else {			
			m_Indices[1] = m_StoredIndices[1];
			m_Indices[2] = m_StoredIndices[2];
		}
		m_Index++;
	}
	bool Degenerate(){
		return (m_Indices[0] == m_Indices[1]) | (m_Indices[0] == m_Indices[2]) | (m_Indices[1] == m_Indices[2]);
	}
	void Reset(){
		m_Offset = 0;
		m_Index = 1;
		m_StoredIndices[0] = m_Indices[0] = m_pfnIndexReader(m_pData, m_Offset);
		m_StoredIndices[1] = m_Indices[1] = m_pfnIndexReader(m_pData, m_Offset);
		m_StoredIndices[2] = m_Indices[2] = m_pfnIndexReader(m_pData, m_Offset);
	}
	inline UINT32 operator[](int index){
		return m_Indices[index];
	}
};

const void* GetIndexPtr(UINT32 offset, SMesh* pMesh){
PTR_WARN
	return ((UINT8*)pMesh->IndexData.Ptr)+offset * pMesh->IndexSize;
}

struct SDrawCallData {
	UINT32	MaterialIndex;
	UINT32	FirstFace;
	UINT32	NumFaces;
};

struct SFace {
	UINT32	v[3];
};

class CVertexRemap{
public:
	INT32*	m_pVertexRemap;
	UINT32*	m_pVertexStart;
	UINT32	m_NumVerts;
	CVertexRemap() : m_pVertexRemap(NULL), m_pVertexStart(NULL){

	}
	~CVertexRemap(){
		if (m_pVertexRemap)
			delete [] m_pVertexRemap;
		if (m_pVertexStart)
			delete [] m_pVertexStart;
	}
	bool Init(SMesh* pMesh){
		UINT32 i, c;
		c = 0;
		m_pVertexStart = new UINT32[pMesh->VertexStreamCount];
		for (i = 0; i < pMesh->VertexStreamCount; i++){
			m_pVertexStart[i] = (UINT32)c;
			c += pMesh->VertexStreams.Ptr[i].VertexCount;
		}
		m_pVertexRemap = new INT32[m_NumVerts = c];
		return true;
	}
	void Reset(){
		if (m_pVertexRemap)
			memset(m_pVertexRemap, 0xFF, sizeof(INT32)*m_NumVerts);
	}
	INT32* operator[](UINT32 index){
		return m_pVertexRemap+m_pVertexStart[index];
	}
};

class CGatheredData {
public:
	//INT32*			m_pVertexRemap;
	SDrawCallData*	m_pDrawCallData;
	SFace*			m_pFaceList;
	UINT32			m_FaceCount;
	UINT32			m_VertexCount;
	INode**			m_ppBoneMap;
	bool*			m_pUsedBones;
	UINT8*			m_pVertexBits;
	CGatheredData() : /*m_pVertexRemap(NULL),*/ m_pDrawCallData(NULL), m_pFaceList(NULL), m_ppBoneMap(NULL), m_pUsedBones(NULL), m_pVertexBits(NULL){}
	~CGatheredData(){
		/*if (m_pVertexRemap)
			delete [] m_pVertexRemap;*/
		if (m_pDrawCallData)
			delete [] m_pDrawCallData;
		if (m_pFaceList)
			delete [] m_pFaceList;
		if (m_ppBoneMap)
			delete [] m_ppBoneMap;
		if (m_pUsedBones)
			delete [] m_pUsedBones;
		if (m_pVertexBits)
			delete [] m_pVertexBits;
	}


	bool GatherData(SMesh* pMesh, SMeshObject* pObj, UINT32 bonecount, CVertexRemap&remap){
		UINT32 i, j, fi, vs;
		SMeshDrawCall* pDC;
		//First gather stage: vertex remap, face counts, vertex count
		//m_pVertexRemap = new INT32[pMesh->VertexStreams.Ptr[0].VertexCount];
		//memset(m_pVertexRemap, 0xFF, sizeof(INT32)*pMesh->VertexStreams.Ptr[0].VertexCount);
		remap.Reset();
		m_pDrawCallData = new SDrawCallData[pObj->NumDrawCalls];
		m_VertexCount = 0;
		m_FaceCount = 0;
		for (i = 0; i < pObj->NumDrawCalls; i++){
			PTR_WARN
			pDC = pObj->Drawcalls.Ptr+i;
			CTriStripReader r(GetIndexPtr(pDC->IndexStart, pMesh), pMesh->IndexSize);
			m_pDrawCallData[i].FirstFace = m_FaceCount;
			vs = pDC->VertexBufferIndex;
			for (j = 0; j < pDC->IndexCount-2; j++, r.Inc()){
				if (r.Degenerate())
					continue;
				m_FaceCount++;
				
				if (remap[vs][r[0]] == -1){
					remap[vs][r[0]] = m_VertexCount++;
				}
				if (remap[vs][r[1]] == -1){
					remap[vs][r[1]] = m_VertexCount++;
				}
				if (remap[vs][r[2]] == -1){
					remap[vs][r[2]] = m_VertexCount++;
				}
			}
			m_pDrawCallData[i].NumFaces = m_FaceCount-m_pDrawCallData[i].FirstFace;
			m_pDrawCallData[i].MaterialIndex = pDC->MaterialIndex;
		}
		//Second stage: facelist
		m_pFaceList = new SFace[m_FaceCount];
		fi = 0;
		for (i = 0; i < pObj->NumDrawCalls; i++){
			PTR_WARN
			pDC = pObj->Drawcalls.Ptr+i;
			CTriStripReader r(GetIndexPtr(pDC->IndexStart, pMesh), pMesh->IndexSize);
			for (j = 0; j < pDC->IndexCount-2; j++, r.Inc()){
				if (r.Degenerate())
					continue;
				m_pFaceList[fi].v[0] = r[0];
				m_pFaceList[fi].v[1] = r[1];
				m_pFaceList[fi].v[2] = r[2];
				fi++;
			}
		}
		//Third stage: bone map
		m_pUsedBones = new bool[bonecount];
		memset(m_pUsedBones, 0, sizeof(bool)*bonecount);
		m_pVertexBits = new UINT8[remap.m_NumVerts];
		memset(m_pVertexBits, 0, sizeof(UINT8)*remap.m_NumVerts);
		return true;
	}
};

#define UV_SCALE	(1.0/1024.0)

bool AnyReferencedStreamSupports(SMesh* pMesh, SMeshObject* pObj, UINT32 decl){
	UINT32 i;
	SInputLayout* pLayout;
	SInputLayoutElement* pElem;
	for (i = 0; i < pObj->NumDrawCalls; i++){
		pLayout = GetIL(pMesh->VertexStreams.Ptr[pObj->Drawcalls.Ptr[i].VertexBufferIndex].InputLayout);
		if (pLayout){
			pElem = GetElement(pLayout, decl);
			if (pElem && pElem->Stream == 0)
				return true;
		}
	}
	return false;
}

bool ImportObject(SMesh* pMesh, SMeshObject* pObject, MCHAR* pName, INode** ppBones, UINT32 bonecount, CVertexRemap &remap, Matrix3 &ntm){
	CGatheredData data;
	SInputLayout* pLayout;
	SInputLayoutElement* pElem, *pElem2;
	INode* pNode;
	UINT8 bc, bo;
	TriObject* pObj;
	UINT32 i, j, k, l, d;
	UINT32 fidx;
	SInputValue v, v2;
	Matrix3 tm;
	UINT32 vs;
	tm.SetRotateX(DegToRad(90.0f));
	if (!data.GatherData(pMesh, pObject, bonecount, remap))
		return false;

	pObj = CreateTriObjectAndAddToRoot(&pNode);
	if (!pObj)
		return false;
	Mesh* m = &pObj->mesh;
	m->setNumFaces(data.m_FaceCount);
	m->setNumVerts(data.m_VertexCount);
	{
		for (d = 0; d < pObject->NumDrawCalls; d++){
			vs = pObject->Drawcalls.Ptr[d].VertexBufferIndex;
			pLayout = GetIL(pMesh->VertexStreams.Ptr[vs].InputLayout);
			if (!pLayout){
				return pObj->DeleteThis(), false;
			}
			pElem = GetElement(pLayout, SRVDecl_Position0);
			if (!pElem || pElem->Stream != 0)
				return pObj->DeleteThis(), false;
			PTR_WARN
			CStridedPointer p(pMesh->VertexStreams.Ptr[vs].VertexData.Ptr, pMesh->VertexStreams.Ptr[vs].Stream0Stride);

			for (j = 0; j < data.m_pDrawCallData[d].NumFaces; j++){
				for (l = 0; l < 3; l++){
					fidx = data.m_pFaceList[data.m_pDrawCallData[d].FirstFace+j].v[l];
					ToValue(v, p[fidx], pElem);
#ifdef SR3IMP_FIXAXIS
					m->verts[remap[vs][fidx]] = tm.PointTransform(Point3(-v.x, v.y, v.z));
#else //SR3IMP_FIXAXIS
					m->verts[remap[vs][fidx]] = Point3(v.x, v.y, v.z);
#endif //!SR3IMP_FIXAXIS
				}
				m->faces[data.m_pDrawCallData[d].FirstFace+j].v[0] = remap[vs][data.m_pFaceList[data.m_pDrawCallData[d].FirstFace+j].v[0]];
				m->faces[data.m_pDrawCallData[d].FirstFace+j].v[2] = remap[vs][data.m_pFaceList[data.m_pDrawCallData[d].FirstFace+j].v[1]];
				m->faces[data.m_pDrawCallData[d].FirstFace+j].v[1] = remap[vs][data.m_pFaceList[data.m_pDrawCallData[d].FirstFace+j].v[2]];
				m->faces[data.m_pDrawCallData[d].FirstFace+j].setMatID(data.m_pDrawCallData[d].MaterialIndex);
			}
		}
	}
	m->buildBoundingBox();

	for (k = SRVDecl_Texcoord0; k <= SRVDecl_Texcoord7; k++){
		if (!AnyReferencedStreamSupports(pMesh, pObject, k))
			continue;
		for (d = 0; d < pObject->NumDrawCalls; d++){
			vs = pObject->Drawcalls.Ptr[d].VertexBufferIndex;
			//Only need to bail once for this (position stage)
			pLayout = GetIL(pMesh->VertexStreams.Ptr[vs].InputLayout);
			pElem = GetElement(pLayout, k);
			
			m->setMapSupport(k-SRVDecl_Texcoord0+1);
			m->setNumMapFaces(k-SRVDecl_Texcoord0+1, data.m_FaceCount);
			m->setNumMapVerts(k-SRVDecl_Texcoord0+1, data.m_VertexCount);
			if (!pElem || pElem->Stream != 0){

				for (j = 0; j < data.m_pDrawCallData[d].NumFaces; j++){
					for (l = 0; l < 3; l++){
						fidx = data.m_pFaceList[data.m_pDrawCallData[d].FirstFace+j].v[l];
						m->mapVerts(k-SRVDecl_Texcoord0+1)[remap[vs][fidx]] = UVVert(0.0, 0.0, 0.0);
					}
					m->mapFaces(k-SRVDecl_Texcoord0+1)[data.m_pDrawCallData[d].FirstFace+j].t[0] = remap[vs][data.m_pFaceList[data.m_pDrawCallData[d].FirstFace+j].v[0]];
					m->mapFaces(k-SRVDecl_Texcoord0+1)[data.m_pDrawCallData[d].FirstFace+j].t[2] = remap[vs][data.m_pFaceList[data.m_pDrawCallData[d].FirstFace+j].v[1]];
					m->mapFaces(k-SRVDecl_Texcoord0+1)[data.m_pDrawCallData[d].FirstFace+j].t[1] = remap[vs][data.m_pFaceList[data.m_pDrawCallData[d].FirstFace+j].v[2]];
				}
			} else {
				PTR_WARN
				CStridedPointer p(pMesh->VertexStreams.Ptr[vs].VertexData.Ptr, pMesh->VertexStreams.Ptr[vs].Stream0Stride);

				for (j = 0; j < data.m_pDrawCallData[d].NumFaces; j++){
					for (l = 0; l < 3; l++){
						fidx = data.m_pFaceList[data.m_pDrawCallData[d].FirstFace+j].v[l];
						ToValue(v, p[fidx], pElem);
						m->mapVerts(k-SRVDecl_Texcoord0+1)[remap[vs][fidx]] = UVVert(v.x*UV_SCALE, 1.0-(v.y*UV_SCALE), 0.0);
					}
					m->mapFaces(k-SRVDecl_Texcoord0+1)[data.m_pDrawCallData[d].FirstFace+j].t[0] = remap[vs][data.m_pFaceList[data.m_pDrawCallData[d].FirstFace+j].v[0]];
					m->mapFaces(k-SRVDecl_Texcoord0+1)[data.m_pDrawCallData[d].FirstFace+j].t[2] = remap[vs][data.m_pFaceList[data.m_pDrawCallData[d].FirstFace+j].v[1]];
					m->mapFaces(k-SRVDecl_Texcoord0+1)[data.m_pDrawCallData[d].FirstFace+j].t[1] = remap[vs][data.m_pFaceList[data.m_pDrawCallData[d].FirstFace+j].v[2]];
				}
			}
		}
	}
	
	m->SpecifyNormals();
	MeshNormalSpec* pSpec = m->GetSpecifiedNormals();
	
	pSpec->Clear();
	pSpec->SetNumNormals(data.m_VertexCount);
	pSpec->SetNumFaces(data.m_FaceCount);
	for (d = 0; d < pObject->NumDrawCalls; d++){
		vs = pObject->Drawcalls.Ptr[d].VertexBufferIndex;
		pLayout = GetIL(pMesh->VertexStreams.Ptr[vs].InputLayout);
		pElem = GetElement(pLayout, SRVDecl_Normal0);
		if (!pElem || pElem->Stream != 0)
			continue;
		PTR_WARN
		CStridedPointer p(pMesh->VertexStreams.Ptr[vs].VertexData.Ptr, pMesh->VertexStreams.Ptr[vs].Stream0Stride);

		for (j = 0; j < data.m_pDrawCallData[d].NumFaces; j++){
			for (l = 0; l < 3; l++){
				fidx = data.m_pFaceList[data.m_pDrawCallData[d].FirstFace+j].v[l];
				ToValue(v, p[fidx], pElem);
				//m->verts[remap[vs][fidx]] = tm.PointTransform(Point3(-v.x, v.y, v.z));
				
#ifdef SR3IMP_FIXAXIS
				pSpec->Normal(remap[vs][fidx]) = tm.PointTransform(Point3(-(v.x*2.0f-1.0f), v.y*2.0f-1.0f, v.z*2.0f-1.0f));
#else //SR3IMP_FIXAXIS
				pSpec->Normal(remap[vs][fidx]) = Point3((v.x*2.0f-1.0f), v.y*2.0f-1.0f, v.z*2.0f-1.0f);
#endif //!SR3IMP_FIXAXIS
				pSpec->SetNormalExplicit(remap[vs][fidx], true);
			}
			pSpec->SetNormalIndex(data.m_pDrawCallData[d].FirstFace+j, 0, remap[vs][data.m_pFaceList[data.m_pDrawCallData[d].FirstFace+j].v[0]]);
			pSpec->SetNormalIndex(data.m_pDrawCallData[d].FirstFace+j, 2, remap[vs][data.m_pFaceList[data.m_pDrawCallData[d].FirstFace+j].v[1]]);
			pSpec->SetNormalIndex(data.m_pDrawCallData[d].FirstFace+j, 1, remap[vs][data.m_pFaceList[data.m_pDrawCallData[d].FirstFace+j].v[2]]);
			pSpec->GetFaceArray()[data.m_pDrawCallData[d].FirstFace+j].SpecifyAll();
		}
	}

	if (AnyReferencedStreamSupports(pMesh, pObject, SRVDecl_BlendWeight) && ppBones && bonecount > 0){
		
		CSkin skin;
		float w[4];
		INode* b[4];
#if 0
		const MCHAR* n[4];
#endif //0
		double s;
		if (skin.Create(pNode)){
			for (i = 0; i < pObject->NumDrawCalls; i++){
				bc = pObject->BoneData.Ptr[i].Ptr[0];
				bo = pObject->BoneData.Ptr[i].Ptr[1];
				for (k = 0; k < bc; k++){
					if (!data.m_pUsedBones[pMesh->BoneRemapIndices.Ptr[bo+k]]){
						data.m_pUsedBones[pMesh->BoneRemapIndices.Ptr[bo+k]] = true;
						skin.AddBone(ppBones[pMesh->BoneRemapIndices.Ptr[bo+k]], true);
						//LogFormated(L"Adding bone %S\n", ppBones[pMesh->BoneRemapIndices.Ptr[bo+k]]->GetName());
						//pPrev = ppBones[];
					}
				}
			}


			for (d = 0; d < pObject->NumDrawCalls; d++){
				vs = pObject->Drawcalls.Ptr[d].VertexBufferIndex;
				pLayout = GetIL(pMesh->VertexStreams.Ptr[vs].InputLayout);
				pElem2 = GetElement(pLayout, SRVDecl_BlendWeight);
				pElem = GetElement(pLayout, SRVDecl_BlendIndices);
				if (!pElem || pElem->Stream != 0 || !pElem2 || pElem2->Stream != 0)
					continue;
				PTR_WARN
				CStridedPointer p(pMesh->VertexStreams.Ptr[vs].VertexData.Ptr, pMesh->VertexStreams.Ptr[vs].Stream0Stride);

				for (j = 0; j < data.m_pDrawCallData[d].NumFaces; j++){
					for (l = 0; l < 3; l++){
						fidx = data.m_pFaceList[data.m_pDrawCallData[d].FirstFace+j].v[l];
						if (!(data.m_pVertexBits[remap.m_pVertexStart[vs]+fidx] & 1)){
							data.m_pVertexBits[remap.m_pVertexStart[vs]+fidx] |= 1;
							ToValue(v, p[fidx], pElem);
							ToValue(v2, p[fidx], pElem2);
							s = w[0] = v2.x;
							s += w[1] = v2.y;
							s += w[2] = v2.z;
							s += w[3] = v2.w;
							s = 1.0/s;
							w[0] = (float)(w[0]*s);
							w[1] = (float)(w[1]*s);
							w[2] = (float)(w[2]*s);
							w[3] = (float)(w[3]*s);

							b[0] = bo+v.ix < pMesh->BoneRemapIndexCount ? ppBones[pMesh->BoneRemapIndices.Ptr[bo+v.ix]] : NULL;
							b[1] = bo+v.iy < pMesh->BoneRemapIndexCount ? ppBones[pMesh->BoneRemapIndices.Ptr[bo+v.iy]] : NULL;
							b[2] = bo+v.iz < pMesh->BoneRemapIndexCount ? ppBones[pMesh->BoneRemapIndices.Ptr[bo+v.iz]] : NULL;
							b[3] = bo+v.iw < pMesh->BoneRemapIndexCount ? ppBones[pMesh->BoneRemapIndices.Ptr[bo+v.iw]] : NULL;
#if 0
							n[0] = b[0] ? b[0]->GetName() : NULL;
							n[1] = b[1] ? b[1]->GetName() : NULL;
							n[2] = b[2] ? b[2]->GetName() : NULL;
							n[3] = b[3] ? b[3]->GetName() : NULL;
#endif //0
							skin.AddWeights(w, b, 4, remap[vs][fidx]);
						}
					}
				}
			}
		}
	}

	if (pName)
		pNode->SetName(pName);
	//pNode->SetUserPropInt(_M("InputLayout"), pMesh->VertexStreams.Ptr[0].InputLayout);
	pNode->SetNodeTM(0, ntm);

	return true;
}



SRigHeader* LoadRig(CBufferReader &r){
	UINT32 i;
	char* pNames;
	SRigHeader* pHdr;
	pHdr = (SRigHeader*)r.GetCurrent();
	r.SeekR(sizeof(SRigHeader));
	InitUPointer(r.GetCurrent(), pHdr->Field_38);
	r.SeekR(4*pHdr->BoneCount);
	r.Align<8>();
	InitTUPointer(r.GetCurrent(), pHdr->Bones);
	r.SeekR(pHdr->BoneCount*0x28);
	InitTUPointer(r.GetCurrent(), pHdr->Helpers);
	r.SeekR(pHdr->HelperCount*0x40);

	pNames = (char*)r.GetCurrent();

	for (i = 0; i < pHdr->BoneCount; i++){
		if (pHdr->Bones.Ptr[i].Name.Offset == -1)
			pHdr->Bones.Ptr[i].Name.Ptr = NULL;
		else
			pHdr->Bones.Ptr[i].Name.Ptr = pNames + pHdr->Bones.Ptr[i].Name.Offset;
	}
	for (i = 0; i < pHdr->HelperCount; i++){
		if (pHdr->Helpers.Ptr[i].Name.Offset == -1)
			pHdr->Helpers.Ptr[i].Name.Ptr = NULL;
		else
			pHdr->Helpers.Ptr[i].Name.Ptr = pNames + pHdr->Helpers.Ptr[i].Name.Offset;
	}
	return pHdr;
}


void CreateBone(const char* pName, INode** ppNode){
	INode *pNode;
	//Matrix3 mat, mr, mt, xm, ym, zm;

	SimpleObject2* pBoneObj = (SimpleObject2*)GetCOREInterface()->CreateInstance( GEOMOBJECT_CLASS_ID, BONE_OBJ_CLASSID );
	pBoneObj->pblock2->SetValue(boneobj_length, 0, 0.02f);
	pBoneObj->pblock2->SetValue(boneobj_width, 0, 0.01f);
	pBoneObj->pblock2->SetValue(boneobj_height, 0, 0.01f);
	
	
	pNode = GetCOREInterface()->CreateObjectNode(pBoneObj);
	pNode->SetBoneFreezeLen(TRUE);
	pNode->SetBoneNodeOnOff(TRUE, 0);
	CStringConvert s;
	pNode->SetName((MCHAR*)s.UTF8ToM(pName));
	if (ppNode)
		*ppNode = pNode;
}

void DoImportRig(SRigHeader* pHdr, INode***pppNode, UINT32* pBoneCount){
	UINT32 i;
	INode** ppNodes = new INode*[pHdr->BoneCount];
	Matrix3 am;
	SRigBone* pBone;
	am.SetRotateX(DegToRad(-90));
	Matrix3 mr, mt;
	Point3 t, r;
	for (i = 0; i < pHdr->BoneCount; i++){
		PTR_WARN
		pBone = pHdr->Bones.Ptr+i;
		t.x = pBone->tx;
		t.y = pBone->ty;
		t.z = pBone->tz;
		r.x = pBone->rx;
		r.y = pBone->ry;
		r.z = pBone->rz;
		
		mr = (RotateXMatrix(r.x)*RotateYMatrix(r.y))*RotateZMatrix(r.z);
		mt.SetTranslate(t);
		PTR_WARN
		CreateBone(pBone->Name.Ptr, &ppNodes[i]);
		ppNodes[i]->SetBoneNodeOnOff(TRUE, 0);
		ppNodes[i]->SetNodeTM(0, mt*am);
	}
	for (i = 0; i < pHdr->BoneCount; i++){
		PTR_WARN
		pBone = pHdr->Bones.Ptr+i;
		if (pBone->Parent != -1)
			ppNodes[pBone->Parent]->AttachChild(ppNodes[i], 1);
		ppNodes[i]->SetBoneAutoAlign(TRUE);
		
		ppNodes[i]->SetRenderable(FALSE);
		ppNodes[i]->ResetBoneStretch(0);
	}
	if (pppNode){
		if (!pBoneCount){
			pppNode = NULL;
			delete [] ppNodes;
		} else{
			*pppNode = ppNodes;
			*pBoneCount = pHdr->BoneCount;
		}
	} else		
		delete [] ppNodes;
}

bool DoImportRigFile(const char* pName, INode***pppNode, UINT32* pBoneCount){
	CMemoryBuffer rigbuf;
	if (!LoadFileToBuffer(pName, rigbuf))
		return IMPEXP_FAIL;
	CBufferReader r(&rigbuf);

	SRigHeader* pRig = LoadRig(r);
	DoImportRig(pRig, pppNode, pBoneCount);
	return true;
}

struct SImportOption{
	UINT16	LOD;
	UINT32	Prop;
	bool	Import;
	INT32	PartIndex;
	SImportOption(): LOD(0xFFFF), Prop(0xFFFFFFFF), Import(true), PartIndex(-1){}
};

INT_PTR CALLBACK OptionDlgProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);

class CImportOptions {
public:
	SImportOption*				m_pObjectOptions;
	UINT32						m_NumOptions;
	bool						m_MatchExistingBones;
	TStringBufferA<MAX_PATH>	m_Rig;
	CStringA					m_ModelName;
	CStringA					m_Path;
	bool						m_DoRig;
	UINT32						m_NumLods;
	CImportOptions() : m_pObjectOptions(NULL), m_NumOptions(0), m_MatchExistingBones(true), m_DoRig(false){
	}
	~CImportOptions(){
		if (m_pObjectOptions)
			delete [] m_pObjectOptions;
	}
	bool Init(SMeshHeader &hdr){
		UINT32 i, j;
		m_NumOptions = hdr.m_pMesh.Ptr->NumObjects;
		m_pObjectOptions = new SImportOption[m_NumOptions];
		INT16* pIdx;
		PTR_WARN
		pIdx = hdr.LodMeshIndices.Ptr;
		if (pIdx){
			for (i = 0; i < hdr.NumUpgradeLods; i++){
				for (j = 0; j < hdr.LodCount; j++){
					//if (hdr.UpgradeNames.Ptr)
					if (pIdx[i*hdr.LodCount+j] != -1){
						if (hdr.UpgradeNames.Ptr)
							m_pObjectOptions[pIdx[i*hdr.LodCount+j]].Prop = hdr.UpgradeNames.Ptr[i];
						m_pObjectOptions[pIdx[i*hdr.LodCount+j]].LOD = j;
						m_pObjectOptions[pIdx[i*hdr.LodCount+j]].PartIndex = (INT32)i;
					}
				}
			}
		}
		m_NumLods = hdr.LodCount;
		m_DoRig = hdr.m_pMesh.Ptr->Flags & 2 && hdr.RigBoneCount > 0 ? true : false;
		return true;
	}
	bool ShowUI(){
		SplitPath(m_ModelName, &m_Path, NULL, NULL);
		if (DialogBoxParamA(g_hInstance, MAKEINTRESOURCEA(IDD_IMPORT_OPTION), GetCOREInterface()->GetMAXHWnd(), OptionDlgProc, (LPARAM)this) == -1)
			return false;
		return true;
	}
};



bool GetOpenFile(CStringBufferA &buf, const char* pInitialDir, int *FileOffset, int *ExtensionOffset, const char* pFilter){
	char lBuf[MAX_PATH] = "";
	OPENFILENAMEA of;

	memset(&of, 0, sizeof(OPENFILENAMEA));

	of.lStructSize = sizeof(OPENFILENAMEA);
	of.hwndOwner = NULL;
	of.hInstance = NULL;
	of.lpstrFilter = pFilter;// L"String table\0*.tdb\0\0";
	of.lpstrCustomFilter = NULL;
	of.nMaxCustFilter = 0;
	of.nFilterIndex = 0;
	of.lpstrFile = lBuf;
	of.nMaxFile = MAX_PATH;
	of.lpstrFileTitle = NULL;
	of.nMaxFileTitle = 0;
	of.lpstrInitialDir = pInitialDir;
	of.lpstrTitle = NULL;
	of.Flags = OFN_FILEMUSTEXIST; /*--*/
	of.nFileOffset = 0;
	of.nFileExtension = 0;
	of.lpstrDefExt = NULL;
	of.lCustData = NULL;
	of.lpfnHook = NULL;
	of.lpTemplateName = NULL;	

	if (!GetOpenFileNameA(&of)) return false;
	if (FileOffset)
		*FileOffset = of.nFileOffset;
	if (ExtensionOffset)
		*ExtensionOffset = of.nFileExtension;
	buf = lBuf;
	return true;
}

INT_PTR CALLBACK OptionDlgProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam){
	HWND hRadio;
	HWND hList;
	UINT32 i, s;
	LVITEMA item;
	CImportOptions* pOptions;
	TStringBufferA<MAX_PATH> buf;
	LVCOLUMNA col;
	pOptions = (CImportOptions*)(LONG_PTR)GetWindowLongPtrA(hDlg, GWLP_USERDATA);
	switch (uMsg){
	case WM_INITDIALOG:
		pOptions = (CImportOptions*)lParam;
		SetWindowLongPtrA(hDlg, GWLP_USERDATA, (LONG_PTR)lParam);
		hRadio = GetDlgItem(hDlg, IDC_Y_UP);
		if (hRadio)
			SendMessage(hRadio, BM_SETCHECK, BST_CHECKED, 0);
		hList = GetDlgItem(hDlg, IDC_LODS);
		if (hList){
			memset(&item, 0, sizeof(LVITEMA));
			//SetWindowLongA(hList, GWL_EXSTYLE, GetWindowLongA(hList, GWL_EXSTYLE) | LVS_EX_CHECKBOXES);
			ListView_SetExtendedListViewStyle(hList, LVS_EX_CHECKBOXES);
			col.mask = LVCF_TEXT | LVCF_WIDTH;
			col.cx = 50;
			col.pszText = "Object";
			SendMessage(hList, LVM_INSERTCOLUMNA, (WPARAM)0, (LPARAM)&col);
			col.cx = 50;
			col.pszText = "LOD";
			SendMessage(hList, LVM_INSERTCOLUMNA, (WPARAM)1, (LPARAM)&col);
			col.cx = 50;
			col.pszText = "Prop";
			SendMessage(hList, LVM_INSERTCOLUMNA, (WPARAM)2, (LPARAM)&col);



			for (i = 0; i < pOptions->m_NumOptions; i++){
				item.iItem = i;
				//Object
				item.iSubItem = 0;
				item.mask = LVIF_TEXT;
				buf.Print("%d", i);
				item.pszText = buf.GetBufferPtr();
				SendMessageA(hList, LVM_INSERTITEMA, 0, (LPARAM)&item);
				//LOD
				item.iSubItem = 1;
				if (pOptions->m_pObjectOptions[i].LOD != 0xFFFF)
					buf.Print("%u", pOptions->m_pObjectOptions[i].LOD);
				else
					buf = "";
				item.pszText = buf.GetBufferPtr();
				SendMessageA(hList, LVM_SETITEMA, 0, (LPARAM)&item);
				//Prop			
				item.iSubItem = 2;
				if (pOptions->m_pObjectOptions[i].Prop != 0xFFFFFFFF)
					buf.Print("%u", pOptions->m_pObjectOptions[i].Prop);
				else
					buf = "";
				item.pszText = buf.GetBufferPtr();
				SendMessageA(hList, LVM_SETITEMA, 0, (LPARAM)&item);

				ListView_SetCheckState(hList, i, pOptions->m_pObjectOptions[i].Import);
			}
			hList = GetDlgItem(hDlg, IDC_LOD_SELECT);
			if (hList){
				if (pOptions->m_NumLods == 0)
					EnableWindow(hList, FALSE);
				for (i = 0; i < pOptions->m_NumLods; i++){
					buf.Print("LOD %u", i);
					SendMessageA(hList, CB_ADDSTRING, (WPARAM)0, (LPARAM)buf());
				}
			}
			if (!pOptions->m_DoRig){
				EnableWindow(GetDlgItem(hDlg, IDC_RIG_NAME), FALSE);
				EnableWindow(GetDlgItem(hDlg, IDC_RIG_SELECT), FALSE);
			}
			SetWindowTextA(GetDlgItem(hDlg, IDC_MODEL_NAME), pOptions->m_ModelName);
		}
		/*hRadio = GetDlgItem(hDlg, IDC_USE_LAYERS);
		if (hRadio)
			SendMessageA(hRadio, BM_SETCHECK, pOptions->UseLayers ? BST_CHECKED : BST_UNCHECKED, 0);*/
		break;
	case WM_COMMAND:
		{
			switch (LOWORD(wParam)){
			case IDC_LODS_ALL:
				if (hList = GetDlgItem(hDlg, IDC_LODS)){
					for (i = 0; i < pOptions->m_NumOptions; i++){
						ListView_SetCheckState(hList, i, 1);
					}
				}
				break;
			case IDC_LODS_NONE:
				if (hList = GetDlgItem(hDlg, IDC_LODS)){
					for (i = 0; i < pOptions->m_NumOptions; i++){
						ListView_SetCheckState(hList, i, 0);
					}
				}
				break;
			case IDC_LOD_SELECT:
				if (HIWORD(wParam) == CBN_SELCHANGE){
					s = (int)SendMessageA((HWND)lParam, CB_GETCURSEL, 0, 0);
					hList = GetDlgItem(hDlg, IDC_LODS);
					if (hList){
						if (s >= 0 && s < pOptions->m_NumLods){
							for (i = 0; i < pOptions->m_NumOptions; i++){
								if (pOptions->m_pObjectOptions[i].LOD != -1){
									if (pOptions->m_pObjectOptions[i].LOD == s){
										ListView_SetCheckState(hList, i, 1);
									} else {
										ListView_SetCheckState(hList, i, 0);
									}
								}
							}
						}
					}
				}
				break;
			case IDOK:

				/*if (SendMessageA(GetDlgItem(hDlg, IDC_Z_UP), BM_GETCHECK, 0, 0) == BST_CHECKED)
					pOptions->ZUp = true;*/
				hList = GetDlgItem(hDlg, IDC_LODS);
				if (hList){
					for (i = 0; i < pOptions->m_NumOptions; i++){
						pOptions->m_pObjectOptions[i].Import = ListView_GetCheckState(hList, i) ? true : false;
					}
				}


				//pOptions->UseLayers = SendMessageA(GetDlgItem(hDlg, IDC_USE_LAYERS), BM_GETCHECK, 0, 0) == BST_CHECKED;
				EndDialog(hDlg, (INT_PTR)1);
				break;
			case IDC_RIG_SELECT:
				if (GetOpenFile(buf, pOptions->m_Path, NULL, NULL, "SR3 PC rig\0*.rig_pc\0\0")){
					pOptions->m_Rig = buf;
					SetWindowTextA(GetDlgItem(hDlg, IDC_RIG_NAME), buf);
				}
				break;
			case IDCANCEL:
				EndDialog(hDlg, -1);
				break;
			default:
				return FALSE;
			}
		}
		break;
	case WM_CLOSE:
		EndDialog(hDlg, -1);
		break;
	default:
		return FALSE;
	}
	return TRUE;
}

bool DoImportModel(const char* pName){
	UINT32 i;
	CModelState state;
	INode** ppBones;
	UINT32 bonecount;
	const char* pPartName;

	if (!state.Init(pName))
		return false;
	SMesh* pMesh;
	PTR_WARN
	pMesh = state.m_Data.m_pMesh.Ptr;
	MStringBuffer<128> nbuf;
	CImportOptions options;
	CVertexRemap remap;
	options.m_ModelName = pName;
	Matrix3 tm;
	options.Init(state.m_Data);
	if (options.ShowUI()){
		ppBones = NULL;
		bonecount = 0;
		PTR_WARN
		if (!remap.Init(state.m_Data.m_pMesh.Ptr))
			return false;
		if (options.m_DoRig && *options.m_Rig()){
			DoImportRigFile(options.m_Rig(), &ppBones, &bonecount);
		}
		for (i = 0; i < pMesh->NumObjects; i++){
			if (options.m_pObjectOptions[i].Import){
				pPartName = state.GetPartName(options.m_pObjectOptions[i].PartIndex);
				if (false && pPartName){
					nbuf = pPartName;
					if (options.m_pObjectOptions[i].LOD != 0xFFFF){
						nbuf.AppendF(_M(" LOD %u"), options.m_pObjectOptions[i].LOD);
					}
					tm = state.GetPartTransform(options.m_pObjectOptions[i].PartIndex);
				} else {
					nbuf = "";
					if (options.m_pObjectOptions[i].Prop != 0xFFFFFFFF){
						nbuf.AppendF(_M("Prop %u "), options.m_pObjectOptions[i].Prop);
					}
					if (options.m_pObjectOptions[i].LOD != 0xFFFF){
						nbuf.AppendF(_M("LOD %u "), options.m_pObjectOptions[i].LOD);
					}
					if (options.m_pObjectOptions[i].PartIndex != -1){
						nbuf.AppendF(_M("Part %d "), options.m_pObjectOptions[i].PartIndex);
					}
					nbuf.AppendF(_M("Object %u"), i);
					tm.IdentityMatrix();
				}
				PTR_WARN
				if (!ImportObject(pMesh, pMesh->Objects.Ptr+i, (MCHAR*)nbuf(), ppBones, bonecount, remap, tm)){
					if (ppBones)
						delete []ppBones;
					return false;
				}
			}
		}
		if (ppBones)
			delete []ppBones;
	}
	return true;
}

class CMdlImport : public SceneImport {
public:
	CMdlImport();
	~CMdlImport();
	int ExtCount();
	const MCHAR *Ext(int n);
	const MCHAR *LongDesc();
	const MCHAR *ShortDesc();
	const MCHAR *AuthorName();
	const MCHAR *CopyrightMessage();
	const MCHAR *OtherMessage1();
	const MCHAR *OtherMessage2();
	unsigned int Version();
	void ShowAbout(HWND hWnd);
	int DoImport(const MCHAR *name,ImpInterface *i,Interface *gi, BOOL suppressPrompts=FALSE);
	static Class_ID	s_ClassID;
};

CMdlImport::CMdlImport(){
}

CMdlImport::~CMdlImport(){
}

int CMdlImport::ExtCount(){
	return 2;	
}

const MCHAR *CMdlImport::Ext(int n){
	switch (n){
	case 0: return _M_TEXT("ccmesh_pc");
	case 1: return _M_TEXT("gcmesh_pc");
	default:
		return NULL;
	}
}

const MCHAR *CMdlImport::LongDesc(){
	return _M_TEXT("SR3 ccmesh_pc importer");
}

const MCHAR *CMdlImport::ShortDesc(){
	return _M_TEXT("MDL import");
}

const MCHAR *CMdlImport::AuthorName(){
	return _M_TEXT("Sir Kane");
}

const MCHAR *CMdlImport::CopyrightMessage(){
	return _M_TEXT("Copyright (c) 2013 Sir Kane");
}

const MCHAR *CMdlImport::OtherMessage1(){
	return NULL;
}

const MCHAR *CMdlImport::OtherMessage2(){
	return NULL;
}

unsigned int CMdlImport::Version(){
	return 100;
}

void CMdlImport::ShowAbout(HWND hWnd){
}

int CMdlImport::DoImport(const MCHAR *name,ImpInterface *i,Interface *gi, BOOL suppressPrompts/*=FALSE*/){
	CStringConvert s;
	return DoImportModel(s.MToUTF8(name)) ? IMPEXP_SUCCESS : IMPEXP_FAIL;
}

Class_ID	CMdlImport::s_ClassID = Class_ID(0x324d0664, 0x826623a);

int CMDLImpClassDesc::IsPublic(){
	return 1;
}

void* CMDLImpClassDesc::Create(BOOL loading /*= FALSE*/){
	return new CMdlImport();
}

const MCHAR* CMDLImpClassDesc::ClassName(){
	return _M_TEXT("SR3MdlImport");
}

SClass_ID CMDLImpClassDesc::SuperClassID(){
	return SCENE_IMPORT_CLASS_ID;
}

Class_ID CMDLImpClassDesc::ClassID(){
	return CMdlImport::s_ClassID;
}

const MCHAR* CMDLImpClassDesc::Category(){
	return _M_TEXT("Scene import");
}

CMDLImpClassDesc	CMDLImpClassDesc::s_Instance;

/*
CCar
*/


class CCarImport : public SceneImport {
public:
	CCarImport();
	~CCarImport();
	int ExtCount();
	const MCHAR *Ext(int n);
	const MCHAR *LongDesc();
	const MCHAR *ShortDesc();
	const MCHAR *AuthorName();
	const MCHAR *CopyrightMessage();
	const MCHAR *OtherMessage1();
	const MCHAR *OtherMessage2();
	unsigned int Version();
	void ShowAbout(HWND hWnd);
	int DoImport(const MCHAR *name,ImpInterface *i,Interface *gi, BOOL suppressPrompts=FALSE);
	static Class_ID	s_ClassID;
};

CCarImport::CCarImport(){
}

CCarImport::~CCarImport(){
}

int CCarImport::ExtCount(){
	return 2;	
}

const MCHAR *CCarImport::Ext(int n){
	switch (n){
	case 0: return _M_TEXT("ccar_pc");
	case 1: return _M_TEXT("gcar_pc");
	default:
		return NULL;
	}
}

const MCHAR *CCarImport::LongDesc(){
	return _M_TEXT("SR3 ccar_pc importer");
}

const MCHAR *CCarImport::ShortDesc(){
	return _M_TEXT("Car import");
}

const MCHAR *CCarImport::AuthorName(){
	return _M_TEXT("Sir Kane");
}

const MCHAR *CCarImport::CopyrightMessage(){
	return _M_TEXT("Copyright (c) 2013 Sir Kane");
}

const MCHAR *CCarImport::OtherMessage1(){
	return NULL;
}

const MCHAR *CCarImport::OtherMessage2(){
	return NULL;
}

unsigned int CCarImport::Version(){
	return 100;
}

void CCarImport::ShowAbout(HWND hWnd){
}

int CCarImport::DoImport(const MCHAR *name,ImpInterface *i,Interface *gi, BOOL suppressPrompts/*=FALSE*/){
	CStringConvert s;
	return DoImportModel(s.MToUTF8(name)) ? IMPEXP_SUCCESS : IMPEXP_FAIL;
}

Class_ID	CCarImport::s_ClassID = Class_ID(0x4698750b, 0x324e575d);

int CCarImpClassDesc::IsPublic(){
	return 1;
}

void* CCarImpClassDesc::Create(BOOL loading /*= FALSE*/){
	return new CCarImport();
}

const MCHAR* CCarImpClassDesc::ClassName(){
	return _M_TEXT("SR3MldImport");
}

SClass_ID CCarImpClassDesc::SuperClassID(){
	return SCENE_IMPORT_CLASS_ID;
}

Class_ID CCarImpClassDesc::ClassID(){
	return CCarImport::s_ClassID;
}

const MCHAR* CCarImpClassDesc::Category(){
	return _M_TEXT("Scene import");
}

CCarImpClassDesc	CCarImpClassDesc::s_Instance;

/*
Rig import
*/

class CRigImport : public SceneImport {
public:
	CRigImport();
	~CRigImport();
	int ExtCount();
	const MCHAR *Ext(int n);
	const MCHAR *LongDesc();
	const MCHAR *ShortDesc();
	const MCHAR *AuthorName();
	const MCHAR *CopyrightMessage();
	const MCHAR *OtherMessage1();
	const MCHAR *OtherMessage2();
	unsigned int Version();
	void ShowAbout(HWND hWnd);
	int DoImport(const MCHAR *name,ImpInterface *i,Interface *gi, BOOL suppressPrompts=FALSE);
	static Class_ID	s_ClassID;
};

CRigImport::CRigImport(){
}

CRigImport::~CRigImport(){
}

int CRigImport::ExtCount(){
	return 1;	
}

const MCHAR *CRigImport::Ext(int n){
	switch (n){
	case 0: return _M_TEXT("rig_pc");
	default:
		return NULL;
	}
}

const MCHAR *CRigImport::LongDesc(){
	return _M_TEXT("Valve Source MDL importer");
}

const MCHAR *CRigImport::ShortDesc(){
	return _M_TEXT("MDL import");
}

const MCHAR *CRigImport::AuthorName(){
	return _M_TEXT("Sir Kane");
}

const MCHAR *CRigImport::CopyrightMessage(){
	return _M_TEXT("Copyright (c) 2013 Sir Kane");
}

const MCHAR *CRigImport::OtherMessage1(){
	return NULL;
}

const MCHAR *CRigImport::OtherMessage2(){
	return NULL;
}

unsigned int CRigImport::Version(){
	return 100;
}

void CRigImport::ShowAbout(HWND hWnd){
}

int CRigImport::DoImport(const MCHAR *name,ImpInterface *i,Interface *gi, BOOL suppressPrompts/*=FALSE*/){
	/*CModelState state;
#if defined(_UNICODE) && !defined(_UNICODE_MODULE_FOR_MBCS_MAX)
	TStringBufferA<MAX_PATH> buf;
	buf = name;
	if (!state.Init(buf, suppressPrompts ? true : false))
		return IMPEXP_FAIL;
#else //_UNICODE
	if (!state.Init(name, suppressPrompts ? true : false))
		return IMPEXP_FAIL;
#endif //!_UNICODE
	return DoImportMdl(state) ? IMPEXP_SUCCESS : IMPEXP_FAIL;*/
	CStringConvert s;
	
	CMemoryBuffer rigbuf;
	if (!LoadFileToBuffer(s.MToUTF8(name), rigbuf))
		return IMPEXP_FAIL;
	CBufferReader r(&rigbuf);

	SRigHeader* pRig = LoadRig(r);
	DoImportRig(pRig, NULL, NULL);
	return IMPEXP_SUCCESS;
}

Class_ID	CRigImport::s_ClassID = Class_ID(0x11b35c48, 0x33b17bab);

int CRigImpClassDesc::IsPublic(){
	return 1;
}

void* CRigImpClassDesc::Create(BOOL loading /*= FALSE*/){
	return new CRigImport();
}

const MCHAR* CRigImpClassDesc::ClassName(){
	return _M_TEXT("SR3RigImport");
}

SClass_ID CRigImpClassDesc::SuperClassID(){
	return SCENE_IMPORT_CLASS_ID;
}

Class_ID CRigImpClassDesc::ClassID(){
	return CRigImport::s_ClassID;
}

const MCHAR* CRigImpClassDesc::Category(){
	return _M_TEXT("Scene import");
}

CRigImpClassDesc	CRigImpClassDesc::s_Instance;