#ifndef _SR3IL_H_
#define _SR3IL_H_

#include <Windows.h>
#include <d3d9.h>

enum {
	SRVDecl_Position0 = 0,
	SRVDecl_Position1,
	SRVDecl_Position2,
	SRVDecl_Position3,
	SRVDecl_Position4,
	SRVDecl_Position5,
	SRVDecl_Position6,
	SRVDecl_Position7,
	SRVDecl_Texcoord0,
	SRVDecl_Texcoord1,
	SRVDecl_Texcoord2,
	SRVDecl_Texcoord3,
	SRVDecl_Texcoord4,
	SRVDecl_Texcoord5,
	SRVDecl_Texcoord6,
	SRVDecl_Texcoord7,
	SRVDecl_Normal0,
	SRVDecl_Normal1,
	SRVDecl_Normal2,
	SRVDecl_Normal3,
	SRVDecl_Normal4,
	SRVDecl_Normal5,
	SRVDecl_Normal6,
	SRVDecl_Normal7,
	SRVDecl_Tangent0,
	SRVDecl_Tangent1,
	SRVDecl_Tangent2,
	SRVDecl_Tangent3,
	SRVDecl_Tangent4,
	SRVDecl_Tangent5,
	SRVDecl_Tangent6,
	SRVDecl_Tangent7,
	SRVDecl_Binormal0,
	SRVDecl_Binormal1,
	SRVDecl_Binormal2,
	SRVDecl_Binormal3,
	SRVDecl_Binormal4,
	SRVDecl_Binormal5,
	SRVDecl_Binormal6,
	SRVDecl_Binormal7,
	SRVDecl_Color0,
	SRVDecl_Color1,
	SRVDecl_Color2,
	SRVDecl_Color3,
	SRVDecl_Color4,
	SRVDecl_Color5,
	SRVDecl_Color6,
	SRVDecl_Color7,
	SRVDecl_BlendIndices,
	SRVDecl_BlendWeight,
	SRVDecl_PSize,
	SRVDecl_Last,
};

enum {
	SRVType_Float1,
	SRVType_Float2,
	SRVType_Float3,
	SRVType_Float4,
	SRVType_Float16_2,
	SRVType_Float16_4,
	SRVType_UByte4,
	SRVType_UByte4N,
	SRVType_Short2N,
	SRVType_Short2N_A,
	SRVType_Short2,
	SRVType_Short2_A,
	SRVType_UByte4N_A,
	SRVType_UByte4N_B,
	SRVType_D3DCOLOR,
	SRVType_Short4,
	SRVType_UByte4_A,
	SRVType_Last,
};

struct SInputValue {
	float	x;
	float	y;
	float	z;
	float	w;
	int		ix;
	int		iy;
	int		iz;
	int		iw;
	int		n; //Number of elements
};

class CStridedPointer {
protected:
	UINT8 const*	m_pData;
	UINT8 const*	m_pCurrent;
	UINT32			m_Stride;
	CStridedPointer(){}
public:
	CStridedPointer(const void* pData, UINT32 stride) : m_pData((UINT8 const*)pData), m_Stride(stride){
		m_pCurrent = m_pData;
	}
	CStridedPointer(CStridedPointer &b): m_pData(b.m_pData), m_pCurrent(b.m_pCurrent), m_Stride(b.m_Stride){	
	}
	const void* operator ()(){
		return m_pCurrent;
	}
	const void* operator[](UINT32 index){
		return m_pData + (index*m_Stride);
	}
	CStridedPointer &operator++(){
		m_pCurrent += m_Stride;
		return *this;
	}
	CStridedPointer operator++(int){
		CStridedPointer b(*this);
		m_pCurrent += m_Stride;
		return b;
	}
};


struct SInputLayoutElement {
	UINT32	Usage;
	UINT32	Type;
	UINT8	Stream;
	UINT8	Offset; //0xFF for sequential
	UINT8	Field_A;
	UINT8	Field_B;
	UINT32	Field_C;
	enum { AutoOffset = 0xFF };
};

struct SInputLayout {
	SInputLayoutElement*	Elements;
	UINT32					Count;
};

SInputValue &ToValue(SInputValue &val, const void* p, SInputLayoutElement* pElement);

/*
^(.*),(?=\r?$)
->
case $1: return "$1";
*/


const char* GetSRVDecl(UINT32 type);
const char* GetSRVType(UINT32 type);
extern SInputLayout*	g_InputLayouts[110];

SInputLayout* GetIL(UINT32 ID);
SInputLayoutElement* GetElement(SInputLayout *pIL, UINT32 usage);


#endif //!_SR3IL_H_
