#include <stdlib.h>
#include <stdio.h>
#include <Windows.h>
#include "Base64.h"

static const unsigned char s_Base64EncodeTable[64] = {
	'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', //+26: 26
	'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', //+26: 52
	'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '+', '/', //+12: 64
};

static const char s_Base64DecodeTable[256] = {
	/* 00 ... */ -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	/* 20 ... */ -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 62, -1, -1, -1, 63, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, -1, -1, -1, -1, -1, -1,
	/* 40 ... */ -1,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, -1, -1, -1, -1, -1,
	/* 60 ... */ -1, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, -1, -1, -1, -1, -1,
	/* 80 ... */ -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	/* A0 ... */ -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	/* C0 ... */ -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	/* E0 ... */ -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
};

#define BASE64_UNREACHABLE	__assume(0)

typedef struct _Base64Value {
	union {
		struct {
			DWORD	O2 : 8;
			DWORD	O1 : 8;
			DWORD	O0 : 8;
		};
		struct {
			DWORD	S3 : 6;
			DWORD	S2 : 6;
			DWORD	S1 : 6;
			DWORD	S0 : 6;
		};
		DWORD	V;
	};
} Base64Value;

int Base64Encode(const void* _src, int _srclen, void* _dst, int _dstlen){
	unsigned char* pDst;
	const unsigned char* pSrc;
	char tmp[2] = { 0, 0 };

	if (!_src || !_dstlen)
		return Base64Encode_InvalidPointer;
	if (_srclen < 0)
		return Base64Encode_InvalidLen;

	if (_dstlen < Base64EncodeLen(_srclen))
		return Base64Encode_BufferTooSmall;
	pDst = (unsigned char*)_dst;
	pSrc = (const unsigned char*)_src;

	for (;_srclen > 2; _srclen -= 3, pSrc += 3){
		*(pDst++) = s_Base64EncodeTable[(pSrc[0] >> 2)];
		*(pDst++) = s_Base64EncodeTable[((pSrc[0] & 3) << 4) | (pSrc[1] >> 4)];
		*(pDst++) = s_Base64EncodeTable[((pSrc[1] & 0xF) << 2) | (pSrc[2] >> 6)];
		*(pDst++) = s_Base64EncodeTable[pSrc[2] & 0x3F];
	}
	switch (_srclen){
	case 2:
		tmp[1] = pSrc[1];
		pDst[2] = s_Base64EncodeTable[(tmp[1] & 0xF) << 2];
	case 1:
		tmp[0] = pSrc[0];
		pDst[0] = s_Base64EncodeTable[(tmp[0] >> 2)];
		pDst[1] = s_Base64EncodeTable[((tmp[0] & 3) << 4) | (tmp[1] >> 4)];
		memcpy(pDst+_srclen+1, "==", 3-_srclen);
		pDst += 4;
	case 0:
		break;
	default: BASE64_UNREACHABLE;
	}
	return (int)(pDst-(unsigned char*)_dst);
}

int Base64Decode(const void* _src, int _srclen, void* _dst, int _dstlen){
	unsigned char* pDst;
	const unsigned char* pSrc;
	char tmp[2] = { 0, 0 };
	unsigned char c;
	Base64Value v;
	int padding;

	if (!_src || !_dstlen)
		return Base64Decode_InvalidPointer;
	if (_srclen < 4 || (_srclen & 3) != 0)
		return Base64Encode_InvalidLen;

	if (_dstlen < Base64DecodeLen(_srclen))
		return Base64Encode_BufferTooSmall;
		
	pDst = (unsigned char*)_dst;
	pSrc = (const unsigned char*)_src;

	c = 0;
	padding = (pSrc[_srclen-1] == '=') + (pSrc[_srclen-2] == '=');
	
	for (_srclen -= padding;_srclen > 3;_srclen -= 4){
		c |= v.S0 = s_Base64DecodeTable[*(pSrc++)];
		c |= v.S1 = s_Base64DecodeTable[*(pSrc++)];
		c |= v.S2 = s_Base64DecodeTable[*(pSrc++)];
		c |= v.S3 = s_Base64DecodeTable[*(pSrc++)];
		*(pDst++) = v.O0;
		*(pDst++) = v.O1;
		*(pDst++) = v.O2;
	}
	if (c == -1)
		return Base64Decode_InvalidCharacter;
	v.V = 0;
	switch (padding){
	case 1:
		c |= v.S2 = s_Base64DecodeTable[pSrc[2]];
		c |= v.S1 = s_Base64DecodeTable[pSrc[1]];
		pDst[1] = v.O1;
	case 2:
		c |= v.S0 = s_Base64DecodeTable[pSrc[0]];
		c |= v.S1 = s_Base64DecodeTable[pSrc[1]];
		pDst[0] = v.O0;
		pDst += 3-padding;
	case 0:
		break;
	default:
		BASE64_UNREACHABLE;
	}
	return (int)(pDst-(unsigned char*)_dst);
}
