#ifndef _BINXML_H_
#define _BINXML_H_

#include <windows.h>

#define bxml_swapuint32_m(v) ((((v) & 0xFF) << 24) | (((v) & 0xFF00) << 8) | (((v) & 0xFF0000) >> 8) | (((v) & 0xFF000000) >> 24))

typedef struct _SXMLTag{
	UINT8	ID;
	UINT32	Size;
	void*	Data;
}SXMLTag, *lpSXMLTag;

class CBinXMLData{
private:
	unsigned char*	m_pData;
	UINT32			m_Size;
	UINT32			m_Offset;
	CBinXMLData(){
	}
	bool HasData(UINT32 len){
		return (m_Offset+len) <= m_Size;
	}
	template<typename T> bool Read_t(T&obj){
		if ((m_Offset+sizeof(T)) > m_Size)
			return false;
		memcpy(&obj, m_pData+m_Offset, sizeof(T));
		m_Offset += sizeof(T);
		return true;
	}
public:
	CBinXMLData(void*data, UINT32 size) : m_pData((unsigned char*)data), m_Size(size), m_Offset(0){
	}
	CBinXMLData(lpSXMLTag tag) : m_pData((unsigned char*)(tag->Data)), m_Size(tag->Size), m_Offset(0){
	}
	int ReadTag(lpSXMLTag tag){
		UINT8 b, i;
		UINT32 l;
		if (!Read_t(tag->ID))
			return 1;
		tag->Size = 0;
		if (!Read_t(b))
			return -1;
		if (b < 0x80){
			tag->Size = b;
		} else if (b == 0xFF) {
			if (!Read_t(l))
				return -1;
			tag->Size = bxml_swapuint32_m(l);
		} else {
			tag->Size = (b & 0x7F) << 8;
			if (!Read_t(b))
				return -1;
			tag->Size |= b;
		}
		tag->Data = m_pData+m_Offset;
		if (!HasData(tag->Size))
			return -1;
		m_Offset += tag->Size;
		return 0;
	}
};

#endif //!_BINXML_H_
