#include "Config.h"
#include <stdio.h>
#include <stdlib.h>
#include "Memory.h"


inline bool IsWhitespace(int c){
	return ((c == ' ') | (c == '\t') | (c == '\n') | (c == '\t'));
}

CConfig::CConfig(){
}

bool CConfig::GetLineEnd(char* pStr, char** ppEnd, char** ppNextStart){
	int ch, len;
	if (!pStr || !ppEnd)
		return false;
	*ppEnd = NULL;
	for (ch = PeekUTF8l_i(pStr, &len);ch; ch = PeekUTF8l_i(pStr, &len)){
		if (ch == '\r' || ch == '\n'){
			*ppEnd = pStr;
			break;
		}
		pStr += len;
	}
	if (!*ppEnd){
		*ppEnd = pStr;
		if (ppNextStart)
			*ppNextStart = NULL;
	} else if (ppNextStart){
		if (*pStr){
			for (ch = PeekUTF8l_i(pStr, &len);ch && (ch == '\r' || ch == '\n'); ch = PeekUTF8l_i(pStr, &len)){
				pStr += len;
			}
			*ppNextStart = pStr;
		} else
			*ppNextStart = NULL;
	}
	return true;
}

char* CConfig::SkipWS(char* pStr){
	int ch;
	int len;
	for (ch = PeekUTF8l_i(pStr, &len);ch; ch = PeekUTF8l_i(pStr, &len)){
		if (!IsWhitespace(ch))
			break;
		pStr += len;
	}
	return pStr;
}

char* CConfig::FindStr(char* pStr, char* pCompare){
	bool q;
	char c;
	size_t len;
	len = strlen(pCompare);
	q = false;
	for (c = *pStr; c; c = *(++pStr)){
		if (!q){
			if (strncmp(pStr, pCompare, len) == 0)
				return pStr;
		}
		q ^= (c == '\"');
		if (c == '\\')
			pStr += (*(pStr+1) == '\"');
	}
	return NULL;
}

char* CConfig::InverseGetLastWS(char* pStart, char* pStr){
	int ch, len;
	for (pStr = UTF8Prev(pStart, pStr);pStr;pStr = UTF8Prev(pStart, pStr)){
		ch = PeekUTF8l_i(pStr, &len);
		if (!IsWhitespace(ch))
			return ch == 0 ? pStr : pStr+len;
	}
	return pStart;
}

void CConfig::ReadQuoted(char* pStr, CStringBufferA &res){
	int ch;
	bool q;
	q = false;
	for (ch = GetUTF8_i(pStr); ch; ch = GetUTF8_i(pStr)){
		switch (ch){
			case '\"':
				q ^= true;
				break;
			case '\\':
				if (PeekUTF8_i(pStr) == '\"'){
					GetUTF8_i(pStr);
					ch = '\"';
				}
			default:
				res.AppendChr(ch);
		}
	}
}

bool CConfig::Parse(const char* pFileName, bool (*pfnCallback)(const char* key, const char* value, void* ctx), void* ctx){
	FILE*f;
	long size;
	char* pCache;
	char* pNextLine;
	char* pLineEnd;
	char* pCur;
	char* pEqual;
	char* pComment;
	char* pComment2;
	TStringBufferA<128> key;
	TStringBufferA<128> value;
	if (fopen_s(&f, pFileName, "rb") != 0)
		return false;
	fseek(f, 0, SEEK_END);
	size = ftell(f);
	fseek(f, 0, SEEK_SET);

	pCache = (char*)g_pAllocator(size+1);
	if (!pCache)
		return fclose(f), false;
	
	if (fread(pCache, 1, size, f) != size)
		return g_pFreeer(pCache), fclose(f), false;
	fclose(f);
	pCache[size] = 0;

	pCur = pCache;
	for (;*pCur;){
		if (!GetLineEnd(pCur, &pLineEnd, &pNextLine))
			break;
		*pLineEnd = 0;
		pCur = SkipWS(pCur);
		pEqual = FindStr(pCur, "=");
		pComment = FindStr(pCur, "--");
		pComment2 = FindStr(pCur, ";");
		if (pComment2 > pComment)
			pComment = pComment2;
		if (pComment)
			pLineEnd = pComment;
		if(pComment == pCur){
			if (!pNextLine)
				break;
			pCur = pNextLine;
			continue;
		}
		pComment = pLineEnd;
		if (pEqual){
			*pEqual = 0;
			pEqual++;
			pComment = pEqual-2;
		}
		pComment = InverseGetLastWS(pCur, pComment);
		if (*pComment)
			*pComment = 0;
		key.Clear();
		ReadQuoted(pCur, key);
		value.Clear();
		if (pEqual){
			pCur = SkipWS(pEqual);
			pComment = InverseGetLastWS(pCur, pLineEnd);
			if (pComment)
				*pComment = NULL;
			ReadQuoted(pCur, value);
		}

		if (!pfnCallback(key, value, ctx))
			break;



		pCur = pNextLine;
		if (!pNextLine)
			break;
	}
	
	return g_pFreeer(pCache), true;
}


void CSimpleConfig::AddValue(const char* pKey, const char* pValue){
	CConfigValue* pVal;
	pVal = new CConfigValue();
	if (!pVal)
		return;
	pVal->m_Key = pKey;
	pVal->m_Value = pValue;
	if (!m_Values.Add_Head(pVal))
		delete pVal;
}

bool CSimpleConfig::ParseCallback(const char* key, const char* value, CSimpleConfig* ctx){
	ctx->AddValue(key, value);
	return true;
}

CSimpleConfig::CSimpleConfig(){
}

CSimpleConfig::~CSimpleConfig(){
	Clear();
}

void CSimpleConfig::Clear(){
	m_Values.Remove_All(SList<CConfigValue>::DefaultCallback);
}

bool CSimpleConfig::Load(const char* pPath){
	CConfig cfg;
	Clear();
	return cfg.TParse(pPath, CSimpleConfig::ParseCallback, this);
}

const char* CSimpleConfig::GetValue(const char* pKey){
	GenericSLNode<CConfigValue>* pNode;
	for (pNode = m_Values.Get_Head(); pNode; pNode = pNode->m_pNext){
		if (_stricmp(pKey, pNode->m_pData->m_Key) == 0)
			return pNode->m_pData->m_Value;
	}
	return NULL;
}
