#ifndef _FASTLOCK_H_
#define _FASTLOCK_H_

#include <Windows.h>
#include <intrin.h>

//Doesn't check against multiple locks from same thread
class CSingletonFastLock {
protected:
	volatile LONG	m_Spin;
public:
	inline CSingletonFastLock(){
		m_Spin = 0;
	}
	inline ~CSingletonFastLock(){
	}
	inline void Lock(){
		for (;_InterlockedCompareExchange(&m_Spin, 1, 0) != 0;)
			;
	}
	inline void Unlock(){
		_InterlockedExchange(&m_Spin, 0);
	}
};

//Save to lock again if already locked by a thread
class CFastLock {
protected:
	CSingletonFastLock	m_ThreadIDLock;
	DWORD				m_OwningThreadID;
	volatile LONG		m_Spin;
public:
	inline CFastLock(){
		m_OwningThreadID = 0;
		m_Spin = 0;
	}
	inline ~CFastLock(){
	}
	inline void Lock(){
		DWORD tid;
		tid = GetCurrentThreadId();
		m_ThreadIDLock.Lock();
		if (m_OwningThreadID == tid){
			_InterlockedIncrement(&m_Spin);
			m_ThreadIDLock.Unlock();
			return;
		}
		m_ThreadIDLock.Unlock();
		for (;_InterlockedCompareExchange(&m_Spin, 1, 0) != 0;)
			;
		m_ThreadIDLock.Lock();
		m_OwningThreadID = tid;
		m_ThreadIDLock.Unlock();
	}
	inline void Unlock(){
		DWORD tid;
		tid = GetCurrentThreadId();
		m_ThreadIDLock.Lock();
		if (m_OwningThreadID == tid){
			if (_InterlockedDecrement(&m_Spin) == 0)
				m_OwningThreadID = 0;
		}
		m_ThreadIDLock.Unlock();
	}
	class CFastLockRegion {
	private:
		CFastLock&	m_Lock;
		inline CFastLockRegion() : m_Lock(*((CFastLock*)NULL)){}
	public:
		inline CFastLockRegion(CFastLock &lock) : m_Lock(lock){
			m_Lock.Lock();
		}
		inline ~CFastLockRegion(){
			m_Lock.Unlock();
		}
	};
};

#endif //!_FASTLOCK_H_
