#ifndef _FILEFIND_H_
#define _FILEFIND_H_


#include "Lists.h"
#include "StringBuffer.h"
#include "String.h"

enum FindResult {
	FindResult_FullPath,
	FindResult_RelativePath,
	FindResult_Name,	
};

class CFindFile {
private:
	SList<CStringA>				m_Files;
	int							m_Count;
	int							m_Index;
	GenericSLNode<CStringA>*	m_pCurrentNode;
	void InternalFind(CStringBufferA &filter, CStringBufferA &subpath, CStringBufferA &temp, CStringBufferA &name, const char* pFilter, FindResult rmode){
		int len, len2;
		HANDLE hFind;
		WIN32_FIND_DATAA fd;
		CStringA* pString;
		filter = temp;
		filter.Append(pFilter);
		hFind = FindFirstFileA(filter, &fd);
		if (hFind != INVALID_HANDLE_VALUE){
			do {
				if (!(fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)){
					pString = new CStringA;
					if (pString){
						switch (rmode){
						case FindResult_FullPath:
							name.Print("%s%s", temp(), fd.cFileName);
							(*pString) = name;
							break;
						case FindResult_RelativePath:
							name.Print("%s%s", subpath(), fd.cFileName);
							(*pString) = name;
							break;
						case FindResult_Name:
							(*pString) = fd.cFileName;
							break;
						}
						if (!m_Files.Add_Tail(pString))
							delete pString;
					}
				}
			} while (FindNextFileA(hFind, &fd));
			FindClose(hFind);
		}
		filter = temp;
		filter.Append("*");
		hFind = FindFirstFileA(filter, &fd);
		if (hFind != INVALID_HANDLE_VALUE){
			do {
				if (fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY && (_stricmp(fd.cFileName, "..") != 0 && _stricmp(fd.cFileName, ".") != 0)){
					len = temp.GetOffset();
					len2 = subpath.GetOffset();
					temp.AppendF("%s\\", fd.cFileName);
					subpath.AppendF("%s\\", fd.cFileName);
					InternalFind(filter, subpath, temp, name, pFilter, rmode);
					temp.SetOffset(len);
					temp.ZeroTerminate();

					subpath.SetOffset(len2);
					subpath.ZeroTerminate();
				}
			} while (FindNextFileA(hFind, &fd));
			FindClose(hFind);
		}
	}
public:
	CFindFile(){
		m_Count = m_Index = 0;
	}
	~CFindFile(){
		Clear();
	}
	void Clear(){
		m_Files.Remove_All(SList<CStringA>::DefaultCallback);
	}
	int Find(const char* pDir, const char* pFilter, FindResult resultmode = FindResult_FullPath, bool recursive = false){
		CStringBufferA filter, name, temp, subpath;
		int len, len2;
		HANDLE hFind;
		WIN32_FIND_DATAA fd;
		filter = pDir;
		temp = filter.Right(1);
		CStringA* pString;
		Clear();
		if (temp[0] != '\\' && temp[0] != '/')
			filter.Append("\\");
		temp = filter;
		filter.Append(pFilter);
		hFind = FindFirstFileA(filter, &fd);
		if (hFind != INVALID_HANDLE_VALUE){
			do {
				if (!(fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)){
					pString = new CStringA;
					if (pString){
						switch (resultmode){
						case FindResult_FullPath:
							name.Print("%s%s", temp(), fd.cFileName);
							(*pString) = name;
							break;
						case FindResult_RelativePath:
							name.Print("%s%s", subpath(), fd.cFileName);
							(*pString) = name;
							break;
						case FindResult_Name:
							(*pString) = fd.cFileName;
							break;
						}
						if (!m_Files.Add_Tail(pString))
							delete pString;
					}
				}/* else if (recursive && (stricmp(fd.cFileName, "..") != 0 && stricmp(fd.cFileName, ".") != 0)){
					len = temp.GetOffset();
					temp.AppendF("%s\\", fd.cFileName);
					subpath.Print("%s\\", fd.cFileName);
					InternalFind(filter, subpath, temp, name, pFilter, resultmode);
					temp.SetOffset(len);
					temp.ZeroTerminate();
				}*/
			} while (FindNextFileA(hFind, &fd));
			FindClose(hFind);
		}
		if (recursive){
			filter = temp;
			filter.Append("*");
			hFind = FindFirstFileA(filter, &fd);
			if (hFind != INVALID_HANDLE_VALUE){
				do {
					if (fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY && (stricmp(fd.cFileName, "..") != 0 && stricmp(fd.cFileName, ".") != 0)){
						len = temp.GetOffset();
						temp.AppendF("%s\\", fd.cFileName);
						subpath.Print("%s\\", fd.cFileName);
						InternalFind(filter, subpath, temp, name, pFilter, resultmode);
						temp.SetOffset(len);
						temp.ZeroTerminate();
					}
				} while (FindNextFileA(hFind, &fd));
				FindClose(hFind);
			}
		}
		m_pCurrentNode = m_Files.Get_Head();
		return (m_Count = m_Files.Get_Count());
	}
	const char* GetNext(){
		const char* pStr;
		if (!m_pCurrentNode)
			return NULL;
		pStr = (*(m_pCurrentNode->m_pData));
		m_pCurrentNode = m_pCurrentNode->m_pNext;
		m_Index++;
		return pStr;
	}
	void Reset(){
		m_pCurrentNode = m_Files.Get_Head();
		m_Index = 0;
	}
	const char* Get(int index){
		GenericSLNode<CStringA>* pNode;
		pNode = m_Files[index];
		return pNode ? (*(pNode->m_pData)) : (const char*)NULL;
	}
	const char* operator()(){
		return GetNext();
	}
	const char* operator[](int index){
		return Get(index);
	}
	int GetCount(){
		return m_Count;
	}
};

#endif //!_FILEFIND_H_
