#ifndef _RESACCESSOR_H_
#define _RESACCESSOR_H_

#include <Windows.h>

class CResource {
private:
	HRSRC	m_hResource;
	HGLOBAL	m_hGlobal;
	DWORD	m_Size;
	void*	m_pData;
	DWORD	m_RefCount;
	HMODULE	m_hModule;
public:
	CResource() : m_hResource(NULL), m_hGlobal(NULL), m_pData(NULL), m_Size(NULL) {
		m_RefCount = 0;
	}
	~CResource(){
		Discard();
	}
	bool Acquire(HMODULE hModule, LPCSTR lpName, LPCSTR lpType){
		if (Discard()){
			m_hResource = FindResourceA(hModule, lpName, lpType);
			if (!m_hResource)
				return false;
			m_hGlobal = LoadResource(hModule, m_hResource);
			if (!m_hGlobal)
				return m_hResource = NULL, false;
			m_hModule = hModule;
			return true;
		}
		return false;
	}
	bool Discard(){
		if (!m_hResource)
			return true;
		if (m_pData)
			return false;
		FreeResource(m_hGlobal);
		m_hGlobal = NULL;
		m_hResource = NULL;
		return true;
	}
	void* Lock(DWORD* lpSize){
		if (m_pData){
			m_RefCount++;
			if (lpSize) *lpSize = m_Size;
			return m_pData;
		}
		if (m_hResource == NULL)
			return NULL;
		m_RefCount = 1;
		m_pData = LockResource(m_hGlobal);
		m_Size = SizeofResource(m_hModule, m_hResource);
		if (lpSize)
			*lpSize = m_Size;
		return m_pData;

	}
	void Unlock(){
		if (!m_pData)
			return;
		if (--m_RefCount == 0){
			UnlockResource(m_hResource);
			m_pData = NULL;
			m_Size = 0;			
		}
	}
};

#endif //!_RESACCESSOR_H_
