#include <Windows.h>
#include "Str.h"

wchar_t* UTF16Compose(int _chr, wchar_t* _dst, int _buflen){
	if (_chr < 1 || _chr > WCH_UTF16_LAST || !_dst || _buflen < 1 || (_chr >= WCH_UTF16_HIGH_FIRST && _chr <= WCH_UTF16_LOW_LAST))
		return NULL;
	if ((_chr >= 0 && _chr < WCH_UTF16_HIGH_FIRST) || (_chr > WCH_UTF16_LOW_LAST && _chr < 0x10000)){
		*_dst = (wchar_t)_chr;
		return _dst+1;
	}
	if (_buflen < 2)
		return NULL;
	_chr -= 0x10000;
	*(_dst++) = WCH_UTF16_HIGH_FIRST+(_chr >> 10);
	*(_dst++) = WCH_UTF16_LOW_FIRST+(_chr & 0x3FF);
	return _dst;
}

int UTF16Decompose(const wchar_t* _str, int &_len, bool &_valid){
	int chr;
	if (!_str)
		return 0;
	
	//Surrogate pair
	if (*_str >= WCH_UTF16_HIGH_FIRST && *_str <= WCH_UTF16_HIGH_LAST){
		chr = (*_str-WCH_UTF16_HIGH_FIRST) << 10;
		_str++;
		if (*_str < WCH_UTF16_LOW_FIRST || *_str > WCH_UTF16_LOW_LAST)
			return _valid = false, _len = 1, UTF16_INVALIDCODEPOINT;
		chr = *_str-WCH_UTF16_LOW_FIRST+0x10000;
		_len = 2;
		_valid = true;
		return chr;
	}
	_len = 1;
	_valid = true;
	return *_str;
}

inline bool ValidCodepoint(int ch){
	return (((ch >= WCH_UTF16_HIGH_FIRST)-(ch <= WCH_UTF16_LOW_LAST))|(ch > WCH_UTF16_LAST)|(1-((ch >= 192)-(ch <= 193)))|(1-((ch >= 245)-(ch <= 255)))) != 0;
}

char* UTF8Compose(int _ch, char* _buf, int _buflen){
	DWORD bpos;
	if (_ch < 0){
		return _buf;
	}
	if (_ch <= 0x7F){
		*_buf = (char)_ch;
		return ++_buf;
	}
	_BitScanReverse(&bpos, *(DWORD*)&_ch);
	bpos = ((bpos-6)/5)+1;
	if ((int)bpos+1>=_buflen){
		return _buf;
	}

	_buf += bpos;
	switch (bpos-1){
	case 4: UTF8_PUTTRAILI(_ch, _buf);
	case 3: UTF8_PUTTRAILI(_ch, _buf);
	case 2: UTF8_PUTTRAILI(_ch, _buf);
	case 1: UTF8_PUTTRAILI(_ch, _buf);
	case 0: UTF8_PUTTRAILI(_ch, _buf);
		UTF8_PUTLEAD(_ch, _buf, bpos);
		return _buf+bpos+1;
	default:UTF_UNREACHABLE;
	}
}

int UTF8Decompose(const char*_str, int &_len, bool &_valid){
	unsigned char c;
	unsigned char* pStr;
	int ch;
	DWORD bpos;
	DWORD ntrails;
	pStr = (unsigned char*)_str;
	c = *(pStr++);
	if (c <= 0x7F){
		return _len = 1, _valid = true, c;
	}
	if (c == 0xFF){
		return _len = 1, _valid = false, UTF8_INVALIDCODEPOINT;
	}
	_BitScanReverse(&bpos, ~(DWORD)(0xFFFFFF00 | c));
	ntrails = 6-bpos;
	ch = c & (0xFF >> (8-bpos));
	switch (ntrails-1){
	case 4: UTF8_GETTRAIL(ch, pStr, c, __utf8_invalid);
	case 3: UTF8_GETTRAIL(ch, pStr, c, __utf8_invalid);
	case 2: UTF8_GETTRAIL(ch, pStr, c, __utf8_invalid);
	case 1: UTF8_GETTRAIL(ch, pStr, c, __utf8_invalid);
	case 0: UTF8_GETTRAIL(ch, pStr, c, __utf8_invalid);
		break;
	default: UTF_UNREACHABLE;
	}
#ifdef UTF8_VALIDATELENGTH
	_BitScanReverse(&bpos, *(DWORD*)&ch);
	bpos = ((bpos-6)/5)+1;
	if (bpos != ntrails)
		goto __utf8_invalid;
#endif //UTF8_VALIDATELENGTH
#ifdef UTF8_VALIDATECODEPOINTS
	if (!ValidCodepoint(ch))
		goto __utf8_invalid;
#endif //!UTF8_VALIDATECODEPOINTS
	return _len = (int)(pStr-(unsigned char*)_str), _valid = true, ch;
__utf8_invalid:
	return _len = (int)(pStr-(unsigned char*)_str), _valid = false, UTF8_INVALIDCODEPOINT;
}

char* UTF8strchr(char *_str, int _chr){
	int ch;
	char *str;
	for (str = _str, ch = GetUTF8_i(_str); ch; str = _str, ch = GetUTF8_i(_str)){
		if (ch == _chr)
			return str;
	}
	return NULL;
}

const char* UTF8strchr(const char *_str, int _chr){
	int ch;
	const char *str;
	for (str = _str, ch = GetUTF8_i(_str); ch; str = _str, ch = GetUTF8_i(_str)){
		if (ch == _chr)
			return str;
	}
	return NULL;
}
