#ifndef _STR_H_
#define _STR_H_

#define WCH_UTF16_HIGH_FIRST    ((WCHAR)0xd800)
#define WCH_UTF16_HIGH_LAST     ((WCHAR)0xdbff)
#define WCH_UTF16_LOW_FIRST     ((WCHAR)0xdc00)
#define WCH_UTF16_LOW_LAST      ((WCHAR)0xdfff)
#define WCH_UTF16_LAST			((int)0x10FFFF)

#define UTF8TRAIL				(0x80)
#define UTF8FIRST				(0x80)
#define UTF8TRAILMASK			(0x3F)

#define GetTrail(_chr)			((_chr)&0x3F)
#define IsTrail(_chr)			(((_chr)&0xC0) == UTF8TRAIL)
#define IsUTF8har(_chr)			(((_chr)&UTF8FIRST) == UTF8FIRST)

#define UTF_UNREACHABLE	__assume(0)

#define UTF8_TRAIL			0x80
#define UTF8_TRAILMASK		0xC0
#define UTF8_TRAILVALMASK	0x3F

#define UTF8_LEAD(_bytecount) (0xFE << (6-(_bytecount)))
#define UTF8_LEADVALMASK(_ch, _bytecount) ((_ch) & (0xFF >> (2+(_bytecount))))

#define UTF8_PUTTRAILI(_ch, _buf) { *(unsigned char*)((_buf)--) = UTF8_TRAIL | ((_ch) & UTF8_TRAILVALMASK); (_ch) >>= 6; }
#define UTF8_PUTLEAD(_ch, _buf, _bytecount) { *(unsigned char*)((_buf)) = (UTF8_LEAD(_bytecount) | UTF8_LEADVALMASK(_ch, _bytecount)); }
#define UTF8_GETTRAIL(_ch, _str, _tmp, _ivalidlabel) { (_tmp) = *(_str); if (((_tmp) & UTF8_TRAILMASK) != UTF8_TRAIL) goto _ivalidlabel; (_ch) <<= 6; (_ch) |= (_tmp) & UTF8_TRAILVALMASK; (_str)++; }

//Enables validation of UTF8 sequence lengths, returning invalid codepoint for overlong sequences
#define UTF8_VALIDATELENGTH
//Enables validation of codepoints, rejecting codepoints in the surrogate ranges and other invalid codepoints
#define UTF8_VALIDATECODEPOINTS
//This character is returned by the decomposer if it encounters an invalid UTF8 sequence
#define UTF8_INVALIDCODEPOINT	0xFFFD
//This character is returned by the decomposer if it encounters an invalid UTF16 sequence
#define UTF16_INVALIDCODEPOINT	0xFFFD

wchar_t* UTF16Compose(int _chr, wchar_t* _dst, int _buflen);
int UTF16Decompose(const wchar_t* _str, int &_len, bool &_valid);
//int GetUTF8(const char*_str, int *_len, bool *_valid);
char* UTF8strchr(char *_str, int _chr);
const char* UTF8strchr(const char *_str, int _chr);
//char* BuildUTF8(char* _dst, int _chr);
int GetUTF8ByteLen(int _chr);

char* UTF8Compose(int _ch, char* _buf, int _buflen);
int UTF8Decompose(const char*_str, int &_len, bool &_valid);

inline int GetUTF8_i(char* &_str){
	int chr;
	bool valid;
	int len;
	chr = UTF8Decompose(_str, len, valid);
	_str += len;
	
	return valid ? chr : 0;
}

inline int GetUTF8_i(const char* &_str){
	int chr;
	bool valid;
	int len;
	chr = UTF8Decompose(_str, len, valid);
	_str += len;
	
	return valid ? chr : 0;
}

inline int GetUTF8l_i(char* &_str, int &_len){
	int chr;
	bool valid;
	chr = UTF8Decompose(_str, _len, valid);
	_str += _len;	
	return valid ? chr : 0;
}

inline int GetUTF8l_i(const char* &_str, int &_len){
	int chr;
	bool valid;
	chr = UTF8Decompose(_str, _len, valid);
	_str += _len;
	
	return valid ? chr : 0;
}



inline int PeekUTF8_i(const char* _str){
	int chr;
	bool valid;
	int len;
	chr = UTF8Decompose(_str, len, valid);	
	return valid ? chr : 0;
}

inline int PeekUTF8l_i(const char* _str, int *_len){
	int chr;
	bool valid;
	chr = UTF8Decompose(_str, *_len, valid);
	return valid ? chr : 0;
}

inline bool PutUTF8(char *_str, int _chr){
	char* _ptr;
	_ptr = UTF8Compose(_chr, _str, 6);
	return _str != _ptr;
}

inline int GetUTF16_i(wchar_t* &_str){
	int len, chr;
	bool valid;
	chr = UTF16Decompose(_str, len, valid);
	_str += len;
	return valid ? chr : 0;
}

inline int GetUTF16_i(const wchar_t* &_str){
	int len, chr;
	bool valid;
	chr = UTF16Decompose(_str, len, valid);
	_str += len;
	return valid ? chr : 0;
}

inline int GetUTF16l_i(wchar_t* &_str, int &_len){
	int chr;
	bool valid;
	chr = UTF16Decompose(_str, _len, valid);
	_str += _len;
	return valid ? chr : 0;
}

inline int GetUTF16l_i(const wchar_t* &_str, int &_len){
	int chr;
	bool valid;
	chr = UTF16Decompose(_str, _len, valid);
	_str += _len;
	return valid ? chr : 0;
}


inline int PeekUTF16_i(wchar_t* _str){
	int len, chr;
	bool valid;
	chr = UTF16Decompose(_str, len, valid);
	return valid ? chr : 0;
}

inline int PeekUTF16_i(const wchar_t* _str){
	int len, chr;
	bool valid;
	chr = UTF16Decompose(_str, len, valid);
	return valid ? chr : 0;
}

inline int PeekUTF16l_i(wchar_t* _str, int &_len){
	int chr;
	bool valid;
	chr = UTF16Decompose(_str, _len, valid);
	return valid ? chr : 0;
}

inline int PeekUTF16l_i(const wchar_t* _str, int &_len){
	int chr;
	bool valid;
	chr = UTF16Decompose(_str, _len, valid);
	return valid ? chr : 0;
}

inline bool UTF8IsLead(char c){
	unsigned char u = *(unsigned char*)&c;
	return ((c >= 194)-(c <= 244)) == 0;
}

inline const char* UTF8Prev(const char* pStart, const char* pCurrent){
	int ch;
	pCurrent--;
	while (pCurrent > pStart){
		ch = PeekUTF8_i(pCurrent);
		if (ch < 0x7F || UTF8IsLead(ch))
			return pCurrent;
		pCurrent--;
	}
	return NULL;
}

inline char* UTF8Prev(char* pStart, char* pCurrent){
	int ch;
	pCurrent--;
	while (pCurrent > pStart){
		ch = PeekUTF8_i(pCurrent);
		if (ch < 0x7F || UTF8IsLead(ch))
			return pCurrent;
		pCurrent--;
	}
	return NULL;
}

inline const char* NextUTF8_i(const char* &pStr){
	int len;
	bool valid;
	UTF8Decompose(pStr, len, valid);
	pStr += len;
	return pStr;
}

#endif //!_STR_H_
