#ifndef _T4_H_
#define _T4_H_

#include "StringBuffer.h"
#include "Lists.h"
#include "UTFStreams.h"

typedef bool (*fnT4ExecuteCallback)(CStringBufferA &stream, const char* pParam, void* ctx);

class CSimpleT4{
protected:
	class CExecuteContext {
	public:
		fnT4ExecuteCallback	m_pfnCallback;
		void*				m_pUserCtx;
	};
	class CT4Chunk {
	public:
		virtual bool Execute(CStringBufferA &stream, CExecuteContext&ctx){return false;}
	};
	class CT4TextChunk : public CT4Chunk{
	public:
		CStringBufferA	m_String;
		virtual bool Execute(CStringBufferA &stream, CExecuteContext&ctx){
			stream.Append(m_String);
			return true;
		}
	};
	class CT4ExecChunk : public CT4Chunk{
	public:
		CStringBufferA	m_Param;
		virtual bool Execute(CStringBufferA &stream, CExecuteContext&ctx){
			if (!ctx.m_pfnCallback)
				return true;
			//stream.Append(m_String);
			return ctx.m_pfnCallback(stream, m_Param, ctx.m_pUserCtx);
		}
	};
	SList<CT4Chunk>	m_Chunks;
public:
	inline CSimpleT4(){
	}
	inline ~CSimpleT4(){
		m_Chunks.Remove_All(SList<CT4Chunk>::DefaultCallback);
	}
	inline void Clear(){
		m_Chunks.Remove_All(SList<CT4Chunk>::DefaultCallback);
	}
	bool Parse(const char* pText){
		Clear();
		CUTF8Stream stream(pText);
		int ch, ch2, ch3, ch4;
		CT4ExecChunk* pExecChunk;
		CT4TextChunk* pTextChunk;
		pTextChunk = NULL;
		const char* pEnd;
		CStringBufferA buf;
		for(;;){
			ch = stream.Get();
			if (ch == 0)
				break;
			if (ch == '$'){
				ch2 = stream.Get();
				if (ch2 == '#'){
					CUTF8Stream stream2(stream.GetCurrent());
					for (;;){
						ch3 = stream2.Get();
						if (ch3 == 0)
							break;
						if (ch3 == '#'){
__check_next:
							pEnd = stream2.GetCurrent();
							ch4 = stream2.Get();
							if (ch4 == '#'){
								ch3 = ch4;
								goto __check_next;
							}
							if (ch4 == '$'){
								pTextChunk = NULL;
								pExecChunk = new CT4ExecChunk();
								if (!pExecChunk)
									return false;
								if (!m_Chunks.Add_Tail(pExecChunk))
									return delete pExecChunk, false;
								pExecChunk->m_Param.AppendN(stream.GetCurrent(), pEnd-stream.GetCurrent()-1);
								stream = stream2.GetCurrent();
								goto __found_exec;
							}
						}
					}
				} else if (ch2 == '$'){
					if (!pTextChunk){
						pTextChunk = new CT4TextChunk();
						if (!pTextChunk)
							return false;
						if (!m_Chunks.Add_Tail(pTextChunk))
							return delete pTextChunk, false;
					}
					pTextChunk->m_String.AppendChr('$');
					continue;
				}
				if (!pTextChunk){
					pTextChunk = new CT4TextChunk();
					if (!pTextChunk)
						return false;
					if (!m_Chunks.Add_Tail(pTextChunk))
						return delete pTextChunk, false;
				}
				pTextChunk->m_String.AppendChr('$');
				pTextChunk->m_String.AppendChr(ch2);
				continue;
			}
			if (!pTextChunk){
				pTextChunk = new CT4TextChunk();
				if (!pTextChunk)
					return false;
				if (!m_Chunks.Add_Tail(pTextChunk))
					return delete pTextChunk, false;
			}
			pTextChunk->m_String.AppendChr(ch);
__found_exec:
			;
		}
		return true;
	}
	bool Execute(CStringBufferA &stream, fnT4ExecuteCallback pCallback, void* pUserCtx = NULL){
		GenericSLNode<CT4Chunk>* pNode;
		CExecuteContext ctx;
		ctx.m_pfnCallback = pCallback;
		ctx.m_pUserCtx = pUserCtx;
		for (pNode = m_Chunks.Get_Head(); pNode; pNode = pNode->m_pNext){
			if (!pNode->m_pData->Execute(stream, ctx))
				return false;
		}
		return true;
	}
};

#endif //!_T4_H_
