#ifndef _UTF8STREAMS_H_
#define _UTF8STREAMS_H_
#include "Str.h"

class CUTF8Stream {
private:
	const char* m_pData;
	const char*	m_pInitial;
	int			m_LastPeek;
	int			m_LastLength;
	inline CUTF8Stream(){}
	inline CUTF8Stream(const CUTF8Stream &b){
		m_pData = b.m_pData;
		m_pInitial = b.m_pInitial;
		InternalFetch();
	}
	inline void InternalFetch(){
		int len;
		bool valid;
		m_LastPeek = UTF8Decompose(m_pData, m_LastLength, valid);
	}
	inline void InternalIncrementAndFetch(){
		if (m_LastPeek == 0)
			return;
		m_pData += m_LastLength;
		InternalFetch();
	}
public:
	inline CUTF8Stream(const char* pData) : m_pData(pData), m_pInitial(pData){
		InternalFetch();
	}
	inline CUTF8Stream &operator=(const char* pData){ m_pData = m_pInitial = pData; InternalFetch(); return *this;}
	inline int Peek(){
		return m_LastPeek;
	}
	inline int Get(){
		int ch;
		ch = m_LastPeek;
		return InternalIncrementAndFetch(), ch;
	}
	inline const char* GetCurrent(){ return m_pData; }
	inline void Reset(){ m_pData = m_pInitial; InternalFetch(); };
	inline void UpdateInitial(){ m_pInitial = m_pData; }
	inline CUTF8Stream &operator++(){
		InternalIncrementAndFetch();
		return *this;
	}
	inline CUTF8Stream &operator++(int){
		InternalIncrementAndFetch();
		return *this;
	}
};

class CUTF16Stream {
private:
	const wchar_t*	m_pData;
	const wchar_t*	m_pInitial;
	int			m_LastPeek;
	int			m_LastLength;
	inline CUTF16Stream(){}
	inline CUTF16Stream(const CUTF16Stream &b){
		m_pData = b.m_pData;
		m_pInitial = b.m_pInitial;
		InternalFetch();
	}
	inline void InternalFetch(){
		int len;
		bool valid;
		m_LastPeek = UTF16Decompose(m_pData, m_LastLength, valid);
	}
	inline void InternalIncrementAndFetch(){
		if (m_LastPeek == 0)
			return;
		m_pData += m_LastLength;
		InternalFetch();
	}
public:
	inline CUTF16Stream(const wchar_t* pData) : m_pData(pData), m_pInitial(pData){
		InternalFetch();
	}
	inline CUTF16Stream &operator=(const wchar_t* pData){ m_pData = m_pInitial = pData; InternalFetch(); return *this;}
	inline int Peek(){
		return m_LastPeek;
	}
	inline int Get(){
		int ch;
		ch = m_LastPeek;
		return InternalIncrementAndFetch(), ch;
	}
	inline const wchar_t* GetCurrent(){ return m_pData; }
	inline void Reset(){ m_pData = m_pInitial; InternalFetch(); };
	inline void UpdateInitial(){ m_pInitial = m_pData; }
	inline CUTF16Stream &operator++(){
		InternalIncrementAndFetch();
		return *this;
	}
	inline CUTF16Stream &operator++(int){
		InternalIncrementAndFetch();
		return *this;
	}
};

#endif //!_UTF8STREAMS_H_
