#ifndef _VECTORANDLISTTEMPLATECLASSES_H_
#define _VECTORANDLISTTEMPLATECLASSES_H_
#include <stdio.h>
#include <windows.h>
#include <stdlib.h>
//#include "BitVectors.h"

template<typename T> class VectorClass {
public:
	T *m_pVector;
	int m_nItems;
    bool m_Safe_For_Thread_Access;
	bool m_Is_Locally_Allocated;


	VectorClass(int nItems, T const *pItems){
		m_nItems = nItems;
		m_Is_Locally_Allocated = false;
		m_Safe_For_Thread_Access = true;
		if (nItems){
			if (pItems)
				m_pVector = (T *)pItems;
			else
				m_pVector = new T[nItems]();
		}
	}
	virtual ~VectorClass(){
		this->Clear();
	}
	virtual bool operator==(VectorClass<T> const &comp){
		if (m_nItems != comp.m_nItems || (!m_pVector || !comp.m_pVector)) return false;
		if (m_nItems < 1 ) return true;
		for (int i = 0; i < m_nItems; i++){
			if (m_pVector[i] != comp.m_pVector[i])
				return false;
		}
		return true;
	}
	virtual bool Resize(int nItems, T const *pItems){
		T *v_pVector = NULL;
		register int v_nItems;
		if (!nItems){
			Clear();
			return true;
		} else {
			m_Safe_For_Thread_Access = false;
			if (!pItems)
				v_pVector = new T[nItems]();
			else
				v_pVector = (T*)pItems;
			m_Safe_For_Thread_Access = true;
			if (!v_pVector)
				return false;
			if (m_pVector){
				v_nItems = (m_nItems <= nItems ? m_nItems : nItems);
				if (v_nItems > 1){
					for (int i = 0; i < v_nItems; i++)
						v_pVector[i] = m_pVector[i];
					if (m_Is_Locally_Allocated && m_pVector)
						delete[]m_pVector;
				}
			}
			m_pVector = v_pVector;
			m_nItems = nItems;
			if (pItems)
				m_Is_Locally_Allocated = false;
			else
				m_Is_Locally_Allocated = true;
			return true;

		}
	}
	virtual void Clear(){
		if (m_pVector && m_Is_Locally_Allocated){
			delete[] m_pVector;
			m_pVector = NULL;
		}
		m_Is_Locally_Allocated = NULL;
		m_nItems = NULL;
	}
	virtual int ID(T const &instance){
		if (!m_Safe_For_Thread_Access) return 0;
		if (m_nItems < 1){
			return -1;
		}
		for (int i = 0; i < m_nItems; i++){
			if (memcmp(&m_pVector[i], &instance, sizeof(T)) == 0)
				return i;
		}
		return -1;
	}
	virtual int ID_(T const *ptr){
		if (!m_Safe_For_Thread_Access) return 0;
		if (m_nItems < 1){
			return -1;
		}
		return (int)((char*)ptr - (char*)m_pVector) >> 2;
	}
	inline T& operator[](int index){
		if (index < 0 || index >= m_nItems){
			//debug shit here
		}
		return m_pVector[index];
	}
};

template<typename T> class DynamicVectorClass : public VectorClass<T> {
public:
	int m_nItemCount;
    int m_nResizePadding;

	DynamicVectorClass(int nItems = 0, T const *pItems = 0 ) : VectorClass<T>(nItems, pItems){		
		m_nItemCount = 0;
		m_nResizePadding = 10;
	}
	virtual ~DynamicVectorClass(){
		this->Clear();
	}
	virtual bool operator==(VectorClass<T> const &comp){
		if (m_nItems != comp.m_nItems || (!m_pVector || !comp.m_pVector)) return false;
		if (m_nItems < 1 ) return true;
		for (int i = 0; i < m_nItems; i++){
			if (m_pVector[i] != comp.m_pVector[i])
				return false;
		}
		return true;
	}
	virtual bool Resize(int nItems, T const *pItems){
		if (__super::Resize(nItems, pItems)){
			if (m_nItemCount > m_nItems)
				m_nItemCount = m_nItems;
			return true;
		} else return false;
	}
	virtual void Clear(){
		m_nItemCount = 0;
		__super::Clear();
	}
	virtual int ID(T const &instance){
		if (!m_Safe_For_Thread_Access) return 0;
		if (m_nItemCount < 1){
			return -1;
		}
		for (int i = 0; i < m_nItemCount; i++){
			if (memcmp(&m_pVector[i], &instance, sizeof(T)) == 0)
				return i;
		}
		return -1;
	}
	virtual int ID_(T const *ptr){
		if (!m_Safe_For_Thread_Access) return 0;
		if (m_nItems < 1){
			return -1;
		}
		return (int)((char*)ptr - (char*)m_pVector) >> 2;
	}
	inline T& operator[](int index){
		if (index < 0 || index >= m_nItems){
			//debug shit here
		}
		return m_pVector[index];
	}
	bool Add(T const &instance){
		if (m_nItemCount >= m_nItems){
			if ((m_Is_Locally_Allocated == false && m_nItems > 0) || m_nResizePadding <= 0)
				return false;
			if (this->Resize(m_nItems + m_nResizePadding, NULL) == false)
				return false;
		}
		m_pVector[m_nItemCount++] = instance;
		return true;
	}
	bool Delete(int index){
		if (index < 0 || index >= m_nItemCount)
			return false;
		m_nItemCount--;
		for (int i = index; i < m_nItemCount; i++)		
			m_pVector[i] = m_pVector[i+1];
		return true;
	}
	bool Insert(int index, T const &instance){
		if (index == 0 || index >= m_nItemCount)
			return false;
		if (m_nItemCount >= m_nItems){
			if ((!m_Safe_For_Thread_Access && m_nItems != 0) || m_nResizePadding <= 0) return false;
			if (!this->Resize(m_nItems + m_nResizePadding, NULL)) return false;

		}
		if (index < m_nItemCount){
			memmove(&m_pVector[index+1], &m_pVector[index], (m_nItemCount-index)*sizeof(T));
		}
		m_pVector[index] = instance;
		m_nItemCount++;
		return true;
	}
};

template<typename T> class SimpleVecClass {
public:
	T *m_pVector;
	int m_nItems;

	SimpleVecClass(){
		m_nItems = 0;
		m_pVector = NULL;
	}
	virtual ~SimpleVecClass(){
		if (m_pVector)
			delete [] m_pVector;
	}
	virtual bool Resize(int nItems){
		T *v_pVector = NULL;
		register int v_nItems;
		if (nItems == m_nItems) return true;
		if (!nItems){
			if (m_pVector)
				delete[] m_pVector;
			m_pVector = NULL;
			m_nItems = 0;
			return true;
		} else {
			v_pVector = new T[nItems]();
			if (!v_pVector)
				return false;
			if (m_pVector){
				v_nItems = (m_nItems <= nItems ? m_nItems : nItems);
				if (v_nItems > 1){
					for (int i = 0; i < v_nItems; i++)
						v_pVector[i] = m_pVector[i];
					if (m_pVector)
						delete[]m_pVector;
				}
			}
			m_pVector = v_pVector;
			m_nItems = nItems;
			return true;
		}
	}
	virtual bool Uninitialised_Grow(int nItems){
		if (nItems <= m_nItems) return true;
		if (m_pVector)
			delete [] m_pVector;
		m_pVector = new T[nItems];
		m_nItems = nItems;
		return true;
	}
	inline T& operator[](int index){
		if (index < 0 || index >= m_nItems){
			//debug shit here
		}
		return m_pVector[index];
	}
};

template<typename T> class SimpleDynVecClass : public SimpleVecClass<T> {
public:
	int m_nItemCount;

	SimpleDynVecClass(){
		m_nItemCount = 0;
	}
	virtual ~SimpleDynVecClass(){
	}
	virtual bool Resize(int nItems){
		T *v_pVector = NULL;
		register int v_nItems;
		if (nItems != m_nItems){
			if (!nItems){
				if (m_pVector)
					delete[] m_pVector;
				m_pVector = NULL;
				m_nItems = 0;
			} else {
				v_pVector = new T[nItems]();
				if (m_pVector && v_pVector){
					v_nItems = (m_nItems <= nItems ? m_nItems : nItems);
					if (v_nItems > 1){
						for (int i = 0; i < v_nItems; i++)
							v_pVector[i] = m_pVector[i];
						if (m_pVector)
							delete[]m_pVector;
					}
				}
				m_pVector = v_pVector;
				m_nItems = nItems;
			}
		}
		if (m_nItems < m_nItemCount)
			m_nItemCount = m_nItems;
		return true;
	}
	virtual bool Uninitialised_Grow(int nItems){
		if (nItems <= m_nItems) return true;
		if (m_pVector)
			delete [] m_pVector;
		m_pVector = new T[nItems];
		m_nItems = nItems;
		return true;
	}
	inline T& operator[](int index){
		if (index < 0 || index >= m_nItems){
			//debug shit here
		}
		return m_pVector[index];
	}
};

template<typename T, int align = 32, int (_cdecl *SortProc)(void*ctx, T&a, T&b) = NULL> class TVector {
private:
	T*		m_pFirst;
	T*		m_pLast;
	size_t	m_Size;
	void Resize(size_t size, T** ppPtr = NULL, bool expand = false){
		T* pArray;
		if (size <= m_Size)
			return;
		T* pPtr;
		size = ((size+align-1)/align)*align;
		m_Valid.Resize(size);
		alloc_nt(&pArray, size);
		memcpy_s(pArray, size*sizeof(T), m_pFirst, (m_pLast-m_pFirst)*sizeof(T));
		if (m_pFirst)
			g_pFreeer(m_pFirst);
		if (ppPtr)
			*ppPtr = pArray+((*ppPtr)-m_pFirst);

		m_pLast = pArray+ (expand ? size : (m_pLast-m_pFirst));
		m_pFirst = pArray;
		if (m_pFirst && size > m_Size){
			memset_r(m_pLast, m_pFirst+size);
			for (pPtr = m_pLast; pPtr < m_pFirst+size; pPtr++)
				*pPtr = T();
		}
		m_Size = size;
	}
	
public:
	typedef int (_cdecl *pfnSortProc)(void*ctx, T&a, T&b);
private:
	struct VectorCtx {
		void*		ctx;
		pfnSortProc	proc;
	};
	static int InternalSort(void*ctx, const void*a, const void*b){
		VectorCtx *pctx;
		pctx = (VectorCtx*)ctx;
		return pctx->proc(pctx->ctx, *(T*)a, *(T*)b);
	}
public:
	static T	StdValue;
	template<typename T> class _Iterator {
	private:
		T*	m_pPtr;
	public:
		_Iterator(){
			m_pPtr = NULL;
		}
		_Iterator(T* ptr){
			m_pPtr = ptr;
		}
		_Iterator(const _Iterator<T> &b){
			m_pPtr = b.m_pPtr;
		}
		_Iterator<T> &operator=(const _Iterator<T> &b){
			m_pPtr = b.m_pPtr;
			return *this;
		}
		_Iterator<T> &operator=(T* b){
			return m_pPtr=b, *this;
		}
		/*T* operator(){
			return m_pPtr;
		}*/
		T &operator *(){
			return *m_pPtr;
		}
		T* operator->(){
			return m_pPtr;
		}
		_Iterator<T> &operator++(){
			m_pPtr++;
			return *this;
		}
		_Iterator<T> &operator--(){
			m_pPtr++;
			return *this;
		}
		_Iterator<T> operator++(int){
			_Iterator<T> tmp;
			tmp = *this;
			++*this;
			return tmp;
		}
		_Iterator<T> operator--(int){
			_Iterator<T> tmp;
			tmp = *this;
			--*this;
			return tmp;
		}
		bool operator==(const _Iterator<T> &b){
			return (m_pPtr == b.m_pPtr);
		}
		bool operator!=(const _Iterator<T> &b){
			return (m_pPtr != b.m_pPtr);
		}
	};
	typedef _Iterator<T> Iterator;
	INLINE Iterator GetIterator(){
		return Iterator(m_pFirst);
	}
	INLINE Iterator Begin(){
		return m_pFirst;
	}
	INLINE Iterator End(){
		return m_pLast;
	}
	INLINE T* First(){
		return m_pFirst;
	}
	INLINE T* Last(){
		return m_pLast;
	}
	/*TVector<T, align> &operator=(const TVector<T, align> &b){
		m_pFirst = b.m_pFirst;
		m_pLast = b.m_pLast;
		m_Size = b.m_Size;
	}*/
	TVector(){
		m_pFirst = m_pLast = NULL;
		m_Size = 0;
	}
	TVector(const TVector<T, align> &b){
		Copy(b);
	}
	~TVector(){
		Clear();
	}
	void Clear(){
		T*p;
		if (m_pFirst != m_pLast){
			for (p = m_pFirst; p < m_pLast; p++)
				p->~T();
		}
		if (m_pFirst)
			g_pFreeer(m_pFirst);
		m_pFirst = m_pLast = NULL;
		m_Size = 0;
	}
	INLINE size_t Count(){
		return ((INT_PTR)(m_pLast-m_pFirst));
	}
	INLINE T& Add(const T&item){
		Resize(m_pLast-m_pFirst+1);
		*m_pLast = item;
		return *(m_pLast++);
	}
	INLINE T& Insert(const T&item, size_t _where){
		return Insert(item, m_pFirst+_where);
	}
	T& Insert(const T&item, T* _where){
		if (_where == m_pLast)
			return Add(item);
		Resize(m_pLast-m_pFirst+1, &_where);
		memmove_r(_where+1, _where, m_pLast);
		*_where = item;
		return ++m_pLast, *_where;
	}
	INLINE T&Get(size_t index){
		return *(m_pFirst+index);
	}
	INLINE T&operator[](size_t index){
		return *GetSafeNew(index);
	}
	INLINE void Sort(pfnSortProc proc, void *ctx = NULL){
		VectorCtx vctx;
		vctx.ctx = ctx;
		vctx.proc = proc;
		qsort_s(m_pFirst, m_pLast-m_pFirst, sizeof(T), InternalSort, &vctx);
	}
	INLINE bool Sort(void* ctx = NULL){
		if (!SortProc)
			return false;
		Sort(SortProc, ctx);
		return true;
	}
	INLINE bool IsValid(T*p){
		//Could add align check
		return (p >= m_pFirst && p < m_pLast);
	}
	bool IsValidClamped(T*&p){
		if (p < m_pFirst)
			return p = m_pFirst, false;
		if (p > m_pLast)
			return p = m_pLast, false;
		return true;
	}
	T* FindAdjacent(T*first = NULL, T*last = NULL){
		T* next;
		if (!first)
			first = m_pFirst;
		if (!last)
			last = m_pLast;
		if (first != last){
			next = first+1;
			for (;next != last;){
				if (*(first++) == *(next++))
					return first;
			}
		}
		return last;
	}
	INLINE T* GetSafe(int index){
		return (index < 0 || index >= (int)Count() || !m_Valid.Get(index)) ? NULL : (m_pFirst + index);
	}
	INLINE T* GetSafeNew(int index){
		if (index < 0)
			return NULL;
		if (index < (int)Count()){
			m_Valid.Set(index);
			return m_pFirst+index;
		}
		if (m_Valid.Get(index))
			(m_pFirst+index)->~T();
		Resize(index+1, NULL, true);
		m_Valid.Set(index);
		*(m_pFirst+index) = T();
		return m_pFirst+index;
	}
	INLINE int GetIndex(const T*p){
		if (p >= m_pFirst && p < m_pLast)
			return p-m_pFirst;
		return -1;
	}
	INLINE T* Erase(int index){
		return Erase(m_pFirst+index);
	}
	T* Erase(T*p){
		if (!IsValid(p))
			return m_pLast;
		p->~T();
		memmove_r(p, p+1, m_pLast);
		m_pLast--;
		return p;
	}
	T* Erase(T*first, T*last){
		T* temp;
		if (last != m_pLast){
			for (temp = first; temp < last; temp++)
				temp->~T();
			memmove_r(first, last, m_pLast);
			m_pLast = m_pLast-last+first;
		} else
			m_pLast = first;
		return first;
	}
	INLINE void Swap(TVector<T, align> &b){
		TVector<T, align> temp;
		temp.DirectCopy(*this);
		DirectCopy(b);
		b.DirectCopy(temp);
	}
	Iterator CopyUnique(Iterator first, Iterator last, Iterator result){
		T value;
		value = *first;
		for (;first != last; ++first){
			if (value != *first)
				*(++result) = value = *first;
		}
		return ++result;
	}
	void INLINE Reserve(size_t size){
		Resize(size);
	}
	void INLINE Copy(const TVector<T, align> &b){
		Reserve(b.m_Size);
		memcpy_r(m_pFirst, b.m_pFirst, b.m_pLast);
	}
	void INLINE Copy(const TVector<T, align>*b){
		Copy(*b);
	}
	void INLINE DirectCopy(const TVector<T, align> &b){
		m_pFirst = b.m_pFirst;
		m_pLast = b.m_pLast;
		m_Size = b.m_Size;
	}
	INLINE TVector<T, align> &operator=(const TVector<T, align> &b){
		Copy(b);
		return *this;
	}
	INLINE TVector<T, align> &operator+=(const TVector<T, align> &b){
		size_t i;
		Reserve(i = (Count()+b.Count()));
		if (!i)
			return *this;
		memcpy_r(m_pLast, b.m_pFirst, b.m_pLast);
		m_pLast += (b.m_pLast-b.m_pFirst);
		return *this;
	}
	INLINE TVector<T, align> &operator+=(const T &b){
		Add(b);
		return *this;
	}

public:
	static void Swap(TVector<T, align> &a, TVector<T, align> &b){
		a.Swap(b);
	}		
};

template<typename T, int align, int (_cdecl *SortProc)(void*ctx, T&a, T&b)> T TVector<T, align, SortProc>::StdValue;


template<typename T> class TSafeVectorAllocator {
public:
	void __cdecl Allocate(T(**ppObject)){
	}
	void __cdecl Destroy(T (**ppObject)){
	}
};

template<typename T> class TSimpleSafeVectorAllocator : public TSafeVectorAllocator<T> {
public:
	void __cdecl Allocate(T(**ppObject)){
		if (*ppObject)
			Destroy(ppObject);
		*ppObject = new T();
	}
	void __cdecl Destroy(T (**ppObject)){
		if (*ppObject)
			delete *ppObject;
		*ppObject = NULL;
	}
};

template<typename T, int _align = 256> class TPooledSafeVectorAllocator : public TSafeVectorAllocator<T> {
private:
	ObjectPoolClass<T, _align> Pool;
public:
	void __cdecl Allocate(T(**ppObject)){
		if (*ppObject)
			Destroy(ppObject);
		*ppObject = new (Pool.Allocate_Object_Memory())T();
	}
	void __cdecl Destroy(T (**ppObject)){
		if (*ppObject){
			(*ppObject)->~T();
			Pool.Free_Object_Memory(*ppObject);
		}
		*ppObject = NULL;
	}
};

//Allocated items make their pointers valid even when the vector has changed in size

template<typename T, int align = 32, typename Allocator = TSimpleSafeVectorAllocator, int (_cdecl *SortProc)(void*ctx, T&a, T&b) = NULL> class TSafeVector {
private:
	T**		m_pFirst;
	T**		m_pLast;
	size_t	m_Size;
	static Allocator	s_Allocator;
	void Resize(size_t size, T*** pppPtr = NULL, bool expand = false){
		T** pArray;
		size_t prevsize, temp;
		if (size <= m_Size)
			return;
		//T** ppPtr;
		temp = size;
		size = ((size+align-1)/align)*align;
		alloc_nt(&pArray, size);
		memcpy_s(pArray, size*sizeof(T*), m_pFirst, (m_pLast-m_pFirst)*sizeof(T*));
		if (m_pFirst)
			g_pFreeer(m_pFirst);
		if (pppPtr)
			*pppPtr = pArray+((*pppPtr)-m_pFirst);
		prevsize = m_pLast-m_pFirst;

		m_pLast = pArray+ (expand ? temp : (m_pLast-m_pFirst));
		m_pFirst = pArray;
		if (m_pFirst && size > m_Size){
			memset_r(m_pFirst+prevsize, m_pFirst+size);
		}
		m_Size = size;
	}
	
public:
	typedef int (_cdecl *pfnSortProc)(void*ctx, T&a, T&b);
private:
	struct VectorCtx {
		void*		ctx;
		pfnSortProc	proc;
	};
	static int InternalSort(void*ctx, const void*a, const void*b){
		VectorCtx *pctx;
		pctx = (VectorCtx*)ctx;
		return pctx->proc(pctx->ctx, *(T*)a, *(T*)b);
	}
public:
	template<typename T> class _Iterator {
	private:
		T*	m_pPtr;
	public:
		_Iterator(){
			m_pPtr = NULL;
		}
		_Iterator(T* ptr){
			m_pPtr = ptr;
		}
		_Iterator(const _Iterator<T> &b){
			m_pPtr = b.m_pPtr;
		}
		_Iterator<T> &operator=(const _Iterator<T> &b){
			m_pPtr = b.m_pPtr;
			return *this;
		}
		_Iterator<T> &operator=(T* b){
			return m_pPtr=b, *this;
		}
		/*T* operator(){
			return m_pPtr;
		}*/
		T &operator *(){
			return *m_pPtr;
		}
		T* operator->(){
			return m_pPtr;
		}
		_Iterator<T> &operator++(){
			m_pPtr++;
			return *this;
		}
		_Iterator<T> &operator--(){
			m_pPtr++;
			return *this;
		}
		_Iterator<T> operator++(int){
			_Iterator<T> tmp;
			tmp = *this;
			++*this;
			return tmp;
		}
		_Iterator<T> operator--(int){
			_Iterator<T> tmp;
			tmp = *this;
			--*this;
			return tmp;
		}
		bool operator==(const _Iterator<T> &b){
			return (m_pPtr == b.m_pPtr);
		}
		bool operator!=(const _Iterator<T> &b){
			return (m_pPtr != b.m_pPtr);
		}
	};
	typedef _Iterator<T> Iterator;
	INLINE Iterator GetIterator(){
		return Iterator(m_pFirst);
	}
	INLINE Iterator Begin(){
		return m_pFirst;
	}
	INLINE Iterator End(){
		return m_pLast;
	}
	INLINE T** First(){
		return m_pFirst;
	}
	INLINE T** Last(){
		return m_pLast;
	}
	/*TVector<T, align> &operator=(const TVector<T, align> &b){
		m_pFirst = b.m_pFirst;
		m_pLast = b.m_pLast;
		m_Size = b.m_Size;
	}*/
	TSafeVector(){
		m_pFirst = m_pLast = NULL;
		m_Size = 0;
	}
	/*TSafeVector(const TSafeVector<T, align> &b){
		Copy(b);
	}*/
	~TSafeVector(){
		Clear();
	}
	void Clear(){
		T**p;
		if (m_pFirst != m_pLast){
			for (p = m_pFirst; p < m_pLast; p++)
				s_Allocator.Destroy(p);
		}
		if (m_pFirst)
			g_pFreeer(m_pFirst);
		m_pFirst = m_pLast = NULL;
		m_Size = 0;
	}
	INLINE size_t Count(){
		return ((INT_PTR)(m_pLast-m_pFirst));
	}
	INLINE T* Add(const T&item){
		Resize(m_pLast-m_pFirst+1);
		s_Allocator.Allocate(m_pLast);
		**m_pLast = item;

		return *(m_pLast++);
	}
	INLINE T* Insert(const T&item, size_t _where){
		return Insert(item, m_pFirst+_where);
	}
	T* Insert(const T&item, T* _where){
		if (_where == m_pLast)
			return Add(item);
		Resize(m_pLast-m_pFirst+1, &_where);
		memmove_r(_where+1, _where, m_pLast);
		s_Allocator.Allocate(_where);
		return ++m_pLast, *_where;
	}
	INLINE T*Get(size_t index){
		return *(m_pFirst+index);
	}
	INLINE T*operator[](size_t index){
		return GetSafeNew(index);
	}
	INLINE void Sort(pfnSortProc proc, void *ctx = NULL){
		VectorCtx vctx;
		vctx.ctx = ctx;
		vctx.proc = proc;
		qsort_s(m_pFirst, m_pLast-m_pFirst, sizeof(T), InternalSort, &vctx);
	}
	INLINE bool Sort(void* ctx = NULL){
		if (!SortProc)
			return false;
		Sort(SortProc, ctx);
		return true;
	}
	INLINE bool IsValid(T*p){
		//Could add align check
		return (p >= m_pFirst && p < m_pLast);
	}
	bool IsValidClamped(T*&p){
		if (p < m_pFirst)
			return p = m_pFirst, false;
		if (p > m_pLast)
			return p = m_pLast, false;
		return true;
	}
	T* FindAdjacent(T*first = NULL, T*last = NULL){
		T* next;
		if (!first)
			first = m_pFirst;
		if (!last)
			last = m_pLast;
		if (first != last){
			next = first+1;
			for (;next != last;){
				if (*(first++) == *(next++))
					return first;
			}
		}
		return last;
	}
	INLINE T* GetSafe(int index){
		return (index < 0 || index >= (int)Count()) ? NULL : m_pFirst[index];
	}
	INLINE T** GetSafePP(int index){
		return (index < 0 || index >= (int)Count() || !m_pFirst[index]) ? NULL : (m_pFirst + index);
	}
	INLINE T* GetSafeNew(int index){
		if (index < 0)
			return NULL;
		if (index < (int)Count()){
			if (!m_pFirst[index])
				s_Allocator.Allocate(m_pFirst+index);
			return *(m_pFirst+index);
		}
		Resize(index+1, NULL, true);
		s_Allocator.Allocate(m_pFirst+index);
		if (m_pFirst+index >= m_pLast)
			m_pLast = m_pFirst+index+1;
		return *(m_pFirst+index);
	}
	INLINE T** GetSafeNewPP(int index){
		if (index < 0)
			return NULL;
		if (index < (int)Count()){
			if (!m_pFirst[index])
				s_Allocator.Allocate(m_pFirst+index);
			return m_pFirst+index;
		}
		Resize(index+1, NULL, true);
		s_Allocator.Allocate(m_pFirst+index);
		if (m_pFirst+index >= m_pLast)
			m_pLast = m_pFirst+index+1;
		return m_pFirst+index;
	}
	INLINE int GetIndex(const T**p){
		if (p >= m_pFirst && p < m_pLast)
			return p-m_pFirst;
		return -1;
	}
	INLINE T** Erase(int index){
		return Erase(m_pFirst+index);
	}
	T** Erase(T**p){
		if (!IsValid(p))
			return m_pLast;
		s_Allocator.Destroy(p);
		memmove_r(p, p+1, m_pLast);
		m_pLast--;
		return p;
	}
	T** Erase(T**first, T**last){
		T* temp;
		if (last != m_pLast){
			for (temp = first; temp < last; temp++)
				temp->~T();
			memmove_r(first, last, m_pLast);
			m_pLast = m_pLast-last+first;
		} else
			m_pLast = first;
		return first;
	}
	INLINE void Swap(TVector<T, align> &b){
		TVector<T, align> temp;
		temp.DirectCopy(*this);
		DirectCopy(b);
		b.DirectCopy(temp);
	}
	Iterator CopyUnique(Iterator first, Iterator last, Iterator result){
		T value;
		value = *first;
		for (;first != last; ++first){
			if (value != *first)
				*(++result) = value = *first;
		}
		return ++result;
	}
	void INLINE Reserve(size_t size){
		Resize(size);
	}
	/*void INLINE Copy(const TVector<T, align> &b){
		Reserve(b.m_Size);
		memcpy_r(m_pFirst, b.m_pFirst, b.m_pLast);
	}
	void INLINE Copy(const TVector<T, align>*b){
		Copy(*b);
	}
	void INLINE DirectCopy(const TVector<T, align> &b){
		m_pFirst = b.m_pFirst;
		m_pLast = b.m_pLast;
		m_Size = b.m_Size;
	}
	INLINE TVector<T, align> &operator=(const TVector<T, align> &b){
		Copy(b);
		return *this;
	}
	INLINE TVector<T, align> &operator+=(const TVector<T, align> &b){
		size_t i;
		Reserve(i = (Count()+b.Count()));
		if (!i)
			return *this;
		memcpy_r(m_pLast, b.m_pFirst, b.m_pLast);
		m_pLast += (b.m_pLast-b.m_pFirst);
		return *this;
	}
	INLINE TVector<T, align> &operator+=(const T &b){
		Add(b);
		return *this;
	}*/

public:
	/*static void Swap(TVector<T, align> &a, TVector<T, align> &b){
		a.Swap(b);
	}*/		
};

template<typename T, int align, typename Allocator, int (_cdecl *SortProc)(void*ctx, T&a, T&b)> Allocator	TSafeVector<T, align, Allocator, SortProc>::s_Allocator;

template<typename _first, typename _second> class Pair {
public:
	_first	First;
	_second	Second;
	Pair(const _first first, const _second second){
		First = first;
		Second = second;
	}
	Pair(const Pair<_first, _second> &b){
		First = b.First;
		Second = b.Second;
	}
	bool operator=(const Pair<_first, _second> &b){
		First = b.First;
		Second = b.Second;
	}
};


template<typename _key, typename T> class CMap {
private:
	static const size_t INVALID_INDEX = (size_t)-1;
	template<typename _id, typename T> struct Entry {
		_key	Key;
		T		Val;
		bool operator==(const _key b){
			return (Key == b);
		}
		bool operator<(const _key b){
			return (Key < b);
		}
		bool operator>(const _key b){
			return (Key > b);
		}
		bool operator<=(const _key b){
			return (Key <= b);
		}
		bool operator>=(const _key b){
			return (Key >= b);
		}
		bool operator==(const Entry<_key, T> &b) {
			return true;//(ID == b.ID && Val == b.Val);
		}
		bool operator!=(const Entry<_key, T> &b){
			return !(*this == b);
		}
	};
	TVector<Entry<_key, T>>	m_Map;

	size_t GetSlot(const _key Key){
		size_t start, end, count, current;
		_key key;
		current = (end = ((count = (int)m_Map.Count())-1))/2;
		start = 0;

		if (count > 0){

			while (end > start+1){
				key = m_Map[current].Key;
				if (key == Key){
					return INVALID_INDEX;
				}

				if (Key > key){
					start = current;
					current += (end-current)/2;
				} else {
					end = current;
					current -= (current-start);
				}
			}
			if (m_Map[end].Key > Key){
				if (m_Map[start].Key > Key)
					current = start;
				else
					current = end;
			} else
				current = end+1;
		} else
			current = 0;
		return current;
	}
	size_t GetIndex(const _key Key){
		size_t start, end, count, current;
		_key key;
		current = (end = ((count = (int)m_Map.Count())-1))/2;
		start = 0;

		if (count > 0){
			while (end > start+1){
				key = m_Map[current].Key;
				if (key == Key){
					return current;
				}

				if (Key > key){
					start = current;
					current += (end-current)/2;
				} else {
					end = current;
					current -= (current-start);
				}
			}
		}
		return INVALID_INDEX;
	}
public:
	CMap(){
	}
	~CMap(){
		Clear();
	}
	T* Insert(const _key Key, const T &val){
		size_t i;
		Entry<_key, T> entry;

		i = GetSlot(Key);
		if (i == INVALID_INDEX)
			return NULL;

		entry.Key = Key;
		entry.Val = val;
		return &m_Map.Insert(entry, i).Val;
	}
	T* Find(const _key Key){
		size_t i;
		i = GetIndex(Key);
		return (i != INVALID_INDEX) ? &m_Map[i].Val : NULL;
	}
	void Remove(const _key Key){
		size_t i;
		i = GetIndex(key);
		if (i != INVALID_INDEX)
			m_Map.Erase(i);
	}
	void Clear(){
		m_Map.Erase(&*m_Map.Begin(), &*m_Map.End());
	}
	T*operator[](const _key Key){
		return Find(Key);
	}
};

#endif //_VECTORANDLISTTEMPLATECLASSES_H_