
#include <Windows.h>
#include "XML.h"
#include "Lists.h"
#include "StdLib.h"
#include "FileStream.h"
#include "String.h"
#include "StringBuffer.h"
#include "StringHeap.h"

//#define ADD_REF_ON_NAMED_QUERY

#ifdef MBCS_XML
//Cache type
#define IXML_CT	CStringCacheA
#define IXML_BT	CStringBufferA
#else //MBCS_XML
#define IXML_CT	CStringCacheW
#define IXML_BT	CStringBuffer
#endif //!MBCS_XML

class CXML;

class CXMLNSStack {
private:
	class CXMLNS : public AutoPoolClass<CXMLNS, 32> {
	public:		
		IXML_CT::CRefStr*	m_pNameSpace;
		IXML_CT::CRefStr*	m_pURN;
		CXMLNS(){
			m_pNameSpace = IXML_CT::Get();
			m_pURN = IXML_CT::Get();
		}
		~CXMLNS(){
			IXML_CT::Return(m_pNameSpace);
			IXML_CT::Return(m_pURN);
		}
		inline const IXML_SC &NS(){
			return *m_pNameSpace;
		};
		inline const IXML_SC &URN(){
			return *m_pURN;
		}
	};
	CXMLNS			m_DefaultNS;
	SList<CXMLNS>	m_NameSpaces;
	SList<CXMLNS>	m_Stack;
	CXMLNS* Find(const IXML_ST* URN);
public:
	CXMLNSStack();
	~CXMLNSStack();
	inline int GetStackSize();
	inline int GetNumNameSpaces();
	bool Push(const IXML_ST* pURN);
	void Pop();
	bool PushDef(const IXML_ST* pPrefix, const IXML_ST* pURN);
	bool PushDef(const IXML_STI* pPrefix, const IXML_ST* pURN);
	void PopDef();
	void PopDefs(int count);
	const IXML_ST* GetCurrentPrefix();
};

CXMLNSStack::CXMLNS* CXMLNSStack::Find(const IXML_ST* URN){
	GenericSLNode<CXMLNS>* node;
	for (node = m_NameSpaces.Get_Head(); node; node = node->m_pNext){
		if (node->m_pData->URN() == URN)
			return node->m_pData;
	}
	return NULL;
}

CXMLNSStack::CXMLNSStack(){
}

CXMLNSStack::~CXMLNSStack(){
	m_NameSpaces.Remove_All(SList<CXMLNSStack::CXMLNS>::DefaultCallback, NULL);
	m_Stack.Remove_All();
}

int CXMLNSStack::GetStackSize(){
	return m_Stack.Get_Count();
}

int CXMLNSStack::GetNumNameSpaces(){
	return m_NameSpaces.Get_Count();
}

bool CXMLNSStack::Push(const IXML_ST* pURN){
	CXMLNS* ns;
	if (!*pURN)
		return m_Stack.Add_Head(&m_DefaultNS);
	ns = Find(pURN);
	if (ns){
		m_Stack.Add_Head(ns);
		return true;
	}
	return false;
}

void CXMLNSStack::Pop(){
	m_Stack.Remove_Head();
}

bool CXMLNSStack::PushDef(const IXML_ST* pPrefix, const IXML_ST* pURN){
	CXMLNS* ns;
	ns = new CXMLNS();
	if (!ns) return false;
	ns->m_pNameSpace->Set(pPrefix);
	ns->m_pURN->Set(pURN);
	if (!m_NameSpaces.Add_Head(ns))
		return (delete ns), false;
	return true;
}

bool CXMLNSStack::PushDef(const IXML_STI* pPrefix, const IXML_ST* pURN){
	CXMLNS* ns;
	IXML_SC prefix;
	prefix = pPrefix;
	ns = new CXMLNS();
	if (!ns) return false;
	ns->m_pNameSpace->Set(prefix);
	ns->m_pURN->Set(pURN);
	if (!m_NameSpaces.Add_Head(ns))
		return (delete ns), false;
	return true;
}

void CXMLNSStack::PopDef(){
	m_NameSpaces.Remove_Head();
}

void CXMLNSStack::PopDefs(int count){
	for (;count > 0; count--)
		PopDef();
}

const IXML_ST* CXMLNSStack::GetCurrentPrefix(){
	return m_Stack.Get_Head() ? m_Stack.Get_Head()->m_pData->NS() : m_DefaultNS.NS();
}

class CXMLResultSet : public IXMLResultSet {
public:
	ULONG		m_RefCount;
	IXMLData**	m_ppResults;
	int			m_ResultCount;
	CXML*		m_pXML;
	virtual ULONG AddRef();
	virtual ULONG Release();
	virtual int GetResultCount();
	virtual IXMLData* GetNode(int i);
	virtual void Enum(pfnEnumCallback cb, void* ctx = NULL);

	CXMLResultSet();
	~CXMLResultSet();
	void ClearData();

#ifdef _DEBUG
	static ULONG	s_Count;
#endif //_DEBUG

};

class CXMLData : public IXMLData {
public:
	ULONG				m_RefCount;
	SList2<CXMLData>	m_Children;
	SList2<CXMLData>	m_Attributes;
	CXMLData*			m_pNext;
	IXML_HANDLE			m_Name;
	IXML_HANDLE			m_Value;
	IXML_HANDLE			m_Namespace;
	IXML_HANDLE			m_URI;
	IXML_HANDLE			m_FullName;
	DOMNodeType			m_Type;
	CXML*				m_pXML;
	CXMLData*			m_pParent;
	int					m_StackOffset;
	bool				m_GotNS;
	

	virtual ULONG AddRef();
	virtual ULONG Release();
	virtual const IXML_ST*GetName();
	virtual const IXML_ST*GetValue();
	virtual const IXML_ST*GetNamespace();
	virtual const IXML_ST*GetURI();
	virtual const IXML_ST*GetFullName();
	virtual DOMNodeType GetType();
	virtual IXMLData* GetNext();
	virtual IXMLData* GetFirstChild();
	virtual IXMLData* GetFirstAttribute();
	virtual IXMLResultSet* FindChildren(const IXML_ST* pName);
	virtual IXMLResultSet* FindAttributes(const IXML_ST* pName);
	virtual IXMLData* GetNamedAttribute(const IXML_ST* pName, IXMLData* pPrevious);
	virtual IXMLData* GetNamedNode(const IXML_ST* pName, IXMLData* pPrevious);
	virtual IXMLResultSet* FindFiltered(pfnFilterProc pfnFilter, void *ctx = NULL);
	virtual IXMLResultSet* FindFilteredAttributes(pfnFilterProc pfnFilter, void *ctx = NULL);

	CXMLData(CXML *pXML);
	~CXMLData();
	void FreeData(bool SFree = false);
	bool IsValidChild(IXMLData* pData);
	bool IsValidAttribute(IXMLData* pData);
	bool LoadFromNode(IXMLDOMNode* pNode, CXMLData* pParent, CXMLNSStack *pStack);
	void RecheckXMLNS(CXMLNSStack* pStack);
	void SaveStack(CXMLNSStack *pStack);
	void RestoreStack(CXMLNSStack *pStack);
#ifdef _DEBUG
	static ULONG	s_Count;
#endif //_DEBUG

};

class CXML : public IXML {
public:
	ULONG			m_RefCount;
	CXMLData		m_Root;
	IXML_BT			m_Error;
	IXML_HEAP		m_Heap;
	IXML_BT			m_Buffer;

	virtual ULONG AddRef();
	virtual ULONG Release();
	virtual bool LoadFromFile(const IXML_ST* pPath);
	virtual bool LoadFromStream(IStream* pStream);
	virtual void Close();
	virtual IXMLData* GetRoot() const;
	virtual const IXML_ST* GetXMLError() const;
	CXML();
	~CXML();
	void FreeData();
	bool Initialize(IXMLDOMDocument* pDoc);
	bool InitializeFromNode(IXMLDOMNode* pNode);

#ifdef _DEBUG
	static ULONG	s_Count;
#endif //_DEBUG
};

/*
CXMLResult
*/

ULONG CXMLResultSet::AddRef(){
	m_pXML->AddRef();
	return ++m_RefCount;
}

ULONG CXMLResultSet::Release(){
	if (m_pXML)
		m_pXML->Release();
	if (--m_RefCount){
		return m_RefCount;
	}
	delete this;
	return 0;
}

int CXMLResultSet::GetResultCount() {
	return m_ResultCount;
}

IXMLData* CXMLResultSet::GetNode(int i) {
	return (i >= 0 && i < m_ResultCount) ? m_ppResults[i] : NULL;
}

void CXMLResultSet::Enum(pfnEnumCallback cb, void* ctx){
	int i;
	if (!cb)
		return;
	for (i = 0; i < m_ResultCount && cb(m_ppResults[i], ctx); i++)
		;
}


CXMLResultSet::CXMLResultSet(){
	m_ppResults = NULL;
	m_ResultCount = 0;
	m_RefCount = 1;
#ifdef _DEBUG
	s_Count++;
#endif //_DEBUG
}

CXMLResultSet::~CXMLResultSet(){
	ClearData();
#ifdef _DEBUG
	s_Count--;
#endif //_DEBUG
}

void CXMLResultSet::ClearData(){
	if (m_ppResults){
		/*for (i = 0; i < m_ResultCount; i++)
			m_ppResults[i]->Release();*/
		g_pFreeer(m_ppResults);
		m_ppResults = NULL;
	}
	m_ResultCount = 0;
}

#ifdef _DEBUG
ULONG	CXMLResultSet::s_Count = 0;
#endif //_DEBUG

/*
CXMLData
*/

ULONG CXMLData::AddRef(){
	if (m_pParent)
		m_pParent->AddRef();
	if (m_pXML)
		m_pXML->AddRef();
	return ++m_RefCount;
}

ULONG CXMLData::Release(){	
	ULONG ref;
	ref = --m_RefCount;
	if (m_pParent)
		m_pParent->Release();
	if (m_pXML)
		m_pXML->Release();
	return ref;
}

const IXML_ST*CXMLData::GetName(){
	return m_Name;
}

const IXML_ST*CXMLData::GetValue(){
	return m_Value;
}

const IXML_ST*CXMLData::GetNamespace(){
	return m_Namespace;
}

const IXML_ST*CXMLData::GetURI() {
	return m_URI;
}

const IXML_ST*CXMLData::GetFullName(){
	return m_FullName;
}

DOMNodeType CXMLData::GetType(){
	return m_Type;
}

IXMLData* CXMLData::GetNext(){
	return m_pNext;
}

IXMLData* CXMLData::GetFirstChild(){
	return m_Children.Get_Head();
}

IXMLData* CXMLData::GetFirstAttribute(){
	return m_Attributes.Get_Head();
}

IXMLResultSet* CXMLData::FindChildren(const IXML_ST* pName){
	SList<CXMLData> list;
	CXMLResultSet *pResultSet;
	CXMLData* pNode;
	int i;
	GenericSLNode<CXMLData>* pNode2;
	if (!m_Children.Get_Head())
		return NULL;

	pNode = m_Children.Get_Head();

	while (pNode){
		if (pNode->m_FullName == pName)
			list.Add_Tail(pNode);
		pNode = pNode->m_pNext;
	}
	if (!list.Get_Head())
		return NULL;

	pResultSet = new CXMLResultSet;
	if (!pResultSet)
		return NULL;
	pResultSet->m_ResultCount = list.Get_Count();
	pResultSet->m_ppResults = (IXMLData**)g_pAllocator(sizeof(IXMLData*)*pResultSet->m_ResultCount);
	if (!pResultSet->m_ppResults){
		delete pResultSet;
		return NULL;
	}
	m_pXML->AddRef();
	pResultSet->m_pXML = m_pXML;

	memset(pResultSet->m_ppResults, 0, sizeof(IXMLData*)*pResultSet->m_ResultCount);
	i = 0;
	pNode2 = list.Get_Head();
	while (pNode2){
		pResultSet->m_ppResults[i++] = pNode2->m_pData;
		pNode2 = pNode2->m_pNext;
	}

	return pResultSet;
}

IXMLResultSet* CXMLData::FindAttributes(const IXML_ST* pName){
	SList<CXMLData> list;
	CXMLResultSet *pResultSet;
	CXMLData* pNode;
	int i;
	GenericSLNode<CXMLData>* pNode2;
	if (!m_Attributes.Get_Head())
		return NULL;

	pNode = m_Attributes.Get_Head();

	while (pNode){
		if (pNode->m_FullName == pName)
			list.Add_Tail(pNode);
		pNode = pNode->m_pNext;
	}
	if (!list.Get_Head())
		return NULL;

	pResultSet = new CXMLResultSet;
	if (!pResultSet)
		return NULL;
	pResultSet->m_ResultCount = list.Get_Count();
	pResultSet->m_ppResults = (IXMLData**)g_pAllocator(sizeof(IXMLData*)*pResultSet->m_ResultCount);
	if (!pResultSet->m_ppResults){
		delete pResultSet;
		return NULL;
	}
	m_pXML->AddRef();
	pResultSet->m_pXML = m_pXML;

	memset(pResultSet->m_ppResults, 0, sizeof(IXMLData*)*pResultSet->m_ResultCount);
	i = 0;
	pNode2 = list.Get_Head();
	while (pNode2){
		pResultSet->m_ppResults[i++] = pNode2->m_pData;
		pNode2 = pNode2->m_pNext;
	}

	return pResultSet;
}

IXMLData* CXMLData::GetNamedAttribute(const IXML_ST* pName, IXMLData* pPrevious){
	CXMLData *pData;
	if (pPrevious && IsValidAttribute(pPrevious))
		pData = ((CXMLData*)pPrevious)->m_pNext;
	else
		pData = m_Attributes.Get_Head();
	while (pData){
		if (pData->m_Name == pName){
#ifdef ADD_REF_ON_NAMED_QUERY
			pData->AddRef();
#endif //ADD_REF_ON_NAMED_QUERY
			return pData;
		}
		pData = pData->m_pNext;
	}
	return NULL;
}

IXMLData* CXMLData::GetNamedNode(const IXML_ST* pName, IXMLData* pPrevious){
	CXMLData *pData;
	if (pPrevious && IsValidChild(pPrevious))
		pData = ((CXMLData*)pPrevious)->m_pNext;
	else
		pData = m_Children.Get_Head();
	while (pData){
		if (pData->m_FullName == pName){
#ifdef ADD_REF_ON_NAMED_QUERY
			pData->AddRef();
#endif //!ADD_REF_ON_NAMED_QUERY
			return pData;
		}
		pData = pData->m_pNext;
	}
	return NULL;
}

IXMLResultSet* CXMLData::FindFiltered(pfnFilterProc pfnFilter, void *ctx){
	SList<CXMLData> list;
	CXMLResultSet *pResultSet;
	CXMLData* pNode;
	int i;
	if (!pfnFilter) return NULL;
	GenericSLNode<CXMLData>* pNode2;
	if (!m_Children.Get_Head())
		return NULL;

	pNode = m_Children.Get_Head();

	while (pNode){
		if (pfnFilter(pNode, ctx))
			list.Add_Tail(pNode);
		pNode = pNode->m_pNext;
	}
	if (!list.Get_Head())
		return NULL;

	pResultSet = new CXMLResultSet;
	if (!pResultSet)
		return NULL;
	pResultSet->m_ResultCount = list.Get_Count();
	pResultSet->m_ppResults = (IXMLData**)g_pAllocator(sizeof(IXMLData*)*pResultSet->m_ResultCount);
	if (!pResultSet->m_ppResults){
		delete pResultSet;
		return NULL;
	}
	m_pXML->AddRef();
	pResultSet->m_pXML = m_pXML;

	memset(pResultSet->m_ppResults, 0, sizeof(IXMLData*)*pResultSet->m_ResultCount);
	i = 0;
	pNode2 = list.Get_Head();
	while (pNode2){
		pResultSet->m_ppResults[i++] = pNode2->m_pData;
		pNode2 = pNode2->m_pNext;
	}

	return pResultSet;
}

IXMLResultSet* CXMLData::FindFilteredAttributes(pfnFilterProc pfnFilter, void *ctx){
	SList<CXMLData> list;
	CXMLResultSet *pResultSet;
	CXMLData* pNode;
	int i;
	if (!pfnFilter) return NULL;
	GenericSLNode<CXMLData>* pNode2;
	if (!m_Children.Get_Head())
		return NULL;

	pNode = m_Attributes.Get_Head();

	while (pNode){
		if (pfnFilter(pNode, ctx))
			list.Add_Tail(pNode);
		pNode = pNode->m_pNext;
	}
	if (!list.Get_Head())
		return NULL;

	pResultSet = new CXMLResultSet;
	if (!pResultSet)
		return NULL;
	pResultSet->m_ResultCount = list.Get_Count();
	pResultSet->m_ppResults = (IXMLData**)g_pAllocator(sizeof(IXMLData*)*pResultSet->m_ResultCount);
	if (!pResultSet->m_ppResults){
		delete pResultSet;
		return NULL;
	}
	m_pXML->AddRef();
	pResultSet->m_pXML = m_pXML;

	memset(pResultSet->m_ppResults, 0, sizeof(IXMLData*)*pResultSet->m_ResultCount);
	i = 0;
	pNode2 = list.Get_Head();
	while (pNode2){
		pResultSet->m_ppResults[i++] = pNode2->m_pData;
		pNode2 = pNode2->m_pNext;
	}

	return pResultSet;
}
	
CXMLData::CXMLData(CXML *pXML) : m_Name(pXML->m_Heap), m_Value(pXML->m_Heap), m_Namespace(pXML->m_Heap), m_URI(pXML->m_Heap), m_FullName(pXML->m_Heap){
	m_pXML = pXML;
	m_pParent = NULL;
	m_pNext = NULL;
	m_RefCount = 1;
	m_StackOffset = -1;
	m_GotNS = false;
#ifdef _DEBUG
	s_Count++;
#endif //_DEBUG
}

CXMLData::~CXMLData(){
	/*if (m_pXML)
		m_pXML->Release();
	if (m_pParent)
		m_pParent->Release();*/
	FreeData();
#ifdef _DEBUG
	s_Count--;
#endif //_DEBUG
}

void CXMLData::FreeData(bool SFree){
	CXMLData* pCur;
	CXMLData* pNext;

	pCur = m_Children.Get_Head();
	m_Children.Remove_All();
	while (pCur){
		pNext = pCur->m_pNext;
		delete pCur;
		pCur = pNext;
	}

	pCur = m_Attributes.Get_Head();
	m_Attributes.Remove_All();
	while (pCur){
		pNext = pCur->m_pNext;
		delete pCur;
		pCur = pNext;
	}
	//SFree == selective free; deleting selected nodes, thus it's required to remove strings from the heap, CXML dtor clears entire heap
	if (SFree){
		m_Name.Clear();
		m_Value.Clear();
		m_Namespace.Clear();
		m_URI.Clear();
	}
}


bool CXMLData::IsValidChild(IXMLData* pData){
	CXMLData* pNode;
	pNode = m_Children.Get_Head();
	while (pNode){
		if (pData == pNode)
			return true;
		pNode = pNode->m_pNext;
	}
	return false;
}

bool CXMLData::IsValidAttribute(IXMLData* pData){
	CXMLData* pNode;
	pNode = m_Attributes.Get_Head();
	while (pNode){
		if (pData == pNode)
			return true;
		pNode = pNode->m_pNext;
	}
	return false;
}

bool CXMLData::LoadFromNode(IXMLDOMNode* pNode, CXMLData* pParent, CXMLNSStack* pStack){
	BSTR str;
	HRESULT hr;
	CXMLData* pAttribute;
	CXMLData* pNewNode;
	IXMLData* pTemp;
	IXMLDOMNode* pLocalNode;
	IXMLDOMNodeList* pList;
	IXMLDOMNamedNodeMap* pMap;
	DOMNodeType type;
	VARIANT v;

	m_pParent = pParent;

	FreeData();
	if (!pNode)
		return false;
	AREF2(pNode);
	hr = pNode->get_nodeType(&type);
	if (FAILED(hr))
		return false;

	m_Type = type;
	SaveStack(pStack);

	if (pNode->get_nodeName(&str) != S_OK)
		return RestoreStack(pStack), false;
	m_Name = IXML_CONVERT(str);
	SysFreeString(str);


	if ((type == NODE_TEXT || type == NODE_ATTRIBUTE) && pNode->get_nodeValue(&v) == S_OK && v.vt == VT_BSTR && v.bstrVal){
		m_Value = IXML_CONVERT(v.bstrVal);
		SysFreeString(v.bstrVal);
	}

	

	if (pNode->get_attributes(&pMap) == S_OK){
		AREF(pMap, false);

		while (pMap->nextNode(&pLocalNode) == S_OK){
			AREF(pLocalNode, false);
			pAttribute = new CXMLData(m_pXML);
			if (!pAttribute)
				continue;
			pAttribute->LoadFromNode(pLocalNode, this, pStack);
			m_Attributes.Add_Tail(pAttribute);
		}
	}

	
	if (m_Type == NODE_ATTRIBUTE){
		if (m_Name == IXML_STR("xmlns")){
			if (!pParent->m_GotNS){
				if (pStack->Push(m_Value))
					pParent->m_GotNS = true;
			}
		} else if (pNode->get_prefix(&str) == S_OK){
			if (wcscmp(str, L"xmlns") == 0){
				SysFreeString(str);
				if (pNode->get_baseName(&str) == S_OK){
					pStack->PushDef(str, m_Value);
					SysFreeString(str);
				}
			} else
				SysFreeString(str);
		}
	} else if (m_Type == NODE_ELEMENT){
		RecheckXMLNS(pStack);
		if (pNode->get_prefix(&str) == S_OK){
			m_FullName = m_Name;
			//Already got namespace prefix
			SysFreeString(str);
		} else {
			if (*(pStack->GetCurrentPrefix()) == 0)
				m_FullName = m_Name;
			else{
				m_pXML->m_Buffer.Print(IXML_STR("%s:%s"), pStack->GetCurrentPrefix(), m_Name());
				m_FullName = (const IXML_ST*)m_pXML->m_Buffer;
			}
			//LogFormated(L"Full name: %s\n", m_FullName());
		}
	} else
		m_FullName = m_Name;

	if (m_Type != NODE_ATTRIBUTE && pNode->get_childNodes(&pList) == S_OK){
		AREF(pList, false);
		/*if (pList->get_length(&count) == S_OK && count > 0){
			for (i = 0; i < count; i++){
				if (pList->get_item(i, &pLocalNode) != S_OK || !pLocalNode)
					continue;
				AREF(pLocalNode, false);
				pNewNode = new CXMLData();
				if (!pNewNode)
					continue;
				pNewNode->LoadFromNode(pLocalNode, pXML, this);
				m_Children.Add_Tail(pNewNode);
			}
		}*/
		while (pList->nextNode(&pLocalNode) == S_OK){
			AREF(pLocalNode, false);
			pNewNode = new CXMLData(m_pXML);
			if (!pNewNode)
				continue;
			pNewNode->LoadFromNode(pLocalNode, this, pStack);
			m_Children.Add_Tail(pNewNode);
		}
	}	
	
	if (!*(m_Value()) && (pTemp = GetNamedNode(IXML_STR("#text"), NULL))){
		m_Value = (pTemp->GetType() == NODE_TEXT && pTemp->GetValue()) ? pTemp->GetValue() : IXML_STR("");
	}
	
	/*if (!*(m_Value()) && (pTemp = GetNamedNode(L"#text", NULL))){
		m_Value = (pTemp->GetType() == NODE_TEXT && pTemp->GetValue()) ? pTemp->GetValue() : L"";
	}*/
	
	//LogFormated(L"Value: %s\n\n", m_Value);

	return RestoreStack(pStack), true;
}


void CXMLData::RecheckXMLNS(CXMLNSStack* pStack){
	IXMLData* pAttr;
	if (m_GotNS)
		return; //Already got an NS
	pAttr = GetNamedAttribute(IXML_STR("xmlns"), NULL);
	if (pAttr){
		if (pStack->Push(pAttr->GetValue()))
			m_GotNS = true;
#ifdef ADD_REF_ON_NAMED_QUERY
		pAttr->Release();
#endif //ADD_REF_ON_NAMED_QUERY
	}
}

void CXMLData::SaveStack(CXMLNSStack *pStack){
	if (m_Type == NODE_ATTRIBUTE) return;
	m_StackOffset = pStack->GetNumNameSpaces();
}

void CXMLData::RestoreStack(CXMLNSStack *pStack){
	if (m_Type == NODE_ATTRIBUTE) return;
	if (m_StackOffset == -1)
		return;
	pStack->PopDefs(pStack->GetNumNameSpaces()-m_StackOffset);
	if (m_GotNS)
		pStack->Pop();
	m_StackOffset = -1;
}

#ifdef _DEBUG
ULONG	CXMLData::s_Count = 0;
#endif //_DEBUG

/*
CXML
*/

ULONG CXML::AddRef(){
	return ++m_RefCount;
}

ULONG CXML::Release(){
	if (--m_RefCount)
		return m_RefCount;
	delete this;
	return 0;
}

bool CXML::LoadFromFile(const IXML_ST* pPath){
	CFileStream* pFileStream;
	pFileStream = new CFileStream();
	CFunctionTime xlff("CXML::LoadFromFile");
#ifdef MBCS_XML
	CStringW path;
#endif //MBCS_XML
	if (!pFileStream)
		return false;
	AREF(pFileStream, false);
	{
#ifdef MBCS_XML
		path = pPath;
		if (!pFileStream->Open(path, false, true))
			return false;
#else //MBCS_XML
		if (!pFileStream->Open(pPath, false, true))
			return false;
#endif //!MBCS_XML

		return LoadFromStream(pFileStream);
	}
	/*IXMLDOMDocument *pDoc;
	HRESULT hr;
	VARIANT_BOOL b;
	BSTR xml;
	static VARIANT v;
	Close();
	CoInitialize(NULL);
	hr = CoCreateInstance(CLSID_DOMDocument, NULL, CLSCTX_INPROC_SERVER | CLSCTX_LOCAL_SERVER, IID_IXMLDOMDocument, (void**)&pDoc);
	if (hr != S_OK)
	return false;

	AREF(pDoc, false);
	pDoc->put_async(VARIANT_FALSE);
	xml = SysAllocString(pPath);
	if (!xml)
	return false;
	v.vt = VT_BSTR;
	v.bstrVal = xml;
	hr = pDoc->load(v, &b);
	SysFreeString(xml);
	if (hr != S_OK || !b)
		return false;
	return Initialize(pDoc);*/
}

bool CXML::LoadFromStream(IStream* pStream){
	IXMLDOMDocument *pDoc;
	HRESULT hr;
	VARIANT_BOOL b;
	static VARIANT v;
	IXMLDOMParseError* pError;
	BSTR str;
	Close();
	STATSTG stat;
	//OleInitialize(NULL);
	CoInitialize(NULL);
	long fpos, line, linepos, error;
	AREF2(pStream);
	zmem_t(stat);
	if (pStream->Stat(&stat, STATFLAG_NONAME) != S_OK)
		stat.cbSize.QuadPart = 0;
	hr = CoCreateInstance(CLSID_DOMDocument, NULL, CLSCTX_INPROC_SERVER | CLSCTX_LOCAL_SERVER, IID_IXMLDOMDocument, (void**)&pDoc);
	if (hr != S_OK)
		return false;

	AREF(pDoc, false);
	pDoc->put_async(VARIANT_FALSE);
	//pDoc->put_validateOnParse(VARIANT_FALSE);
	v.vt = VT_UNKNOWN;
	v.punkVal = pStream;
	hr = pDoc->load(v, &b);
	if (hr != S_OK){
		if (pDoc->get_parseError(&pError) == S_OK){
			if (pError->get_errorCode(&error) != S_OK)
				error = 0;
			if (pError->get_filepos(&fpos) != S_OK)
				fpos = 0;
			if (pError->get_line(&line) != S_OK)
				line = 0;
			if (pError->get_linepos(&linepos) != S_OK)
				linepos = 0;
			if (pError->get_reason(&str) == S_OK)
				m_Error.Print(IXML_STR("XML parse error %.8X at line %u, position %u: %s"), error, line, linepos, IXML_SC(str)());
			else
				m_Error.Print(IXML_STR("XML parse error %.8X at line %u, position %u"), error, line, linepos, IXML_SC(str)());
			pError->Release();
		}
	}
	if (hr != S_OK || !b)
		return false;
	
	//Prealloc string heap to size of XML, minimizing buffer allocations while reading
	if (stat.cbSize.LowPart > 0)
		m_Heap.PreallocBuffer((int)stat.cbSize.LowPart);
	return Initialize(pDoc);
}

void CXML::Close(){
	FreeData();
}

IXMLData* CXML::GetRoot() const{
	//((CXMLData*)&m_Root)->AddRef();
	return (IXMLData*)&m_Root;
}

const IXML_ST* CXML::GetXMLError() const {
	return m_Error;
}

CXML::CXML() : m_Root(this){
	m_RefCount = 1;
#ifdef _DEBUG
	s_Count++;
#endif //_DEBUG
}

CXML::~CXML(){
	FreeData();
#ifdef _DEBUG
	s_Count--;
#endif //_DEBUG
}

void CXML::FreeData(){
	m_Root.FreeData();
}

bool CXML::Initialize(IXMLDOMDocument* pDoc){
	HRESULT hr;
	IXMLDOMElement* pElem;
	CXMLNSStack stack;
	AREF2(pDoc);
	hr = pDoc->get_documentElement(&pElem);
	if (hr != S_OK || !pElem)
		return false;
	AREF(pElem, false);
	m_Root.LoadFromNode(pElem, NULL, &stack);
	return true;
}

bool CXML::InitializeFromNode(IXMLDOMNode* pNode){
	CXMLNSStack stack;
	return m_Root.LoadFromNode(pNode, NULL, &stack);
}

#ifdef _DEBUG
ULONG	CXML::s_Count = 0;
#endif //_DEBUG

/*
Misc
*/

IXML * __stdcall CreateIXML(){
	return new CXML;
}

void __stdcall GetCounts(ULONG *_CXML, ULONG *_CXMLData, ULONG *_CXMLResultSet){
#ifdef _DEBUG
	if (_CXML)
		*_CXML = CXML::s_Count;

	if (_CXMLData)
		*_CXMLData = CXMLData::s_Count;

	if (_CXMLResultSet)
		*_CXMLResultSet = CXMLResultSet::s_Count;
#else //#ifdef _DEBUG
	if (_CXML)
		*_CXML = 0;

	if (_CXMLData)
		*_CXMLData = 0;

	if (_CXMLResultSet)
		*_CXMLResultSet = 0;
#endif //!#ifdef _DEBUG
}

bool CreateIXMLFromNode(IXMLDOMNode* pNode, IXML** ppXML){
	CXML *pXML;
	if (!pNode || !ppXML)
		return false;
	*ppXML = NULL;
	pXML = new CXML();
	if (!pXML)
		return false;
	if (!pXML->InitializeFromNode(pNode))
		return pXML->Release(), false;
	*ppXML = pXML;
	return true;
}

bool CreateIXMLFromChildNode(IXMLDOMNode* pNode, IXML** ppXML){
	IXMLDOMNode* pParent;
	if (!pNode || !ppXML)
		return false;
	*ppXML = NULL;
	if (pNode->get_parentNode(&pParent) != S_OK)
		return false;
	return CreateIXMLFromNode(pParent, ppXML) ? (/*pParent->Release(),*/ true) : (pParent->Release(), false);
}

class CXMLAttr;

class CXMLNode : public IXMLNode {
public:
	~CXMLNode();
	bool		m_IsRoot;
	CXMLNode*	m_pFirstChild;
	CXMLAttr*	m_pFirstAttribute;
	IXML_SC		m_Name;
	IXML_SC		m_Text;
	CXMLNode*	m_pNext;
public:
	CXMLNode();
	virtual void DeleteThis();
	virtual const IXML_ST* GetName() const;
	virtual const IXML_ST* GetText() const;
	virtual const IXMLNode* GetNext() const;
	virtual const IXMLNode* GetFirstChild() const;
	virtual const IXMLAttr* GetFirstAttribute() const;

public:
	void LoadFromData(IXMLData* pData);
	void SetIsRoot(bool root);
};

class CXMLAttr : public IXMLAttr {
	friend class CXMLNode;
public:
	~CXMLAttr();
	CXMLNode*	m_pFirstChild;
	CXMLAttr*	m_pNext;
	CXMLNode*	m_pNode;
	IXML_SC		m_Name;
	IXML_SC		m_Text;
public:
	CXMLAttr();
	virtual IXMLNode* GetNode() const;
	virtual const IXML_ST* GetName() const;
	virtual const IXML_ST* GetText() const;
	virtual const IXMLAttr* GetNext() const;
	virtual const IXMLNode* GetFirstChild() const;
public:
	void LoadFromData(IXMLData* pData);
};

IXMLNode* CreateIXMLNode(const IXML_ST* pPath){
	IXML* pXML;
	IXMLData* pData;
	CXMLNode* pRootNode;

	pXML = CreateIXML();
	if (!pXML)
		return NULL;
	{
		AREF3(pXML);
		if (!pXML->LoadFromFile(pPath)){
			return NULL;
		}
		pRootNode = new CXMLNode();
		pRootNode->SetIsRoot(true);
		pData = pXML->GetRoot();
		{
			AREF2(pData);
			pRootNode->LoadFromData(pData);
		}
		return pRootNode;
	}
}

/*
CXMLNode
*/

CXMLNode::~CXMLNode(){
	if (m_pFirstChild)
		delete m_pFirstChild;
	if (m_pFirstAttribute)
		delete m_pFirstAttribute;
	if (m_pNext)
		delete m_pNext;
}

CXMLNode::CXMLNode(){
	m_pFirstAttribute = NULL;
	m_pFirstChild = NULL;
	m_IsRoot = false;
}

void CXMLNode::DeleteThis(){
	if (m_IsRoot)
		delete this;
}

const IXML_ST* CXMLNode::GetName() const{
	return m_Name;
}

const IXML_ST* CXMLNode::GetText() const{
	return m_Text;
}

const IXMLNode* CXMLNode::GetNext() const{
	return m_pNext;
}

const IXMLNode* CXMLNode::GetFirstChild() const{
	return m_pFirstChild;
}

const IXMLAttr* CXMLNode::GetFirstAttribute() const{
	return m_pFirstAttribute;
}

void CXMLNode::LoadFromData(IXMLData* pData){
	IXMLData* pChild;
	CXMLNode* pNode, *pPrevious;
	CXMLAttr* pAttr, *pPreviousAttr;
	pPrevious = m_pFirstChild;
	while (pPrevious && pPrevious->m_pNext)
		pPrevious = pPrevious->m_pNext;
	pPreviousAttr = m_pFirstAttribute;
	while (pPreviousAttr && pPreviousAttr->m_pNext)
		pPreviousAttr = pPreviousAttr->m_pNext;

	{
		AREF2(pData);
		m_Name = pData->GetName();
		if (pData->GetType() == NODE_TEXT){
			m_Text = pData->GetValue();
		} else if (pData->GetFirstChild()){
			{
				for (pChild = pData->GetFirstChild(); pChild; pChild = pChild->GetNext()){
					AREF2(pChild);
					pNode = new CXMLNode;
					if (!pNode)
						continue;
					if (pPrevious){
						pPrevious->m_pNext = pNode;
						pPrevious = pNode;
					} else
						m_pFirstChild = pNode;
					pNode->LoadFromData(pChild);
				}
			}
		}
		if (pData->GetFirstAttribute()){
			{
				for (pChild = pData->GetFirstAttribute(); pChild; pChild = pChild->GetNext()){
					AREF2(pChild);
					pAttr = new CXMLAttr;
					if (!pAttr)
						continue;
					if (pPreviousAttr){
						pPreviousAttr->m_pNext = pAttr;
						pPreviousAttr = pAttr;
					} else
						m_pFirstAttribute = pAttr;
					pAttr->LoadFromData(pChild);
				}
			}
		}
	}
}

void CXMLNode::SetIsRoot(bool root){
	m_IsRoot = root;
}

/*
CXMLAttr
*/

CXMLAttr::CXMLAttr(){
	m_pNext = NULL;
	m_pFirstChild = NULL;
}

CXMLAttr::~CXMLAttr(){
	if (m_pNext)
		delete m_pNext;
	if (m_pFirstChild)
		delete m_pFirstChild;
}

IXMLNode* CXMLAttr::GetNode() const{
	return m_pNode;
}

const IXML_ST* CXMLAttr::GetName() const{
	return m_Name;
}

const IXML_ST* CXMLAttr::GetText() const{
	return m_Text;
}

const IXMLAttr* CXMLAttr::GetNext() const{
	return m_pNext;
}


const IXMLNode* CXMLAttr::GetFirstChild() const{
	return m_pFirstChild;
}

void CXMLAttr::LoadFromData(IXMLData* pData){
	IXMLData* pChild;
	CXMLNode* pNode, *pPrevious;
	pPrevious = m_pFirstChild;
	while (pPrevious && pPrevious->m_pNext)
		pPrevious = pPrevious->m_pNext;

	{
		AREF2(pData);
		m_Name = pData->GetName();
		if (pData->GetType() == NODE_TEXT){
			m_Text = pData->GetValue();
		} else if (pData->GetFirstChild()){
			{
				for (pChild = pData->GetFirstChild(); pChild; pChild = pChild->GetNext()){
					AREF2(pChild);
					pNode = new CXMLNode;
					if (!pNode)
						continue;
					if (pPrevious){
						pPrevious->m_pNext = pNode;
						pPrevious = pNode;
					} else
						m_pFirstChild = pNode;
					pNode->LoadFromData(pChild);
				}
			}
		}
		/*if (pData->GetFirstAttribute()){
			{
				for (pChild = pData->GetFirstAttribute(); pChild; pChild = pChild->GetNext()){
					AREF2(pChild);
					pAttr = new CXMLAttr;
					if (!pAttr)
						continue;
					if (pPreviousAttr){
						pPreviousAttr->m_pNext = pAttr;
						pPreviousAttr = pAttr;
					} else
						m_pFirstAttribute = pAttr;
					pAttr->LoadFromData(pChild);
				}
			}
		}*/
	}
}