
#ifndef _XML_H_
#define _XML_H_

#include <windows.h>
#include "Memory.h"
#include "BaseClasses.h"
#include <stdio.h>
#include <msxml.h>

//When defined, it uses UTF8 strings instead of UTF16 strings in the IXML classes
#define MBCS_XML

#ifdef MBCS_XML
class CStringA;
#define IXML_ST			char
#define IXML_STI		wchar_t
#define IXML_STP		char*
#define IXML_SC			CStringA
#define IXML_STR(_str)	_str
#define IXML_SCAN		sscanf_s
#define IXML_COMPARE	strcmp
#define IXML_COMPAREI	_stricmp
#define IXML_HEAP		CStringHeapA
#define IXML_HANDLE		CStringHandleA
#define IXML_CONVERT	TStringBufferA<128>
#else //MBCS_XML
class CStringW;
#define IXML_ST			wchar_t
#define IXML_STI		char
#define IXML_STP		wchar_t*
#define IXML_SC			CStringW
#define IXML_STR(_str)	L ## _str
#define IXML_SCAN		swscanf_s
#define IXML_COMPARE	wcscmp
#define IXML_COMPAREI	_wcsicmp
#define IXML_HEAP		CStringHeapW
#define IXML_HANDLE		CStringHandleW
#define IXML_CONVERT
#endif //!MBCS_XML
#define IXML_CSTP		const IXML_STP


interface IXMLData;

typedef bool (__stdcall *pfnEnumCallback)(IXMLData* node, void* ctx);

interface IXMLResultSet : public IConstUnknown {
public:
	virtual ULONG AddRef() = 0;
	virtual ULONG Release() = 0;
	virtual int GetResultCount() = 0;
	virtual IXMLData* GetNode(int i) = 0;
	virtual void Enum(pfnEnumCallback cb, void* ctx = NULL) = 0;
	template<typename T> inline bool EnumT(bool (__stdcall *pfnCallback)(IXMLData* node, T* ctx), T *ctx){
		return Enum((pfnEnumCallback)pfnCallback, (void*)ctx);
	}
};

typedef bool (_stdcall *pfnFilterProc)(IXMLData* pNode, void* ctx);


interface IXMLData : public IConstUnknown {
public:
	virtual ULONG AddRef() = 0;
	virtual ULONG Release() = 0;
	virtual const IXML_ST* GetName() = 0;
	virtual const IXML_ST* GetValue() = 0;
	virtual const IXML_ST* GetNamespace() = 0;
	virtual const IXML_ST* GetURI() = 0;
	virtual const IXML_ST* GetFullName() = 0;
	virtual DOMNodeType GetType() = 0;
	virtual IXMLData* GetNext() = 0;
	virtual IXMLData* GetFirstChild() = 0;
	virtual IXMLData* GetFirstAttribute() = 0;
	virtual IXMLResultSet* FindChildren(const IXML_ST* pName) = 0;
	virtual IXMLResultSet* FindAttributes(const IXML_ST* pName) = 0;
	virtual IXMLData* GetNamedAttribute(const IXML_ST* pName, IXMLData* pPrevious) = 0;
	virtual IXMLData* GetNamedNode(const IXML_ST* pName, IXMLData* pPrevious) = 0;
	virtual IXMLResultSet* FindFiltered(pfnFilterProc pfnFilter, void *ctx = NULL) = 0;
	virtual IXMLResultSet* FindFilteredAttributes(pfnFilterProc pfnFilter, void *ctx = NULL) = 0;
	template<typename T> inline IXMLResultSet* FindFilteredT(bool (__stdcall *pfnFilter)(const IXMLData* pNode, T* ctx), T *ctx){
		return FindFiltered((pfnFilterProc)pfnFilter, (void*)ctx);
	}
	template<typename T> inline IXMLResultSet* FindFilteredAttributesT(bool (__stdcall *pfnFilter)(const IXMLData* pNode, T* ctx), T *ctx){
		return FindFilteredAttributes((pfnFilterProc)pfnFilter, (void*)ctx);
	}
};

class CXMLFilter{
protected:
	static bool _stdcall FilterCallback(IXMLData* pNode, void* ctx){
		return ((CXMLFilter*)ctx)->Filter(pNode);
	}
public:
	inline CXMLFilter(){};
	virtual ~CXMLFilter(){};
	virtual bool Filter(IXMLData* pNode) = 0;
	inline IXMLResultSet* Run(IXMLData* pNode){
		return pNode->FindFiltered(FilterCallback, this);
	}
};

class CXMLAttributeFilter{
protected:
	static bool _stdcall FilterCallback(IXMLData* pNode, void* ctx){
		return ((CXMLAttributeFilter*)ctx)->Filter(pNode);
	}
public:
	inline CXMLAttributeFilter(){};
	virtual ~CXMLAttributeFilter(){};
	virtual bool Filter(IXMLData* pNode) = 0;
	inline IXMLResultSet* Run(IXMLData* pNode){
		return pNode->FindFilteredAttributes(FilterCallback, this);
	}
};

interface IXML : public IConstUnknown {
public:
	virtual ULONG AddRef() = 0;
	virtual ULONG Release() = 0;
	virtual bool LoadFromFile(const IXML_ST* pPath) = 0;
	virtual bool LoadFromStream(IStream* pStream) = 0;
	virtual void Close() = 0;
	virtual IXMLData* GetRoot() const = 0;

	virtual const IXML_ST* GetXMLError() const = 0;
};

interface IXMLAttr;

interface IXMLNode {
public:
	virtual void DeleteThis() = 0;
	virtual const IXML_ST* GetName() const = 0;
	virtual const IXML_ST* GetText() const = 0;
	virtual const IXMLNode* GetNext() const = 0;
	virtual const IXMLNode* GetFirstChild() const = 0;
	virtual const IXMLAttr* GetFirstAttribute() const = 0;
};

interface IXMLAttr {
public:
	virtual IXMLNode* GetNode() const = 0;
	virtual const IXML_ST* GetName() const = 0;
	virtual const IXML_ST* GetText() const = 0;
	virtual const IXMLAttr* GetNext() const = 0;
	virtual const IXMLNode* GetFirstChild() const = 0;
};


IXML *__stdcall CreateIXML();

IXMLNode* CreateIXMLNode(const IXML_ST* pPath);

void __stdcall GetCounts(ULONG *_CXML, ULONG *_CXMLData, ULONG *_CXMLResultSet);
void ClearXMLStringCache();

bool CreateIXMLFromNode(IXMLDOMNode* pNode, IXML** ppXML);
bool CreateIXMLFromChildNode(IXMLDOMNode* pNode, IXML** ppXML);

#if 0
class CXMLAttribute;
class CXMLNode;

class CXMLAttribute : public AutoPoolClass<CXMLAttribute, 128> {
private:
	wchar_t*	m_pName;
	wchar_t*	m_pValue;
public:
	CXMLAttribute(){
		m_pName = m_pValue = NULL;
	}
	~CXMLAttribute(){
		Free();
	}
	void Free(){
		if (m_pName)
			g_pFreeer(m_pName);
		if (m_pValue)
			g_pFreeer(m_pValue);
		m_pName = m_pValue = NULL;
	}
	const wchar_t* GetName() const{
		return m_pName;
	}
	const wchar_t* GetValue() const{
		return m_pValue;
	}
	bool Initialize(IXMLDOMNode* pNode){
		BSTR str;
		Free();
		if (pNode->get_nodeName(&str) != S_OK)
			return false;
		m_pName = StrDupW(str);
		SysFreeString(str);

		if (pNode->get_text(&str) == S_OK){
			m_pValue = StrDupW(str);
			SysFreeString(str);
		}
		return true;
	}
};

class CXMLNode : public AutoPoolClass<CXMLNode, 64> {
private:
	SList<CXMLNode>			m_Children;
	SList<CXMLAttribute>	m_Attributes;
	wchar_t*				m_pTag;
	wchar_t*				m_pText;
private:
public:
	CXMLNode(){
		m_pTag = m_pText = NULL;
	}
	~CXMLNode(){
		Free();
	}
	bool LoadFromNode(IXMLDOMNode *pNode){
		BSTR str;
		HRESULT hr;
		CXMLAttribute* pAttribute;
		CXMLNode* pNewNode;
		IXMLDOMNode* pLocalNode;
		IXMLDOMNodeList* pList;
		IXMLDOMNamedNodeMap* pMap;
		DOMNodeType type;
		VARIANT v;
		Free();
		if (!pNode)
			return false;
		AREF2(pNode);
		hr = pNode->get_nodeType(&type);
		if (hr != S_OK)
			return false;

		if (pNode->get_nodeName(&str) != S_OK)
			return false;
		m_pTag = StrDupW(str);
		SysFreeString(str);

		if (type == NODE_TEXT && pNode->get_nodeValue(&v) == S_OK){
			m_pText = StrDupW(v.bstrVal);
			SysFreeString(v.bstrVal);
		}

		wprintf(L"tag: %s\n", m_pTag);

		if (pNode->get_childNodes(&pList) == S_OK){
			AREF(pList, false);
			while (pList->nextNode(&pLocalNode) == S_OK){
				AREF(pLocalNode, false);
				pNewNode = new CXMLNode();
				if (!pNewNode)
					continue;
				pNewNode->LoadFromNode(pLocalNode);
				m_Children.Add_Head(pNewNode);
			}
		}

		if (pNode->get_attributes(&pMap) != S_OK)
			return true;

		AREF(pMap, false);

		while (pMap->nextNode(&pLocalNode) == S_OK){
			AREF(pLocalNode, false);
			pAttribute = new CXMLAttribute();
			if (!pAttribute)
				continue;
			pAttribute->Initialize(pLocalNode);
			m_Attributes.Add_Head(pAttribute);
		}

		return true;
	}
	void Free(){
		GenericSLNode<CXMLNode>	*pChildNode;
		GenericSLNode<CXMLAttribute> *pAttributeNode;
		if (m_pTag)
			g_pFreeer(m_pTag);
		if (m_pText)
			g_pFreeer(m_pText);
		m_pTag = m_pText = NULL;

		pChildNode = m_Children.Get_Head();
		while (pChildNode){
			delete pChildNode->m_pData;
			pChildNode = pChildNode->m_pNext;
		}
		m_Children.Remove_All();

		pAttributeNode = m_Attributes.Get_Head();
		while (pAttributeNode){
			delete pAttributeNode->m_pData;
			pAttributeNode = pAttributeNode->m_pNext;
		}
		m_Attributes.Remove_All();

	}
	SList<CXMLNode>* GetChildren(){
		return &m_Children;
	}
	SList<CXMLAttribute>* GetAttributes(){
		return &m_Attributes;
	}
	const wchar_t* GetName() const{
		return m_pTag;
	}
	const wchar_t* GetText() const{
		return m_pText;
	}
	const CXMLAttribute* GetAttributeByName(const wchar_t* pName){
		GenericSLNode<CXMLAttribute> *pNode;
		pNode = m_Attributes.Get_Head();
		while (pNode){
			if (_wcsicmp(pName, pNode->m_pData->GetName()) == 0)
				return pNode->m_pData;
			pNode = pNode->m_pNext;
		}
		return NULL;
	}
	const wchar_t* GetAttributeValueByName(const wchar_t* pName){
		GenericSLNode<CXMLAttribute> *pNode;
		pNode = m_Attributes.Get_Head();
		while (pNode){
			if (_wcsicmp(pName, pNode->m_pData->GetName()) == 0)
				return pNode->m_pData->GetValue();
			pNode = pNode->m_pNext;
		}
		return NULL;
	}
};

class CXML {
private:
	CXMLNode	m_Root;
	bool InitializeRoot(IXMLDOMDocument* pDoc){
		HRESULT hr;
		IXMLDOMNodeList *pList;
		IXMLDOMElement* pElem;
		IXMLDOMNode* pNode;
		DOMNodeType type;
		AREF2(pDoc);
		hr = pDoc->get_documentElement(&pElem);
		if (hr != S_OK || !pElem)
			return false;
		AREF(pElem, false);
		m_Root.LoadFromNode(pElem);
		/*hr = pDoc->get_childNodes(&pList);
		if (hr != S_OK)
			return false;
		AREF(pList, false);

		while (pList->nextNode(&pNode) == S_OK){
			AREF(pNode, false);
			pNode->get_nodeType(&type);
			if (type == NODE_DOCUMENT){
				m_Root.LoadFromNode(pNode);
				break;
			}
		}*/
		return true;
	}
public:
	CXML(){
	}
	~CXML(){
		Free();
	}
	void Free(){
		m_Root.Free();
	}
	bool LoadFromFile(const wchar_t *pFile){
		IXMLDOMDocument *pDoc;
		HRESULT hr;
		VARIANT_BOOL b;
		BSTR xml;
		VARIANT v;
		Close();
		CoInitialize(NULL);
		hr = CoCreateInstance(CLSID_DOMDocument, NULL, CLSCTX_INPROC_SERVER | CLSCTX_LOCAL_SERVER, IID_IXMLDOMDocument, (void**)&pDoc);
		if (hr != S_OK)
			return false;

		AREF(pDoc, false);
		pDoc->put_async(VARIANT_FALSE);
		xml = SysAllocString(pFile);
		if (!xml)
			return false;
		v.vt = VT_BSTR;
		v.bstrVal = xml;
		hr = pDoc->load(v, &b);
		SysFreeString(xml);
		if (hr != S_OK || !b)
			return false;

		return InitializeRoot(pDoc);
	}
	bool LoadFromStream(IStream* pStream){
	}
	void Close(){
		Free();
	}
	CXMLNode *GetRoot(){
		return &m_Root;
	}
};
#endif //0


#endif //!_XML_H_
