#ifndef _HTMLWRITE_H_
#define _HTMLWRITE_H_
#include <windows.h>
#include "StdLib.h"
#include "Memory.h"
#include "MemoryStream.h"
#include "Lists.h"
#include "String.h"
#include "StringBuffer.h"


class CXMLAttribute : public AutoPoolClass<CXMLAttribute, 16> {
public:
	CStringA	m_Name;
	CStringA	m_String;
};

class CXMLTag : public AutoPoolClass<CXMLTag, 16> {
public:
	CStringA	m_Tag;
	bool		m_HasChildren;
	CXMLTag(const char* pTag){
		m_HasChildren = false;
		m_Tag = pTag;
	}
};

typedef struct _HTMLEntity {
	wchar_t			m_Char;
	const wchar_t*	m_pName;
	const char*		m_pEntitized;
} HTMLEntity, *lpHTMLEntity;


class CXMLWriter {
private:
	CStringBufferA*			m_pOutput;
	SList<CXMLAttribute>	m_Attributes;
	SList<CXMLTag>			m_Tags;
	CStringBufferA			m_Buffer;
	int						m_Level;
	bool WriteAttributes();
	bool WriteTabs();
public:
	CXMLWriter(CStringBufferA* pOutput);
	~CXMLWriter();
	void FreeData();
	bool AddAttribute(const char* pName, const char* pString);
	bool AddAttributeFmt(const char* pName, const char* pFmt, ...);
	bool OpenTag(const char* pTag);
	bool OpenAttrTag(const char* pTag, ...);
	bool WriteProcessingInstruction(const char* pInstruction);
	bool WriteXMLStart(const char* version = "1.0");
	bool WriteClosedTag(const char* pTag);
	bool WriteClosedAttrTag(const char* pTag, ...);
	bool WriteUnclosedTag(const char* pTag);
	bool CloseTag();
	bool CloseAllTags();
	
	bool WriteRaw(const char* pString);
	bool WriteRawFmt(const char *pFmt, ...);
	bool WriteString(const char* pString);
	bool WriteStringFmt(const char *pFmt, ...);
	bool WriteTaggedString(const char* pTag, const char* pString);
	bool WriteTaggedStringFmt(const char* pTag, const char *pFmt, ...);
	bool Finalize();

public:
	class AutoTag {
	private:
		CXMLWriter*	m_pWriter;
	public:
		AutoTag(const char* pTag, CXMLWriter* pWriter);
		~AutoTag();
	};
};


#define XMLBEGIN(_this) \
{ \
CXMLWriter* __XML__this_ptr = (_this);

#define XMLGET() \
	(__XML__this_ptr)



#define XMLTAG(tag) \
	CXMLWriter::AutoTag COMBINEDYN(__xml_autotag_, __COUNTER__)(tag, __XML__this_ptr);

#define HTAG(tag) \
	CXMLWriter::AutoTag COMBINEDYN(__xml_autotag_, __COUNTER__)(tag, __XML__this_ptr);


#define HTMLATTRIBUTE(name, val, ...) \
	__XML__this_ptr->AddAttribute(attribute, fmt);

#define HTMLATTRIBUTEF(name, fmt, ...) \
	__XML__this_ptr->AddAttributeFmt(attribute, fmt, __VA_ARGS__);


#define HATTR(name, val, ...) \
	__XML__this_ptr->AddAttribute(attribute, fmt, __VA_ARGS__);

#define HATTRF(name, fmt, ...) \
	__XML__this_ptr->AddAttributeFmt(attribute, fmt, __VA_ARGS__);

#define HTMLWRITEPROCESSINGINSTRUCTION(inst) \
	__XML__this_ptr->WriteProcessingInstruction(inst);

#define HTMLWPI(inst) \
	__XML__this_ptr->WriteProcessingInstruction(inst);

#define HTMLCLOSEDTAG(tag) \
	__html__this_ptr->WriteClosedTag(tag);

#define HTMLUNCLOSEDTAG(tag) \
	__XML__this_ptr->WriteUnclosedTag(tag);

#define HCTAG(tag) \
	__XML__this_ptr->WriteClosedTag(tag);

#define HUTAG(tag) \
	__html__this_ptr->WriteUnclosedTag(tag);


#define XMLEND() \
}

#define XMLEND2() \
	__XML__this_ptr->Finalize(); \
}

#endif //!_HTMLWRITE_H_
